require "yaml"
require "classes/compatibility/yaml_loader.rb"

class DftFunctional
	DFT_FUNC_TYPES = [
		:local,
		:gga,
		:meta_gga,
		:hybrid,
		:meta_hybrid,
		:range_separated,
		:double_hybrid,
		:double_hybrid_rs
	]

	attr_accessor :name
	attr_accessor :type
	attr_accessor :progname
	attr_accessor :comment

	def initialize(name, type)
		@name = name.downcase.gsub('-','')
		raise "Unknown type of DFT functional" unless DFT_FUNC_TYPES.include? type
		@type = type
	end
end

class DftSetup

	# List of functionals known to cuby
	# lower case, without "-" characters
	FUNCTIONAL_TYPE = {
		# Local
		"hfs"		=> :local,		# Hartree-Fock Slater
		"xalpha"	=> :local,		# The famous old Slater Xa theory
		"lsd"		=> :local,		# Local spin density (VWN-5A form)
		"vwn3"		=> :local,		# Local spin density (VWN-3)
		"vwn5"		=> :local,		# Local spin density (VWN-5)
		"vwn"		=> :local,		# Synonym for VWN5
		"pwlda"		=> :local,		# Local spin density (PW-LDA)
		# GGA
		"bnull"		=> :gga,		# Becke '88 exchange, no corr.
		"bvwn"		=> :gga,		# Becke '88 exchane, VWN-5 corr.
		"bp"		=> :gga,		# Becke X-Perdew 86 correlation
		"bp86"		=> :gga,		# Synonym for BP
		"pw91"		=> :gga,		# Perdew-Wang GGA-II '91 func.
		"mpwpw"		=> :gga,		# Modified PW with PW correlation
		"mpwlyp"	=> :gga,		# same with LYP correlation
		"blyp"		=> :gga,		# Becke X with LYP correlation
		"gp"		=> :gga,		# Gill '96 X, Perdew '86 corr.
		"glyp"		=> :gga,		# Gill '96 X with LYP correlation
		"pbe"		=> :gga,		# Perdew-Burke-Ernzerhof
		"revpbe"	=> :gga,		# Revised PBE (exchange scaling)
		"rpbe"		=> :gga,		# Revised PBE (functional form of X)
		"pwp"		=> :gga,		# PW91 exchange + P86 correlation
		"olyp"		=> :gga,		# the optimized exchange and LYP
		"opbe"		=> :gga,		# the optimized exchange and PBE
		"xlyp"		=> :gga,		# the Xu/Goddard exchange and LYP
		"b97d"		=> :gga,
		# Meta-GGA
		"tpss"		=> :meta_gga,		# the TPPS functional
		"scan"		=> :meta_gga,		# The Strongly Constrained and Appropriately Normed (SCAN) meta-GGA functional
		"m06l"		=> :meta_gga,
		# Hybrid
		"b1lyp"		=> :hybrid,		# One parameter Hybrid of BLYP
		"b3lyp"		=> :hybrid,		# Three parameter Hybrid of BLYP
		"b3lyp/g"	=> :hybrid,		# B3LYP as implemented in Gaussian
		"b1p"		=> :hybrid,		# Analogous with Perdew exchange
		"b3p"		=> :hybrid,		# Analogous with Perdew exchange
		"b3pw"		=> :hybrid,		# The 3-parameter hybrid version of PW91
		"g1lyp"		=> :hybrid,		# 1 par. analog with Gill 96 X
		"g3lyp"		=> :hybrid,		# 3 par. analog with Gill 96 X
		"g1p"		=> :hybrid,		# similar with P correlation
		"g3p"		=> :hybrid,		# similar with P correlation
		"pbe0"		=> :hybrid,		# 1 parameter version of PBE
		"pwp1"		=> :hybrid,		# 1 parameter version of PWP
		"mpw1pw"	=> :hybrid,		# 1 parameter version of mPWPW
		"mpw1lyp"	=> :hybrid,		# 2 parameter version of mPWLYP
		"pw91_1"	=> :hybrid,		# 1 parameter version of PW91
		"o3lyp"		=> :hybrid,		# 3 parameter version of OLYP
		"x3lyp"		=> :hybrid,		# 3 parameter version of XLYP
		"pw6b95"	=> :hybrid,		# Hybrid functional by Truhlar
		"bhlyp"		=> :hybrid,		# Becke Half and Half
		# Meta-hybrid
		"tpssh"		=> :meta_hybrid,	# hybrid version of TPSS with 10% HF exchange
		"tpss0"		=> :meta_hybrid,	# hybrid version of TPSS with 25% HF exchange
		"m06"		=> :meta_hybrid,	# Minnesota M06 hybrid meta-GGA (27% HF exchange)
		"m062x"		=> :meta_hybrid,	# The M06-2X version with 54% HF exchange
		# Double-hybrid
		"b2plyp"	=> :double_hybrid,	# Grimme's 2006 double hybrid
		"mpw2plyp"	=> :double_hybrid,	# Schwabe/Grimme improved double hybrid
		"pwpb95"	=> :double_hybrid,	# New Grimme double hybrid
		"b2gpplyp"	=> :double_hybrid,	# Gershom Martin's "general purpose" reparameterization
		"b2kplyp"	=> :double_hybrid,	# Gershom Martin's "kinetic" reparameterization
		"b2tplyp"	=> :double_hybrid,	# Gershom Martin's "thermochemistry" reparameterization
		"dsdblyp"	=> :double_hybrid,	# Gershom Martin's "general purpose" functional
		"dsdpbep86"	=> :double_hybrid,	# Gershom Martin's "general purpose" functional
		"dsdpbeb95"	=> :double_hybrid,	# Gershom Martin's "general purpose" functional
		# Range-separated
		"camb3lyp"	=> :range_separated,	# CAM-B3LYP
		"lcblyp"	=> :range_separated,	# LC-BLYP
		"wb97"		=> :range_separated,	# Head-Gordon's fully variable DF wB97
		"wb97x"		=> :range_separated,	# Head-Gordon's DF wB97X with minimal Fock exchange
		"wb97xd3"	=> :range_separated	# Chai's refit incl. D3 correction
	}

	def initialize(functionals_file, grids_file = nil)
		#: Reads yaml files with definitions for each interface
		f = File.open(functionals_file)
		@functionals = YAMLLoader.load(f, [DftFunctional, Symbol])
		f.close
		
		if grids_file
			f = File.open(grids_file)
			@grids = YAMLLoader.load(f)
			f.close
		end
	end

	def functional(settings)
		#: Returns the functional name in format used by the interfaced program
		# Get cuby functional name, without check
		func = functional_internal(settings, false)
		return settings[:functional_custom] if func == 'custom'

		if @functionals.kind_of?(Hash)
			# Old format hash cubyname => progname
			# Check if cuby knows the functional
			func = functional_internal(settings, true)
			# Check whether it is available in the external program
			unless @functionals.has_key?(func)
				Cuby::error("Selected DFT functional not available in the interfaced program")
			end
			return @functionals[func]
		elsif @functionals.kind_of?(Array)
			# New format - avoids built-in database
			return functional_list_search(func).progname
		else
			Cuby::error "Functional definition file in unknown format"
		end
	end

	def grid(settings)
		#: Returns the functional name in format used by the interfaced program
		if settings[:dft_grid] == :custom
			return settings[:dft_grid_custom]
		else
			# Easy: the allowed values are checked by input parser
			return @grids[settings[:dft_grid]]
		end
	end

	def functional_type(settings)
		func = functional_internal(settings, false)
		if @functionals.kind_of?(Hash)
			if func == 'custom'
				return settings[:functional_custom_type]
			end
			return FUNCTIONAL_TYPE[func]
		elsif @functionals.kind_of?(Array)
			# New format - avoids built-in database
			return functional_list_search(func).type
		else
			Cuby::error "Functional definition file in unknown format"
		end
		
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def functional_list_search(name)
		@functionals.each{|func|
			return func if func.name == name
		}
		Cuby::error("Selected DFT functional '#{name}' not available in the interfaced program")
	end

	def functional_internal(settings, check = true)
		#: Returns the functional name in format used by cuby
		if settings[:functional].downcase == "custom"
			return 'custom'
		end

		# Remove extra characters, downcase
		func = settings[:functional].strip.downcase.gsub("-","")
		# Check if cuby knows it
		if check
			unless FUNCTIONAL_TYPE.has_key?(func)
				Cuby::error("Selected DFT functional not recognized by cuby\nif you are sure you want to use it, use the 'custom' option")
			end
		end
		return func
	end
end
