# Cuby-style atom type class
# It is strored as a string but it has some additional methods

# The atom types are built as the atom's element name, colon, and list of elements
# bonded to this atom. The list is ordered by proton number.

# Examples:
#   Hydrogen in water is H:O
#   Oxygen in water is O:H2
#   Carbon in methanol C:H3O

class AtomtypeCuby < String
	def initialize(str)
		super(str)
	end

	def match_exact(pattern)
		return self == pattern
	end

	def match_wildcard(pattern)
		pattern = AtomtypeCuby.new(pattern)

		# Check pattern
		unless pattern =~ /^[^:]+:[^:]*$/
			Cuby::error "Atomtype selection must use pattern element:bonded_elements"
		end

		# Match the element: wildcard or exact match
		pattern_e = pattern.get_element
		match_ele = (pattern_e == "*") || (pattern_e == get_element)

		# Match the neighbours
		match_bonded = false
		bonded = get_bonded_array
		bonded_pattern = pattern.get_bonded_array
		
		# Number of records must be the same
		if bonded.size == bonded_pattern.size
			match_bonded = true
			# After removing actual atoms from the pattern...
			left_pat = bonded_pattern - bonded
			# ... only the stars should be left to make valid match
			left_pat.each{|x|
				if x !~ /^\*/
					match_bonded = false
				end
			}
			
		end

		return match_ele && match_bonded
	end

	#=======================================================================
	# Helpers
	#=======================================================================
	
	def get_element
		ele, bonded = self.split(":")
		return ele
	end

	def get_bonded
		ele, bonded = self.split(":")
		return "" unless bonded # In case there's nothing behind :, return '' instead of nil
		return bonded
	end

	def get_bonded_array
		# Builds array of items with unique names (element + number)
		str = get_bonded
		if str =~ /\*\*/
			Cuby::error("Error parsing atom type '#{self}', there should be only one star,\npossibly followed by a number")
		end
		# Separate the element entries
		entries = []
		while str.size > 0
			if match = /^[A-Z*][a-z]*[0-9]*/.match(str)
				entries <<  match[0]
				str.gsub!(/^[A-Z*][a-z]*[0-9]*/,'')
			else
				Cuby::error("Error parsing atom type '#{self}'")
			end
		end
		# Expand numbers and number all atoms
		expanded = []
		entries.each{|entry|
			if match = /([A-Z*][a-z]*)([0-9]+)/.match(entry)
				name = match[1]
				count = match[2]
				count.to_i.times{ |i| expanded << name + i.to_s }
			else
				expanded << entry + "0"
			end
		}
		return expanded

	end
end
	
