#: Unambiguous canonicalization of molecular topology
#: Reference: David Weininger, Arthur Weininger and Joseph L. Wieninger, J. Chem. Inf. Comput. Sci. 29, 97-101, 1989.

#: The algorithm assigns an unique number to each atom in the molecule
#: (saved here in Atom.properties[:canonical_rank]). This analysis is based
#: purely on topology of the molecule, not the geometry, and thus allows
#: sorting the atoms in an unique way.

require "classes/math/prime_numbers"

module GeometryCanonicalize

	def assign_canonical_ranks(break_equivalence = :none) # => nil
		#: This method assign the canonical rank to each atom in the geometry
		#: (returned as an array). 
		#: The argument specifies how to handle topologicaly equivalent atoms:
		#* :none - same rank is assigned
		#* :order - ranked by their order in the geometry
		#* :geometry - ranked by their distance from molecular center (fails for symmetric molecules)

		# 0th cycle
		connectivity = Connectivity.new(self)
		canonical_invariant = atomic_invariants(connectivity) # calculate invariants
		prev_rank_count, canonical_rank = rank_from_property(canonical_invariant) #invariants_to_ranks
		# prime product
		canonical_primeproduct = prime_product_of_ranks(canonical_rank, connectivity)

		# iterate
		1000.times {|i|
			raise "Maximum of iterations exceeded" if i == 999
			# new ranks
			rank_count, canonical_rank = rank_from_property(canonical_primeproduct)

			# check if invariant, repeat if not
			if prev_rank_count != rank_count
				prev_rank_count = rank_count
				# prime product
				canonical_primeproduct = prime_product_of_ranks(canonical_rank, connectivity)
				next
			end

			# check if all ranks are unique, break ties if not
			if rank_count < self.size
				case break_equivalence
				when :order
					canonical_primeproduct = break_ties(canonical_rank)
					prev_rank_count = rank_count + 1
					next
				when :geometry
					canonical_primeproduct = break_ties_geo(canonical_rank)
					prev_rank_count = rank_count + 1
					next
				when :geometry_xyz
					canonical_primeproduct = break_ties_geo(canonical_rank, true)
					prev_rank_count = rank_count + 1
					next
				when :none
					# Do nothing
				else
					raise "Wrong equivalene beaking mode, use :order or :geometry"
				end
			end

			# everything is OK
			break
		}

		# Delete auxiliary info
		return canonical_rank
	end

	def assign_canonical_ranks!(break_equivalence = :none) # => nil
		ranks = assign_canonical_ranks(break_equivalence)
		each_with_index{|atom, i|
			atom.properties[:topo_rank] = ranks[i]
		}
		return nil
	end

	def sort_by_canonical_rank(break_equivalence = :geometry) # => Geometry
		#: Sort a geometry by the canonical ranks

		# Calculate ranks
		ranks = assign_canonical_ranks(break_equivalence)

		# Sort the ranks
		sorted = ranks.sort

		# New geometry
		new = Geometry.new

		# Fill with sorted atoms
		sorted.each{|rank|
			atom_index = ranks.index(rank)
			new << at(atom_index)
		}

		return new
	end

	def reorder_to_match(template, break_equivalence = :geometry) # => Geometry
		#: Reorder a geometry to use the order from another geometry

		# Calculate ranks
		ranks = assign_canonical_ranks(break_equivalence)
		new_order = template.assign_canonical_ranks(break_equivalence)

		# New template
		new = Geometry.new

		# Fill with new_order atoms
		new_order.each{|rank|
			atom_index = ranks.index(rank)
			new << at(atom_index)
		}

		return new
	end

	#=======================================================================
	# Private methods
	#=======================================================================

	def atomic_invariants(conn)
		#: Create initial atomic invariants from available information
		canonical_invariant = []

		each_with_index {|atom, at_i|

			# All connections
			connections = conn.hybridization(at_i)

			# Not hydrogens
			not_hydrogens = 0
			conn.bound_atoms(at_i).each{|i|
				not_hydrogens += 1 if at(i).element != :H
			}
			
			# Hydrogens
			hydrogens = connections - not_hydrogens

			# Atomic number
			atomic_number = atom.proton_number

			# Charge and charge sign
			charge = 0
			charge_sign = 0

			# Get the invariant
			string_form = sprintf("%1d%02d%02d%1d%1d%1d",
					      connections,
					      not_hydrogens,
					      atomic_number,
					      charge_sign,
					      charge,
					      hydrogens
					     )
			invariant = string_form.to_i

			# Save it
			canonical_invariant[at_i] = invariant
		}
		return canonical_invariant
	end
	private :atomic_invariants

	def rank_from_property(property_array)
		#: Recreate ranks from specified property of the atoms
		canonical_rank = []

		unique_inv = []
		each_with_index{|atom, at_i|
			unique_inv << property_array[at_i] unless unique_inv.include?(property_array[at_i])
		}
		unique_inv.sort!
		each_with_index{|atom, at_i|
			canonical_rank[at_i] =  unique_inv.index(property_array[at_i]) + 1
		}
		return [unique_inv.size, canonical_rank]
	end
	private :rank_from_property

	def prime_product_of_ranks(canonical_rank, connectivity)
		#: The disambiguation step - calculate a product of primes assigned to neighbor atom
		#: ranks

		canonical_primeproduct = []

		each_with_index {|atom, at_i|
			prod = 1
			connectivity.bound_atoms(at_i).each{|i|
				prod *= PrimeNumbers[canonical_rank[i]]
			}
			canonical_primeproduct[at_i] = prod
		}
		return canonical_primeproduct
	end
	private :prime_product_of_ranks

	def break_ties(canonical_rank)
		#: Break equivalence of ranks (one step). Starts with first atom of the lowest
		#: degenerate rank.

		canonical_primeproduct = []

		# get index of lowest identical rank
		ranks = canonical_rank.sort
		lowest = nil
		(ranks.size-2).times {|i|
			if ranks[i] == ranks[i+1]
				lowest = ranks[i]
				break
			end
		}
		lowest_i = canonical_rank.index(lowest)

		# double all ranks, saving them as prime product
		each_index{|at_i|
			canonical_primeproduct[at_i] = canonical_rank[at_i] * 2
		}

		# decrease rank of the saved atom
		canonical_primeproduct[lowest_i] -= 1

		return canonical_primeproduct
	end
	private :break_ties

	def break_ties_geo(canonical_rank, use_xyz = false)
		#: Break equivalence of ranks (one step), based on coordinates. Starts with first atom of the lowest
		#: degenerate rank.

		canonical_primeproduct = []

		# get index of lowest identical rank
		ranks = canonical_rank.sort
		lowest = nil
		(ranks.size-1).times {|i|
			if ranks[i] == ranks[i+1]
				lowest = ranks[i]
				break
			end
		}

		# List of atoms with lowest ranks
		lowest_list = []
		each_with_index{|atom, at_i|
			lowest_list << atom if canonical_rank[at_i] == lowest
		}

		# double all ranks, saving them as prime product
		each_index{|at_i|
			canonical_primeproduct[at_i] = canonical_rank[at_i] * 2
		}

		# Order the list by the selected criteria
		lowest_list.sort!{|a,b| break_ties_geo_compare(a,b, use_xyz)}

		# Select the first atom
		break_atom_i = self.index(lowest_list[0])
		# decrease rank of the selected atom
		canonical_primeproduct[break_atom_i] -= 1

		return canonical_primeproduct
	end
	private :break_ties_geo

	def break_ties_geo_compare(a,b, use_xyz)
		c = center
		dist_comp = (a - c).abs <=> (b - c).abs
		if dist_comp == 0
			if use_xyz
				x = a.x <=> b.x
				return x unless x == 0
				y = a.y <=> b.y
				return y unless y == 0
				z = a.z <=> b.z
				return z
			else
				Cuby::error "Atom equivalence can not be broken using geometry information, symmetry is to high"
			end
		else
			return dist_comp
		end
	end
	private :break_ties_geo_compare

end
