# PDB-specific methods

module GeometryPdb
	
	def is_pdb?
		return at(0).properties.has_key?(:pdb_res_name)
	end

	def residues_as_atomlists
		residues = []
		last = -1
		this_residue = nil
		each_with_index{|atom, i|
			if atom.properties[:pdb_res_no] != last
				# Create new residue
				this_residue = []
				residues << this_residue
				last = atom.properties[:pdb_res_no]
			end
			# Add the atom to the current residue
			this_residue << i
		}
		return residues
	end

	def residues_as_geometries
		residues = []
		last = -1
		this_residue = nil
		each_with_index{|atom, i|
			if atom.properties[:pdb_res_no] != last
				# Create new residue
				this_residue = Geometry.new
				residues << this_residue
				last = atom.properties[:pdb_res_no]
			end
			# Add the atom to the current residue
			this_residue << atom
		}
		return residues
	end

	def renumber_pdb_atoms!
		 each_with_index{|atom, i|
			 atom.properties[:pdb_atom_no] = i+1
		 }
		 return nil
	end

	def pdb_protein_fix_ters!
		res_no = nil
		each_with_index{|atom, i|
			if atom.properties[:pdb_atom_name] == "OXT"
				res_no = atom.properties[:pdb_res_no] 
			end
			if res_no && (atom.properties[:pdb_res_no] == res_no + 1)
				self[i-1].properties[:terafter] = true
				res_no = nil
			end
		}
		return nil
	end

	def renumber_pdb_residues!(start = 1) # => nil
		#: Fix numbering of residues in pdb. Number is increased when:
		#* original residue number changes
		#* residue name changes
		#* TER record is found
		rn = start -1
		last = -1
		last_terafter = false
		last_name = ''
		each{|atom|
			if atom.properties[:pdb_res_no] != last ||
			   atom.properties[:pdb_res_name] != last_name ||
			   last_terafter
				rn += 1
			end
			last = atom.properties[:pdb_res_no]
			last_name = atom.properties[:pdb_res_name]
			last_terafter = atom.properties[:terafter]
			atom.properties[:pdb_res_no] = rn
		}
		return nil
	end

	def pdb_count_residues(strict = false) # => Integer
		#: Returns number of residues in the system
		last_res_no = first.properties[:pdb_res_no] - 1
		count = 0
		each {|atom|
			rn = atom.properties[:pdb_res_no]
			step = rn - last_res_no
			if (step < 0 || step > 1) && strict
				Cuby::error "Error in pdb file, discontinuity in numbering of residues"
			end
			if step != 0
				count += 1
			end
			last_res_no = rn
		}
		return count
	end

	def pdb_name_atoms_by_num
		#: Assign unique names to atoms. The name is created from the element abbreviation and a number (up to two digits).
		#: If there is more than 99 atoms of the same elements, error is raised. 

		numbers = atom_numbering_by_element

		# Get max. number of atoms of one element
		maxnum = numbers.max
		if maxnum > 99
			Cuby::warning "Too many atoms of one element to fit PDB numbering range 1-99, extended format will be used"
			pdb_name_atoms_by_num_long
			return
		end

		each_index{|i|
			atom = at(i)
			e = atom.element
			atom.properties[:pdb_atom_name] = e.to_s + numbers[i].to_s
		}

		return nil
	end

	def pdb_name_atoms_by_num_long
		# Number the atoms using two-character alphanumeric code

		# base for the numbering system
		base = "0123456789abcdefghijklmnopqrstuvwxyz"

		numbers = atom_numbering_by_element
		# Get max. number of atoms of one element
		maxnum = numbers.max
		if maxnum >= base.size**2
			Cuby::error "Too many atoms of the elements to fit extended PDB numbering range 00-zz"
		end

		each_index{|i|
			atom = at(i)
			e = atom.element
			d1 = base[numbers[i] / base.size]
			d2 = base[numbers[i] % base.size]
			atom.properties[:pdb_atom_name] = e.to_s + d1 + d2
		}

		return nil
	end

    def pdb_disulfide_bridges(resnames = ["CYS","CYX"])
      bridges = []
      each_index{|i|
        atom1 = at(i)
        next unless atom1.element == :S
        next unless resnames.include?(atom1.properties[:pdb_res_name])
        i.times{|j|
          atom2 = at(j)
          next unless atom2.element == :S
          next unless resnames.include?(atom2.properties[:pdb_res_name])
          if atom1.distance(atom2) < 3.0
            bridges << [i,j]
          end
        }
      }
      return bridges
    end

	#=======================================================================
	# Protein charge calculation
	#=======================================================================

	PDB_RES_CHARGES = {
		# Neutral amino acids
		"ALA" => 0, "ASN" => 0, "CYS" => 0, "CYX" => 0,
		"GLN" => 0, "GLY" => 0, "HIS" => 0, "HIE" => 0,
		"HID" => 0, "ILE" => 0, "LEU" => 0,        
		"MET" => 0, "PHE" => 0, "PRO" => 0, "SER" => 0, "THR" => 0,
		"TRP" => 0, "TYR" => 0, "VAL" => 0,
		"ASH" => 0, "GLH" => 0,
		# Cationic AA
		"ARG" => 1, "LYS" => 1, "HIP" => 1,
		# Anionic AA
		"ASP" => -1, "GLU" => -1,
		# Cuby caps
		"BCC" => 0, "BNC" => 0, "BCB" => 0, "BCG" => 0,
		# Water
		"WAT" => 0, "HOH" => 0,
		# Ions
		"LI" => 1, "Li" => 1,
		"Na+" => 1, "NA+" => 1, "NA" => 1, "Na" => 1,
		"K+" => 1, "K" => 1,
		"RB" => 1, "Rb" => 1,
		"CS" => 1, "Cs" => 1,
		"MG" => 2, "Mg" => 2,
		"CA" => 2, "Ca" => 2,
		"SR" => 2, "Sr" => 2,
		"BA" => 2, "Ba" => 2,
		"ZN" => 2, "Zn" => 2,
		"F" => -1, "F-" => -1,
		"CL" => -1, "Cl" => -1, "Cl-" => -1, "CL-" => -1,
		"BR" => -1, "Br" => -1, "BR-" => -1, "Br-" => -1
	}

	# List of all amino acids
	PDB_RES_AA = [ 
		"ALA", "ARG", "ASN", "ASP", "CYS", "CYX", "GLN", "GLU", "GLY", "HIS",
		"HIE", "HID", "HIP", "ILE", "LEU", "LYS", "MET", "PHE", "PRO",
		"SER", "THR", "TRP", "TYR", "VAL", "ARG", "LYS", "HIP", "ASP",
		"GLU", "ASH", "GLH"
	]

	# Phosphorylated AA
	PDB_RES_PHOS_AA = [ "S1P", "SEP", "T1P", "TPO", "Y1P", "PTR", "H1D","H2D",
		"H1E", "H2E"
	]

	# neutral caps
	PDB_RES_CAPS = ["ACE", "NME"]
	PDB_CUBY_CAPS = ["BCC", "BNC", "BCB", "BCG"]



	def pdb_protein_charge(custom_res_charges = {})
		# Check if the geometry is PDB
		unless is_pdb?
			Cuby::error "PDB charge calculation requested on geometry that is not PDB"
		end

		# Initialize variables
		charge = 0
		unknown_residues = []

		# Traverse all residues
		residues_as_geometries.each{|resgeo|
			resname = resgeo.first.properties[:pdb_res_name]

			# Residue charge
			if custom_res_charges[resname]
				# First, use custom data
				charge += custom_res_charges[resname]
			elsif PDB_RES_CHARGES[resname]
				# Then try default database of residues
				charge += PDB_RES_CHARGES[resname]
			else
				unknown_residues << resname
			end

			# Protein terminal groups
			if PDB_RES_AA.include?(resname)
				atomnames = resgeo.map{|atom| atom.properties[:pdb_atom_name]}
				if atomnames.include?("OXT")
					# O terminus
					charge -= 1 
				end
				hcount = (atomnames & ["H1","H2","H3"]).size
				if (hcount == 3) || (resname == "PRO" && hcount == 2)
					# N terminus
					charge += 1 
				end
			end
		}

		# Check unknown residues
		if unknown_residues.size > 0
			unknown_residues.uniq!
			Cuby::warning "PDB charge calculation encountered unknown residues: #{unknown_residues.join(', ')}"
		end

		return charge
	end

end
