# Solvent-accessible surface area calculation
require 'tempfile'

module GeometrySasa

	def calculate_sasa(method, settings, radii, probe)
		case method
		when :freesasa_lr
			return calculate_sasa_freesasa_lr(settings, radii, probe)
		else
			raise "Unknown method for SASA calculation selected"
		end
	end

	def calculate_sasa_freesasa_lr(settings, radii, probe)
		# SASA calculation using external freesasa code
		# FreeSASA paper: http://dx.doi.org/10.12688/f1000research.7931.1
		# FreeSASA source: https://github.com/mittinatten/freesasa
		
		# Check the radii
		if radii.class == Hash
			elements = elements_in_system
			elements.each{|ele|
				unless radii[ele]
					raise "Radius for SASA calculation not defined for element #{ele}"
				end
			}
		elsif radii.class == Array
			unless radii.size == self.size
				raise "Array of SASA radii must be of the same size as geometry"
			end
		else
			raise("Radii for sasa have to be either hash element:value or array of values for individual atoms")
		end

		# Prepare temporary files
		tempfile = Cuby.make_tmpname('tmp')
		tempfile_pdb = tempfile + ".pdb"
		tempfile_radii = tempfile + ".radii"
		tempfile_out = tempfile + ".sasa"


		# Geometry and radii
		if radii.class == Hash
			# Write the geometry as a PDB
			self.write_pdb(:file => tempfile_pdb)

			# Write the radii
			f = File.open(tempfile_radii, "w+")
			f.puts "types:"
			elements.each{|ele|
				f.puts "#{ele.to_s} #{radii[ele]} normal"
			}
			f.puts
			f.puts "atoms:"
			elements.each{|ele|
				f.puts "ANY #{ele.to_s} #{ele.to_s}"
			}
			f.puts
			f.close
		else # Array
			# Write the geometry as a PDB
			tempgeo = self.deep_copy
			tempgeo.pdb_name_atoms_by_num
			tempgeo.write_pdb(:file => tempfile_pdb)

			f = File.open(tempfile_radii, "w+")
			f.puts "types:"
			radii.each_index{|i|
				f.puts "TYPE_#{tempgeo[i].properties[:pdb_atom_name]} #{radii[i]} normal"
			}
			f.puts
			f.puts "atoms:"
			radii.each_index{|i|
				f.puts "ANY #{tempgeo[i].properties[:pdb_atom_name]} TYPE_#{tempgeo[i].properties[:pdb_atom_name]} #{radii[i]}"
			}
			f.puts
			f.close
		end

		# Run freesasa
		# option -Y = include hYdrogens
		unless system("#{settings[:freesasa_exe]} -Y -c #{tempfile_radii} -p #{probe} #{tempfile_pdb} > #{tempfile_out}")
			Cuby::error "Running external program 'freesasa' failed"
		end

		# Parse the output
		sasa = nil
		IO.readlines(tempfile_out).each{|line|
			if line =~ /Total *:/
				sasa = line.strip.split[2].to_f
			end
		}

		# Delete temporary files
		FileUtils.rm_f([tempfile_pdb, tempfile_radii, tempfile_out])
		return sasa
	end
end
