# Series of geometries as read from the settings
# It expects 'settings' keyword to point to and .xyz file
# Template from 'geometry_template' is used when this keyword is set
# Labels are read from comment line of the .xyz file


class GeometrySeries < Array
	attr_reader :labels

	ALLOWED_FILETYPES = [:xyz, :mol2]

	def initialize
		@labels = []
	end

	def GeometrySeries.from_file(filename)
		geo_series = GeometrySeries.new

		# Get filetype
		filetype = GeoFile.type_from_filename(filename)

		# Check supported file types
		unless ALLOWED_FILETYPES.include?(filetype)
			raise("GeometrySeries cat read only files of types: #{ALLOWED_FILETYPES.join(', ')}")
		end

		# Read the frames
		f = File.open(filename)
		g = Geometry.new
		while GeoFile.read(g, filetype, :file => f, :read_series => true)
			geo_series << g
			g = Geometry.new
		end
		f.close
	
		return geo_series
	end

	def GeometrySeries.from_settings(settings, label_text="scan:")
		geo_series = GeometrySeries.new

		# Read template
		if settings.set?(:geometry_template)
			template = Geometry.from_file(settings[:geometry_template])
		end

		# Read geometries
		f = File.open(File.expand_path(settings[:geometry]))
		g = Geometry.new
		while g.read_xyz(:file => f)
			if settings.set?(:geometry_template)
				g2 = template.deep_copy
				g2.copy_coordinates_from(g)
				geo_series << g2
			else
				geo_series << g
			end
			# Get label from second line of the xyz file
			label = g.info[:xyz_remark]
			if label =~ /^#{label_text}/
				geo_series.labels << label.gsub(/^#{label_text}\s*/,"")
			end
			g = Geometry.new
		end
		f.close

		if geo_series.labels.size > 0 # at least some labels found
            if geo_series.labels.size != geo_series.size
				# When the number of labels does not correspond with a number of geometries,
				# remove them
				geo_series.labels.clear
			end
		end

		return geo_series
	end

	def size_check(message_head = "Geometry series")
		Cuby::error(message_head + ": no geometries read") if self.size == 0
		Cuby::warning(message_head + ": only one geometry read") if self.size == 1
	end

	def write_using_settings(settings, write_extra_info = false)
		return nil if settings[:geometry_write] == ""

		# Name and type from keywords
		filename = settings[:geometry_write]
		filetype = settings[:geometry_write_format]

		# Auto filetype
		firstgeo = self.first
		if filetype == :auto
			if settings[:geometry_write] == "/dev/stdout"
				# Standard output - use same format for input if possible
				if firtstgeo.info[:filetype] && GeoFile::TYPE_MODULE_W.has_key?(firtstgeo.info[:filetype])
					filetype = firtstgeo.info[:filetype]
				else
					# otherwise use .xyz
					filetype = :xyz
				end
			else
				# Regular file - determine from file name
				filetype = GeoFile.type_from_filename(filename)
				unless filetype
					Cuby::error("Failed to determine geometry file format from file name '#{filename}'")
				end
			end
		end

		# General options
		write_options = {
			:append => settings[:geometry_write_append],
			:extra_columns => settings[:pdb_extra_columns]
		}
		if settings.set?(:connectivity_tolerance)
			Connectivity.set_tolerance(settings[:connectivity_tolerance])
		end

		# Header
		if settings[:geometry_write_cuby_header]
			write_options[:cuby_header] = true
		end


		each{|geo|
			# File-format specific setup
			case filetype
			when :mol2
				# explicit connectivity has to be defined
				case settings[:geometry_write_connectivity]
				when :from_radii
					conn = Connectivity.new(geo, settings[:connectivity_tolerance])
					write_options[:bondlist] = conn.bond_list
				when :from_zmat
					if geo.info[:z_matrix]
						write_options[:bondlist] = geo.info[:z_matrix].bond_list
					else
						Cuby::error("Using z-matrix connectivity was requested but the geometry does not have\na z-matrix associated.")
					end
				end
			when :xyz
				if write_extra_info
					if geo.info[:xyz_remark]
						write_options[:second_line] = geo.info[:xyz_remark]
					end
				end
			end

			# Write, and set append mode for following frames
			geo.write_file(filename, filetype, write_options)
			write_options[:append] = true
		}

		# Return file name
		return filename
	end
end
