################################################################################
#
# Module GeoFile::MOL2
#
# Author: Jan Rezac
# Date created: 2009-08-17
# License: Cuby license
# Description: Tripos .mol2 file reader
# Status: Working
#
################################################################################

#: The .mol2 reader is limited ro reading the xyz coordinates from the file,
#: it ignores the connectivity and other information.

#: Format documentation: http://tripos.com/index.php?family=modules,SimplePage,,,&page=sup_mol2&s=0

module GeoFile
module MOL2
	def MOL2.read(geometry, arguments = {})
		#: Reads .mol2 file. Arguments:
		#* :file - could be either File (file handle) or string (file name)
		#* :read_into_geometry
		#* :read_series - set to 'true' when this is used to read a series of geometries from file

		file = arguments[:file]
		read_into = arguments[:read_into_geometry]

		close = false
		if file.class == String
			close = true
			file = File.new(file,"r")
		end

		unless file.kind_of?(IO) || file.kind_of?(StringIO)
			raise(TypeError,"File must be IO object (is #{file.class})")
		end

		if read_into
			readgeo = Geometry.new
		else
			readgeo = geometry
			geometry.clear # delete old atoms
		end

		header_natom = 0
		header_nbonds = 0
		header_nsubstructures = 0
		molname = ""
		# Seek the MOLECULE section
		while line = file.gets
			line.chomp!
			if line == "@<TRIPOS>MOLECULE"
				# Read name
				molname = file.gets.strip
				# Read number of atoms
				counts = file.gets.split
				header_natom = counts[0].to_i
				header_nbonds = counts[1].to_i
				header_nsubstructures = counts[2].to_i
				break
			end
		end

		# Reached end of file
		if line == nil
			if arguments[:read_series]
				# If a series is read, return false to terminate reading
				return false
			else
				# If a single geometry is read, mising header is an error
				Cuby::error "Invalid MOL2 file - section MOLECULE not found" if line == nil
			end
		end

		Cuby::error "Invalid MOL2 file - no atoms" if header_natom == 0

		# Seek ATOM section
		while line = file.gets
			line.chomp!
			if line == "@<TRIPOS>ATOM"
				break
			end
		end
		Cuby::error "Invalid MOL2 file - section ATOM not found" if line == nil

		# Read the geometry
		# First pass - save data for later
		mol2_data = []
		atomtype_dot = false
		atomname_ele_num = true
		header_natom.times{|i|
			if (line = file.gets) == nil
				Cuby::error('Invalid MOL2 file - it contains less atoms than specified in the header')
			end
			line.chomp!
			id, name, x, y, z, atomtype, sub_no, sub_name, charge, status = line.chomp.strip.split
			mol2_data[i] = [id, name, x, y, z, atomtype, sub_no, sub_name, charge, status]
			# Check for common atomtypes
			if atomtype =~ /\./
				atomtype_dot = true
			end
			# Check for atom names
			if not Atom.is_element_name?(name.sub(/[0-9]+.*/,''))
				atomname_ele_num = false
			end
		}

		# Check if we can guess element from atomtype or atomname
		if atomtype_dot
			# Sybyl atom types - element.type
			Cuby::log.puts_debug("Reading mol2 file - element information read from sybyl atom types")
		elsif atomname_ele_num
			# Atom name is element + number
			Cuby::log.puts_debug("Reading mol2 file - element information read from atom names")
		else
			Cuby::error("Reading mol2 file - unable to get element from atom name or type")
		end

		# Second pass - parse the data
		total_charge = 0.0
		header_natom.times{|i|
			id, name, x, y, z, atomtype, sub_no, sub_name, charge, status = mol2_data[i]
			if atomtype_dot
				element = atomtype.sub(/\..*/,'')
			else
				element = name.sub(/[0-9]+.*/,'')
			end
			atom = Atom.new(element, x.to_f, y.to_f, z.to_f, {:file_index => i})
			atom.properties[:pdb_atom_name] = name
			atom.properties[:pdb_res_name] = sub_name
			atom.properties[:pdb_res_no] = sub_no
			charge = charge.to_f
			atom.properties[:charge] = charge
			atom.properties[:mol2_atomtype] = atomtype
			readgeo << atom

			# Count charge
			total_charge += charge
		}
		geometry.info[:geofile_charge] = total_charge.round

		if read_into
			geometry.coords_from_geometry(readgeo)
		end

		geometry.info[:file_format] = :mol2
		geometry.info[:mol2_name] = molname

		# Seek BOND section
		while line = file.gets
			line.chomp!
			if line == "@<TRIPOS>BOND"
				break
			end
		end
		Cuby::error "Invalid MOL2 file - section BOND not found" if line == nil
		bonds = []
		# read bonds
		header_nbonds.times{|i|
			if (line = file.gets) == nil
				Cuby::error('Invalid MOL2 file - it contains less bonds than specified in the header')
			end
			i, a, b, order = line.strip.split
			a = a.to_i - 1
			b = b.to_i - 1
			bonds << [a, b, order]
		}
		geometry.info[:mol2_bonds] = bonds

		# Seek SUBSTRUCTURE section
		while line = file.gets
			line.chomp!
			if line == "@<TRIPOS>SUBSTRUCTURE"
				break
			end
		end
		unless line == nil
			substructures = []
			# read bonds
			header_nsubstructures.times{|i|
				if (line = file.gets) == nil
					Cuby::error('Invalid MOL2 file - it contains less substructures than specified in the header')
				end
				substructures << line.chomp
			}
			geometry.info[:mol2_substructures] = substructures
		end

		return true
	end

	def MOL2.write(geometry, arguments = {})
		#: Writes geometry in MOL2 format. Arguments:
		#* :file - file handle or name, default is $stdout
		#* :bondlist - list of bonds
		#* :do_not_rename_atoms - use original pdb atom names (must be present)
		#* :append
		#* :comment

		arguments[:file] = $stdout if arguments[:file].nil?
		long = false
		long = arguments[:long] unless arguments[:long].nil?

		file = arguments[:file]

		comment = ""
		comment = arguments[:comment] unless arguments[:comment].nil?

		# Deafults
		unless arguments[:name]
			if geometry.info[:mol2_name]
				arguments[:name] = geometry.info[:mol2_name]
			else
				arguments[:name] = "UNK"
			end
		end

		if geometry.info[:file_format] && geometry.info[:file_format] == :mol2
			unless arguments[:do_not_rename_atoms]
				arguments[:do_not_rename_atoms] = true
			end
		end

		# Make bonds list
		if arguments[:bondlist]
			bonds = arguments[:bondlist]
		else
			if geometry.info[:mol2_bonds]
				bonds = geometry.info[:mol2_bonds]
			else
				Cuby::warning("Writting .mol2 file - connectivity build by cuby")
				conn = Connectivity.new(geometry)
				bonds = conn.bond_list
			end
		end

		# Open the file
		close = false
		if file.class == String
			close = true
			mode="w+"
			mode = "a+" if arguments[:append] == true
			file = File.new(file,mode)
		end
		unless file.kind_of?(IO)
			raise(TypeError,"File must be IO object")
		end

		# Molecule record
		nsubstr = 1
		nsubstr = geometry.info[:mol2_substructures].size if geometry.info[:mol2_substructures]
		file.puts "@<TRIPOS>MOLECULE"
		file.puts arguments[:name]
		file.printf("%5d %5d %5d %5d %5d\n",geometry.size, bonds.size, nsubstr,0,0)
		file.puts "SMALL"
		file.puts "USER_CHARGES"
		file.puts
		file.puts comment

		# Name atoms
		#unless geometry[0].properties[:pdb_atom_name]
		#	geometry.pdb_name_atoms_by_num
		#end
		geometry.pdb_name_atoms_by_num unless arguments[:do_not_rename_atoms]

		# Atoms record
		file.puts "@<TRIPOS>ATOM"
		geometry.each_index{|i|
			atom = geometry[i]
			file.printf("%7d ",i+1) # Atom index
			file.printf("%-8s", atom.properties[:pdb_atom_name] ) # Atom name
			file.printf("%10.4f",atom.x)
			file.printf("%10.4f",atom.y)
			file.printf("%10.4f",atom.z)
			# Atom type
			if atom.properties[:mol2_atomtype]
				file.printf(" %-6s ", atom.properties[:mol2_atomtype])
			else
				file.printf(" %-6s ", atom.element.to_s)
			end
			if atom.properties[:pdb_res_no]
				file.printf("%4d ", atom.properties[:pdb_res_no])	# Substructure no.
			else
				file.printf("%4d ", 1)	# Substructure no.
			end
			if atom.properties[:pdb_res_name]
				file.printf("%-6s", atom.properties[:pdb_res_name]) # Substructure name
			else
				file.printf("%-6s", "UNK") # Substructure name
			end
			if atom.properties[:charge] == nil
				file.printf("%10.4f",0.0) # Charge
			else
				file.printf("%10.4f",atom.properties[:charge]) # Charge
			end
			file.printf("\n")
		}

		# Bonds
		file.puts "@<TRIPOS>BOND"
		bonds.each_index{|i|
			bond = bonds[i]
			file.printf("%6d",i+1)
			file.printf("%5d",bond[0]+1)
			file.printf("%5d",bond[1]+1)
			bo = "1"
			bo = bond[2] if bond[2]
			file.printf(" %-6s",bo)
			file.printf("\n")
		}

		# Substructure
		if geometry.info[:mol2_substructures]
			file.puts "@<TRIPOS>SUBSTRUCTURE"
			geometry.info[:mol2_substructures].each{|line|
				file.puts line
			}
		end



		file.close if close
		return nil
	end

end
end
