################################################################################
#
# Module GeoFile::ZMatrixMopac
#
# Author: Jan Rezac
# Date created: 2009-06-25
# License: Cuby license
# Description: ZMatrixMopac file reader/writer
# Status: Working
#
################################################################################

require "classes/internal_coordinates/z_matrix"

module GeoFile
module ZMatrixMopac
	MOPAC_MULTIPLICITIES = ["SINGLET", "DOUBLET", "TRIPLET", "QUARTET", "QUINTET", "SEXTET", "SEPTET", "OCTET", "NONET"]

	def ZMatrixMopac.read(geometry, arguments = {})
		#* :file - could be either File (file handle) or string (file name)
		#* :read_into_geometry
		#* :do_not_save_zmatrix - if set to true, parent zmatrix is not linked to the geometry

		# sets also variables
		#* geometry.info[:mopac_opt_flags] - mopac's otimization flags for each coordinate
		#* geometry.info[:mopac_symmetry] - mopac's dfinition of coordinate equivalence

		file = arguments[:file]
		read_into = arguments[:read_into_geometry]

		close = false
		if file.class == String
			close = true
			file = File.new(file,"r")
		end

		unless file.kind_of?(IO) || file.kind_of?(StringIO)
			raise(TypeError,"File must be IO object (is #{file.class})")
		end

		if read_into
			readgeo = Geometry.new
		else
			readgeo = geometry
			geometry.clear # delete old atoms
		end

		s = file.read
		file.close if close

		# Check if it is a z-matrix or cartesian coordinates
		line_count = 0
		r3_count = 0
		r2_count = 0
		r1_count = 0
		lines = []
		first_coord = nil
		coord_finished = false
		extra_blocks = []
		extra_block_index = nil
		s.split("\n").each{|line|
			# Skip comments
			next if line =~ /^ *\*/
			if line =~ / *[A-Z,a-z]\S* +-?[0-9]+\.[0-9]+ +[+-]?[0-9] +-?[0-9]+\.[0-9]+ +[+-]?[0-9] +-?[0-9]+\.[0-9]+ +[+-]?[0-9]/
				# Geometry line
				arr = line.strip.split
				if arr.size < 7
					Cuby::error("Cuby reads mopac input files only if the optimization flags are set properly")
				end
				ele, c1, o1, c2, o2, c3, o3, r1, r2, r3 = line.strip.split
				r3_count += 1 if r3 && r3.to_i > 0
				r2_count += 1 if r2 && r2.to_i > 0
				r1_count += 1 if r1 && r1.to_i > 0
				# Save fisrt coordinate
				first_coord = [c1.to_f, c2.to_f, c3.to_f] if line_count == 0
				line_count += 1
				lines << line
			else
				# Calculation specification
				# Get charge
				if match = /CHARGE=(-?[0-9]+)/i.match(line)
					geometry.info[:geofile_charge] = match[1].to_f
				end
				# Multiplicity
				MOPAC_MULTIPLICITIES.each_index{|i|
					if line =~ /#{MOPAC_MULTIPLICITIES[i]}/i
						geometry.info[:geofile_multiplicity] = i+1
					end
				}
			end

			# Blank line after coordinates
			if lines.size > 0 && line =~ /^\s*$/
				coord_finished = true
				extra_block_index = 0
				next
			end

			# Additional block after coordinates
			if coord_finished && line !~ /^\s*$/
				# Create a block if it does not exist
				unless extra_blocks[extra_block_index]
					extra_blocks[extra_block_index] = []
				end
				extra_blocks[extra_block_index] << line
			end

			# End of additional block - update block counter
			if coord_finished && line =~ /^\s*$/
				extra_block_index += 1
			end


		}

		# Identify z-matrix or cartesian
		if line_count == 1
			# One atom - always cartesian
			result = :cartesian
		elsif line_count == 2
			# Two atoms - if there is distance reference, it's a z-matrix
			if r1_count > 0
				result = :zmat
			else
				result = :cartesian
			end
		elsif r3_count == 0 && line_count > 3
			# Cartesian
			result = :cartesian
		elsif r3_count == 0 && line_count == 3 && r2_count == 1
			# Z-matrix of 3 atoms
			result = :zmat
		elsif r3_count + 3 == line_count
			# Z-matrix
			result = :zmat
		end

		# Save z-matrix or cartesian
		case result
		when :cartesian
			Cuby::log.puts_debug "Reading MOPAC file as certesian geometry"
			ZMatrixMopac.read_cartesian(lines, readgeo)
		when :zmat
			Cuby::log.puts_debug "Reading MOPAC file as z-matrix"
			zmat = ZMatrix.from_s_mopac(s)
			zmat.to_geometry.each{|atom| readgeo << atom}
			geometry.info[:z_matrix] = zmat unless arguments[:do_not_save_zmatrix]
		else
			# Unknown / mix
			Cuby::error("The mopac input file is neither z-matrix nor cartesian geometry")
		end

		# Symmetry info
		if geometry.info[:z_matrix] && extra_blocks.size > 0
			geometry.info[:mopac_symmetry] = extra_blocks[0]
		end

		if read_into
			geometry.coords_from_geometry(readgeo)
		end
	end

	def ZMatrixMopac.write(geometry, arguments = {})
		# Arguments:
		#   file
		#   force_cartesian
		#   opt_flags_reset - reset optimization flags (== optimize all) even if different flags are present
		#   write_symmetry (default true)

		# Default values	
		arguments[:write_symmetry] = true if arguments[:write_symmetry].nil?


		# Check if geometry has z-matrix object
		if !geometry.info[:z_matrix] || arguments[:force_cartesian]
			return ZMatrixMopac.write_cart(geometry, arguments)
		end

		# Update the z-matrix from coordinates
		geometry.info[:z_matrix].update_zmat!(geometry)

		arguments[:file] = $stdout if arguments[:file].nil?
		file = arguments[:file]

		close = false
		if file.class == String
			close = true
			mode="w+"
			file = File.new(file,mode)
		end
		unless file.kind_of?(IO)
			raise(TypeError,"File must be IO object")
		end

		# Write header
		if arguments[:cuby_header]
			file.puts ZMatrixMopac.header_line(geometry)
		end
		
		# Write z-matrix
		file.puts geometry.info[:z_matrix].to_s_mopac(:dummy_xx => true, :opt_flags_reset => arguments[:opt_flags_reset])
		file.puts

		# Write mopac symmetry data
		if [:write_symmetry] && geometry.info[:mopac_symmetry]
			geometry.info[:mopac_symmetry].each{|line| file.puts line}
			file.puts
		end

		file.close if close
		return nil
	end

	#-----------------------------------------------------------------------
	# Private
	#-----------------------------------------------------------------------

	def ZMatrixMopac.read_cartesian(lines, geo)
		mopac_opt_flags = []
		lines.each{|line|
			a = line.strip.split
			a[0].gsub!(/\(.*\)/,'') # remove label in parenthesis
			atom = Atom.new(a[0],a[1].to_f,a[3].to_f,a[5].to_f)
			geo << atom
			mopac_opt_flags << [a[2], a[4], a[6]]
		}

		geo.info[:mopac_opt_flags] = mopac_opt_flags
	end

	def ZMatrixMopac.write_cart(geometry, arguments = {})
		arguments[:file] = $stdout if arguments[:file].nil?
		file = arguments[:file]

		close = false
		if file.class == String
			close = true
			mode="w+"
			file = File.new(file,mode)
		end
		unless file.kind_of?(IO)
			raise(TypeError,"File must be IO object")
		end

		# Mopac optimization flags
		if geometry.info[:mopac_opt_flags] && ! arguments[:opt_flags_reset]
			opt_flags = geometry.info[:mopac_opt_flags]
		else
			opt_flags = []
			geometry.each_index{|i|
				opt_flags[i] = ["+1", "+1", "+1"]
			}
		end
		
		# Write header
		if arguments[:cuby_header]
			file.puts ZMatrixMopac.header_line(geometry)
		end
		
		# Write coordinates
		geometry.each_with_index{|atom, i|
			ele = atom.element.to_s
			ele = "XX" if ele == "X"
			#  C     0.33159086 +1   0.00244065 +1  -1.62184664 +1
			file.printf("%3s   %12.8f %2s %12.8f %2s %12.8f %2s\n",
				    ele,
				    atom.x,
				    opt_flags[i][0],
				    atom.y,
				    opt_flags[i][1],
				    atom.z,
				    opt_flags[i][2]
			)
		}

		file.close if close
		return nil
	end

	def ZMatrixMopac.header_line(geometry)
		header = []
		if geometry.info[:geofile_charge]
			header << "CHARGE=#{geometry.info[:geofile_charge].to_i}"
		end
		if geometry.info[:geofile_multiplicity]
			header << MOPAC_MULTIPLICITIES[geometry.info[:geofile_multiplicity] - 1]
		end
		return header.join(" ")
	end
end
end
