################################################################################
#
# CHARMM_DFTB interface
#
# Author: Jan Rezac
# Date created: 2016-05-19
# License: Cuby4 license
# Description: Interface for DFTB calculations in CHARMM
# Status: Works
#
################################################################################

#===============================================================================
#===============================================================================

require "erb" # ERB templating system, used for construction of the input
require "classes/tools/output_parser.rb"
require "interfaces/dftb/classes/dftb_slko_files.rb"
require "interfaces/dftb/classes/hubbard_derivs.rb"

module InterfaceCharmmDftb
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "OK, more features to be implemented"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy]
	MODIFIER = false
	DIRECTORY = "CHDFTB"
	# Methods provided by the interface:
	METHODS = {
		:"scc-dftb" 		=> [:energy],
		:"scc-dftb3" 		=> [:energy]
	}
	SOLVENTS = [:pb]
	#=======================================================================
	
	CHARMM_INP_FN = "charmm.inp"
	CHARMM_OUT_FN = "charmm.out"

	def prepare_interface
		# Initialize the slko files
		@slko_files = DftbSlkoFiles.new(@settings)
		# Check if there are all parameters needed
		@slko_files.check_geometry(@geometry)

		@use_pb_solvent = @settings[:solvent_model] == :pb

		# Load Hubbard derivatives
		if @settings[:method] == :"scc-dftb3"
			datafile = interface_dir + "/../dftb/data/hubbard_derivs.yaml"
			@hubbard_derivs = HubbardDerivs.from_settings(@settings, datafile)
		end

		# Prepare calculation directory or reuse old
		if calc_dir_mkdir(CHARMM_INP_FN, CHARMM_OUT_FN) == :old
			calc_using_input
			return 
		end
		# Write input
		calc_writing_input
		charmm_dftb_write_input
		# Write geometry
		charmm_dftb_write_geometry

	end

	def calculate_interface
		charmm_dftb_write_geometry
		charmm_dftb_run_calculation unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(in_calc_dir(CHARMM_OUT_FN))
		results = charmm_dftb_read_results
		return results
	end

	def cleanup_interface
		calc_dir_delete if @settings[:job_cleanup]
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def charmm_dftb_write_input
		# List of elements in system, ordered
		elements = @geometry.elements_in_system_sorted
		# Atom numbering
		atom_numbers = @geometry.atom_numbering_by_element

		# Input file from template
		if @use_pb_solvent
			template_inp = ERB.new(IO.read(interface_dir + "/templates/solvation.inp.erb"))
		else
			template_inp = ERB.new(IO.read(interface_dir + "/templates/charmm.inp.erb"))
		end
		File.open(in_calc_dir(CHARMM_INP_FN),"w+") {|file|
			file.print(template_inp.result(binding))
		}

		# Residue definition from template
		template_mol = ERB.new(IO.read(interface_dir + "/templates/molecule.rtf.erb"))
		File.open(in_calc_dir("molecule.rtf"),"w+") {|file|
			file.print(template_mol.result(binding))
		}

		# DFTB parameters file
		f = File.open(in_calc_dir("sccdftb.dat"),"w+")
		# SlKo files
		elements.each{|e1|
			elements.each{|e2|
				f.puts "'#{@slko_files.filename(e1, e2)}'"
			}
		}
		# Hubbard derivs
		if @settings[:method] == :"scc-dftb3"
			elements.each{|e|
				f.puts "'#{e.to_s.downcase}' #{@hubbard_derivs[e]}"
			}
		end
		# X-H gamma
		if @settings[:dftb_xh_damping]
			f.puts @settings[:dftb_xh_damping_exp]
		end
		f.close

		# Solvation parameter file
		if @use_pb_solvent
			f = File.open(in_calc_dir("chrg_radius_scc.dat"),"w+")
			f.puts "#{elements.size} 1"
			elements.each{|e|
				f.puts "#{@settings[:charmm_dftb_pb_b][e]} #{@settings[:charmm_dftb_pb_a][e]}"
			}
			f.close
		end
	end

	def charmm_dftb_write_geometry
		# Atom numbering
		atom_numbers = @geometry.atom_numbering_by_element

		# Write file
		f = File.open(in_calc_dir("molecule.crd"),"w+")
		f.puts "*  Geometry for CHARMM"
		f.puts "*  generated by Cuby"
		f.puts "*"
		f.printf("%5d\n", @geometry.size)
		@geometry.each_index{|i|
			atom = @geometry[i]
			atom_name = atom.element.to_s + atom_numbers[i].to_s
			f.printf("%5d%5d DFTB %-4s%10.5f%10.5f%10.5f DFTB 1      0.0000\n", i+1, 1, atom_name, atom.x, atom.y, atom.z)
		}
		f.close
	end

	def charmm_dftb_run_calculation
		# Run calculation
		if @settings[:parallel] > 1
			# Parallel
			if ! system("cd #{calc_dir} && #{@settings[:charmm_exe_mpi]} #{@settings[:parallel]} #{CHARMM_INP_FN} 2> charmm.err > charmm.out")
				Cuby::error ("CHARMM returned nonzero exit code (#{@name})")
			end
		else
			# Serial
			unless system("cd #{calc_dir}; #{@settings[:charmm_exe]} < #{CHARMM_INP_FN} 2> charmm.err > #{CHARMM_OUT_FN}")
				Cuby::error "CHARMM returned nonzero exit code (#{@name})"
			end
		end
	end

	def charmm_dftb_read_results
		results = Results.new
		parser = OutputParser.new(in_calc_dir(CHARMM_OUT_FN), @name)

		parser.add_pattern(:energy, /^ENER> +0 *([-]*[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :last)
		if @use_pb_solvent
			parser.add_pattern(:energy_vac, /^ENER> +0 *([-]*[0-9]+\.[0-9]+)/, :get => :submatch, :multi => :first)
		end

		parser.execute

		results.energy = parser[:energy]
		if @use_pb_solvent
			results.energy_components[:energy_gas_phase] = parser[:energy_vac]
			results.energy_components[:solvation_free_energy] = results.energy - parser[:energy_vac]
		end

		return results
	end

end
