################################################################################
#
# GAMESS interface
#
# Author: Jan Rezac
# Date created: 2016-02-15
# License: Cuby4 license
# Description: Interface for external calculations in GAMESS
# Status: In development
#
################################################################################

#===============================================================================
# Interface to the GAMESS package
# Tested with GAMESS version 18 AUG 2016 (R1)
# http://www.msg.ameslab.gov/GAMESS/
#===============================================================================

require "classes/tools/output_parser.rb"
# Borrowed from DFTB
require "interfaces/dftb/classes/dftb_slko_files.rb"
require "interfaces/dftb/classes/hubbard_derivs.rb"

module InterfaceGamess
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "Only simple implementation for now"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy]
	MODIFIER = false
	DIRECTORY = "GAMESS"
	# Methods provided by the interface:
	METHODS = {
		:am1			=> [:energy],
		:pm3			=> [:energy],
		:rm1			=> [:energy],
		:mndo			=> [:energy],
		:"scc-dftb"		=> [:energy],
		:"scc-dftb3"		=> [:energy]
	}
	SOLVENTS = [:cpcm, :smd]
	#ATOMIC_CHARGES = [:mulliken]
	#=======================================================================
	
	INPUT_FILE = "gamess.inp"
	OUTPUT_FILE = "gamess.out"

	def prepare_interface
		# General setup
		if @settings[:gamess_fmo]
			require "interfaces/gamess/classes/fmo.rb"
			@fmo = FMO.new(@geometry, @settings)
		end

		# Things that have to be done even when old input is reused:

		# Prepare calculation directory
		if calc_dir_mkdir("gamess.header", OUTPUT_FILE) == :old
			calc_using_input # Old input should be reused
			return
		end

		# Write the keyword part of the input
		gamess_write_keywords # This writes file gamess.header

		# Create complete input by combining header, footer and the current geometry
		gamess_write_input_geo

		# Save info on the system upon writing the input
		calc_writing_input
	end

	def calculate_interface
		gamess_run unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(calc_dir+"/"+OUTPUT_FILE)
		# Read results
		if @fmo
			return @fmo.read_results(in_calc_dir(OUTPUT_FILE))
		else
			return gamess_read
		end
	end

	def cleanup_interface
		calc_dir_delete if @settings[:job_cleanup]
	end

	#=======================================================================
	# Private methods
	#=======================================================================

	def gamess_write_keywords
		#: Write the header of the input where the calculation is set up
		#: geometry is added later, so that the header can be modified
		#: by the user after the job is prepared

		# HEADER FILE
		filename = in_calc_dir("gamess.header")
		f = File.open(filename,"w+")

		# Store what user input was used, unused will be added at the end
		used_extra_input = []

		# CONTRL - Run type
		f.puts " $CONTRL"
		f.puts "  SCFTYP=RHF"
		f.puts "  RUNTYP=ENERGY"
		f.puts "  NPRINT=-5"
		f.puts "  ICHARG=#{@settings[:charge]}"
		f.puts "  MULT=#{@settings[:multiplicity]}"
		if @settings[:gamess_extra_input]["CONTRL"]
			@settings[:gamess_extra_input]["CONTRL"].each_line{|l| f.puts "  #{l}"}
			used_extra_input << "CONTRL"
		end
		if [:"scc-dftb", :"scc-dftb3"].include? @settings[:method]
			f.puts "  MAXIT=#{@settings[:dftb_scc_cycles]}"
		else
			f.puts "  MAXIT=#{@settings[:scf_cycles]}"
		end
	       	f.puts " $END"

		# SYSTEM - Memory
		mw_per_core = @settings[:mem] / 8 / @settings[:parallel]
		f.puts " $SYSTEM"
		f.puts "  MWORDS=#{mw_per_core}"
		if @settings[:gamess_extra_input]["SYSTEM"]
			@settings[:gamess_extra_input]["SYSTEM"].each_line{|l| f.puts "  #{l}"}
			used_extra_input << "SYSTEM"
		end
		f.puts " $END"

		# SCF - SCF setup
		f.puts " $SCF"
		f.puts "  DIRSCF=.T."
		if @settings[:gamess_extra_input]["SCF"]
			@settings[:gamess_extra_input]["SCF"].each_line{|l| f.puts "  #{l}"}
			used_extra_input << "SCF"
		end
		if [:"scc-dftb", :"scc-dftb3"].include? @settings[:method]
			f.puts "  CONV=#{@settings[:dftb_scc_convergence]}"
		else
			f.puts "  CONV=1.0e-#{@settings[:scf_convergence]}"
		end
		f.puts " $END"

		# Solvent
		if @settings[:solvent_model] != :none
			f.puts " $PCM"
			f.puts "  SOLVNT=#{@settings[:solvent_name].upcase}"
			case @settings[:solvent_model]
			when :pcm
				#f.puts "  IEF=0" # Does not work with DFTB
				f.puts "  ICAV=1 IDISP=1"
				f.puts "  IFMO=-1" if @fmo
			when :"ief-pcm"
				f.puts "  IEF=1" # There are more variants of IEF-PCM, not tested
				f.puts "  IFMO=-1" if @fmo
			when :cpcm
				f.puts "  IEF=-10"
				f.puts "  ICAV=1"
				f.puts "  IDISP=1"
				f.puts "  IFMO=-1" if @fmo
			when :smd
				f.puts "  IEF=-10"
				f.puts "  SMD=.TRUE."
				f.puts "  IFMO=-1" if @fmo
			end
			if @settings[:gamess_extra_input]["PCM"]
				@settings[:gamess_extra_input]["PCM"].each_line{|l| f.puts "  #{l}"}
				used_extra_input << "PCM"
			end
			f.puts " $END"

			# Custom PCM radii
			if @settings.set?(:gamess_pcm_radii) && @settings.set?(:gamess_pcm_radii_add)
				Cuby::error("Keywords gamess_pcm_radii and gamess_pcm_radii_add can not be used at the same time")
			end
			if @settings.set?(:gamess_pcm_radii)
				f.puts " $PCMCAV"
				f.puts "  ALPHA=1.0"
				# Assign radii on the basis of selections
				mol_radii = @settings.atom_parameters_array(:gamess_pcm_radii, @geometry)
				# Write out radii one by one - does not work on one line, it gets truncated
				mol_radii.each_index{|i|
					f.puts "  RIN(#{i+1})= #{mol_radii[i]}"
				}

				if @settings[:gamess_extra_input]["PCMCAV"]
					@settings[:gamess_extra_input]["PCMCAV"].each_line{|l| f.puts "  #{l}"}
					used_extra_input << "PCMCAV"
				end
				f.puts " $END"
			end
			if @settings.set?(:gamess_pcm_radii_add)
				f.puts " $PCMCAV"
				# Assign radii on the basis of selections
				mol_radii = @settings.atom_parameters_array(:gamess_pcm_radii_add, @geometry, false)
				# Write out radii one by one - does not work on one line, it gets truncated
				mol_radii.each_index{|i|
					f.puts "  RIN(#{i+1})= #{mol_radii[i]}" if mol_radii[i]
				}

				if @settings[:gamess_extra_input]["PCMCAV"]
					@settings[:gamess_extra_input]["PCMCAV"].each_line{|l| f.puts "  #{l}"}
					used_extra_input << "PCMCAV"
				end
				f.puts " $END"
			end
		end

		# Method/basis
		f.puts " $BASIS"
		case @settings[:method]
		when :am1, :pm3, :rm1, :mndo
			f.puts "  GBASIS=#{@settings[:method].to_s.upcase}"
		when :"scc-dftb", :"scc-dftb3"
			f.puts "  GBASIS=DFTB"
		end
		if @settings[:gamess_extra_input]["BASIS"]
			@settings[:gamess_extra_input]["BASIS"].each_line{|l| f.puts "  #{l}"}
			used_extra_input << "BASIS"
		end
		f.puts " $END"

		# FMO
		if @fmo
			@fmo.write_input(f)
		end

		# DFTB
		if [:"scc-dftb", :"scc-dftb3"].include? @settings[:method]
			f.puts " $DFTB"
			f.puts "   SCC=.TRUE."
			if @settings[:method] == :"scc-dftb3"
				f.puts "   DFTB3=.TRUE."
				# set Hubbard derivatives
				datafile = interface_dir + "/../dftb/data/hubbard_derivs.yaml"
				@hubbard_derivs = HubbardDerivs.from_settings(@settings, datafile)
				hd = []
				@geometry.elements_in_system.each{|ele|
					hd << @hubbard_derivs[ele]
				}
				f.puts "   HUBDER(1)=" + hd.join(",")
			else
				f.puts "   DFTB3=.FALSE."
			end
			if @settings[:dftb_xh_damping]
				f.puts "   DAMPXH=.TRUE. DAMPEX=#{@settings[:dftb_xh_damping_exp]}"
			else
				f.puts "   DAMPXH=.FALSE."
			end
			f.puts "   ETEMP=#{@settings[:dftb_e_temp]}" if @settings.set?(:dftb_e_temp)
			if @settings[:gamess_extra_input]["DFTB"]
				@settings[:gamess_extra_input]["DFTB"].each_line{|l| f.puts "  #{l}"}
				used_extra_input << "DFTB"
			end
			f.puts " $END"
		end

		# DFTB Slko files
		f.puts " $DFTBSK"
		slko_files = DftbSlkoFiles.new(@settings)
		slko_files.check_geometry(geometry)
		elements = geometry.elements_in_system.sort{|a,b| PeriodicTable.proton_number(a) <=> PeriodicTable.proton_number(b)}
		elements.each_index{|i|
			filename = slko_files.filename(elements[i], elements[i])
			f.puts "  #{elements[i]} #{elements[i]} \"#{filename}\""
			i.times{|j|
				filename = slko_files.filename(elements[i], elements[j])
				f.puts "  #{elements[i]} #{elements[j]} \"#{filename}\""
				filename = slko_files.filename(elements[j], elements[i])
				f.puts "  #{elements[j]} #{elements[i]} \"#{filename}\""
			}
		}
		if @settings[:gamess_extra_input]["DFTBSK"]
			@settings[:gamess_extra_input]["DFTBSK"].each_line{|l| f.puts "  #{l}"}
			used_extra_input << "DFTBSK"
		end
		f.puts " $END"

		# Additional user-defined blocks
		@settings[:gamess_extra_input].each_pair{|name, contents|
			next if used_extra_input.include?(name)
			f.puts " $#{name}"
			contents.each_line{|l| f.puts "  #{l}"}
			f.puts " $END"
		}

		# Setup
		f.close
	end

	def gamess_write_input_geo
		#: Write final input for mopac, using the previously prepared header and footer
		#: and a current geometry

		# Read header and footer
		header = IO.readlines(in_calc_dir("gamess.header"))

		# Complete input:
		filename = in_calc_dir(INPUT_FILE)
		f = File.open(filename,"w+")

		# Print header from the header file
		f.puts header

		# Write geometry
		if @fmo
			@fmo.write_geo(f)
		else
			f.puts " $DATA"
			f.puts "  cuby"
			f.puts "  C1"
			@geometry.each_index{|i|
				atom = @geometry[i]
				f.printf(" %3s %5.1f %15.10f %15.10f %15.10f\n",atom.element, atom.proton_number.to_f, atom.x, atom.y, atom.z)
			}
			f.puts " $END"
		end
		f.close
	end

	def gamess_run
		#: Set up environment for gamess and run it

		# Build input with the current geometry
		gamess_write_input_geo

		# Test if the executable is set
		Cuby::error "GAMESS executable is not set (see keyword gamess_rungms)" if @settings[:gamess_rungms] == ""

		# Run gamess
		command = "cd #{calc_dir};"
		command << " mkdir tmp;"
		command << " export GAMESS_CUBY_TMP=$PWD/tmp;"
		command << " rm -f gamess.dat; "
		command << " #{@settings[:gamess_rungms]} #{INPUT_FILE} #{@settings[:gamess_exe_version]} #{@settings[:parallel]} 2> gamess.err > #{OUTPUT_FILE};"
		command << " rm -r tmp;"
		unless system(command)
			Cuby::error "GAMESS returned nonzero exit code (calculation #{@name})"
		end

		# Error check

	end

	def gamess_read # => Results
		# Initialize results
		results = Results.new
		parser = OutputParser.new(in_calc_dir(OUTPUT_FILE), @name)

		# Errors
		parser.add_error(/SCF IS UNCONVERGED/, "SCF not converged")

		# Method-dependent patterns
		case @settings[:method]
		when :am1, :pm3, :rm1, :mndo
			# Electronic energy - ignored
			#parser.add_pattern(:sqm_energy, /^ *FINAL R-AM1 ENERGY IS (.*) AFTER/, {:get => :submatch, :multi => :last})
			# Heat of formation
			parser.add_pattern(:sqm_energy_kcal, /^ *HEAT OF FORMATION IS (.*) KCAL/, {:get => :submatch, :multi => :last})
		when :"scc-dftb", :"scc-dftb3"
			# DFTB
			parser.add_pattern(:dftb_energy, /^ *FINAL RDFTB ENERGY IS (.*) AFTER/, {:get => :match_array, :multi => :last})
			parser.add_pattern(:dftb_energy_first, /^ *FINAL RDFTB ENERGY IS (.*) AFTER/, {:get => :match_array, :multi => :first})
			if @settings[:solvent_model] != :none
				parser.add_pattern(:g_in_solvent, /^ *TOTAL FREE ENERGY IN SOLVENT *= (.*) A.U./, {:get => :match_array})
			end
		end

		# Run the parser
		parser.execute 

		# Results for various methods
		case @settings[:method]
		when :am1, :pm3, :rm1, :mndo
			if @settings[:solvent_model] == :none
				results.energy = parser[:sqm_energy_kcal]
			else
				Cuby::error "Reading of the results of SQM calculation in solvent not implemented yet"
			end
		when :"scc-dftb", :"scc-dftb3"
			# DFTB in solvent
			if @settings[:solvent_model] == :none
				# DFTB
				results.energy = parser[:dftb_energy][0] * HARTREE2KCAL
			else
				results.energy_components[:e_dftb_solvent] = parser[:dftb_energy][0] * HARTREE2KCAL
				results.energy_components[:e_dftb_gas_phase] = parser[:dftb_energy_first][0] * HARTREE2KCAL
				results.energy_components[:g_dftb_solvent] = parser[:g_in_solvent][0] * HARTREE2KCAL
				results.energy = results.energy_components[:g_dftb_solvent]
				results.energy_components[:g_solvation] = results.energy - results.energy_components[:e_dftb_gas_phase]

			end
		end

		return results
	end



end
