################################################################################
#
# MOLPRO interface
#
# Author: Jan Rezac
# Date created: 2014-01-20
# License: Cuby4 license
# Description: Interface for external calculations using MOLPRO
# Status: Works
#
################################################################################

#===============================================================================
# Interface to MOLPRO
# http://www.molpro.net/
#===============================================================================

# TBD (compared to cuby3)
# * mp3
# * dft
# * mp2c
# * mp2-f12, ccsd-f12, ccsd(t)-f12
# * lmp2
# * basis set from file
# * basisset_elements
# * auxiliary basis choice
# * point charges
# * open shell calculations

require "erb" # ERB templating system, used for construction of the input
require "classes/tools/output_parser.rb"

module InterfaceMolpro
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :warning
	DEVELOPMENT_STATUS = "Interface works but was not tested yet"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient, :ghost_atoms, :point_charges]
	MODIFIER = false
	DIRECTORY = "MOLPRO"
	# Methods provided by the interface:
	METHODS = {
		:"hf"		=> [:energy, :gradient, :point_charges],
		:"mp2"		=> [:energy, :gradient, :point_charges],
		:"ccsd"		=> [:energy],
		:"ccsd(t)"	=> [:energy],
		:"mp3"		=> [:energy],
		:"dft-sapt"	=> [:energy],
		#:"dft"		=> [:energy, :gradient],
	}
	SOLVENTS = [:cosmo]
	# Input structure
	INPUT_BLOCKS = [
		InputBlock[:molecule_a, :optional, "Definition of the first monomer for SAPT calculation (selection, charge, multiplicity)"],
		InputBlock[:molecule_b, :optional, "Definition of the second monomer for SAPT calculation (selection, charge, multiplicity)"]
	]
	#=======================================================================
	
	MOLPRO_INPUT =  "molpro.com"
	MOLPRO_OUTPUT = "molpro.out"
	MOLPRO_ERROUT = "molpro.err"

	def prepare_interface
		# Check setup
		molpro_check_setup

		# Prepare calculation directory
		if calc_dir_mkdir(MOLPRO_INPUT, MOLPRO_OUTPUT) == :old
			calc_using_input # Old input should be reused
			return
		end

		# Write input
		molpro_write_input

		# Write geometry
		molpro_write_geometry

		# Write point charges
		molpro_write_charges if @point_charges

		# Save info on the system upon writing the input
		calc_writing_input
	end

	def calculate_interface
		# Write current geometry
		molpro_write_geometry

		# Write point charges
		molpro_write_charges if @point_charges

		# Run molpro
		molpro_run unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(in_calc_dir(MOLPRO_OUTPUT))

		# Read results
		if @settings[:method] == :"dft-sapt"
			return molpro_read_sapt
		else
			return molpro_read
		end
	end

	def cleanup_interface
		calc_dir_delete if @settings[:job_cleanup]
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def molpro_write_geometry
		@geometry.write_xyz(:file => in_calc_dir("geo.xyz"))

		# Save also monomers - not used now
		if false
			prepare_dimer_settings
			mono_a = @geometry.geometry_from_selection(@settings[:molecule_a, :selection])
			mono_b = @geometry.geometry_from_selection(@settings[:molecule_b, :selection])

			mono_a.write_xyz(:file => in_calc_dir("geo_a.xyz"))
			mono_b.write_xyz(:file => in_calc_dir("geo_b.xyz"))
		end

	end

	def molpro_check_setup
		if @settings[:density_fitting] == :scf && @settings[:method] != :hf
			Cuby::error "Correlation energy calculation without density fitting can not be performed\nafter a SCF calculation with density fitting"
		end

		if @settings[:method] == :mp2 && [:both, :correlation].include?(@settings[:density_fitting]) && @what.include?(:gradient)
			Cuby::error("Gradient not available for MP2 with density fitting in Molpro")
		end

		if @settings[:method] == :mp3 && @settings[:density_fitting] != :none
			Cuby::error("MP3 with density fitting not available in Molpro")
		end

		if (@settings[:spin_restricted] == :uhf || (@settings[:spin_restricted] == :auto_uhf && @settings[:multiplicity] != 1)) && @settings[:density_fitting] != :none
			Cuby::error("Spin-unrestricted calculations not available with desnity fitting")
		end

		if (@settings[:spin_restricted] == :uhf || (@settings[:spin_restricted] == :auto_uhf && @settings[:multiplicity] != 1)) && ! [:hf, :mp2].include?(@settings[:method])
			Cuby::error("Spin-unrestricted calculations available only for HF and MP2")
		end

		if @settings[:method] == :mp3 && @settings[:multiplicity] != 1
			Cuby::error("Open-shell MP3 calculations not available")
		end
	end
	
	def molpro_write_input
		# Generate the input file from the templates
		file = File.open(in_calc_dir(MOLPRO_INPUT),"w+")

		# Open shell
		if @settings[:spin_restricted] == :rhf || (@settings[:spin_restricted] == :auto_uhf && @settings[:multiplicity] == 1)
			if @settings[:multiplicity] != 1
				openshell = "r"
			else
				openshell = ""
			end
		else
			openshell = "u"
		end

		# Memory
		memory_mw = @settings[:mem].to_i / 8
		memory_mw = memory / @settings[:parallel] if @settings[:parallel_mode] == :mpi

		# Build basis lists to be used in the templates
		elements = @geometry.elements_in_system
		basis_list = "DEFAULT=" + @settings[:basisset]
		@settings.elements_hash(:basisset_elements).each_pair{|element, basis|
			basis_list += "\n" + element.to_s + "=" + basis if elements.include?(element)
		}
		if @settings.set?(:auxiliary_basis_scf)
			auxbasis_scf_list = "DEFAULT=" + @settings[:auxiliary_basis_scf] + "/jkfit"
			@settings.elements_hash(:auxiliary_basis_scf_elements).each_pair{|element, basis|
				if elements.include?(element)
					auxbasis_scf_list += "\n" + element.to_s + "=" + basis + "/jkfit" 
				end
			}
		end
		if @settings.set?(:auxiliary_basis_mp2)
			auxbasis_mp2_list = "DEFAULT=" + @settings[:auxiliary_basis_mp2] + "/mp2fit"
			@settings.elements_hash(:auxiliary_basis_mp2_elements).each_pair{|element, basis|
				auxbasis_mp2_list += "\n" + element.to_s + "=" + basis + "/mp2fit" if elements.include?(element)
			}
		end

		# Setup for all non-SAPT calculations
		unless @settings[:method] == :"dft-sapt"
			# General setup
			template = ERB.new(IO.read(interface_dir + "/templates/general.erb"))
			file.print(template.result(binding))

			# HF setup
			template = ERB.new(IO.read(interface_dir + "/templates/hf.erb"))
			file.print(template.result(binding))
		end

		# Frozen core setup
		frozen_core = "" # Default - use molpro's derfault
		if @settings[:correlation_frozen_core]
			# Use efault unless custom core is specified
			if @settings.set?(:frozen_core_elements)
				if @settings.set?(:reduce_core_element)
					Cuby::error "Keyword reduce_core_element can not combined with frozen_core_elements"
				end
				ele_core = @settings.elements_hash(:frozen_core_elements)
				core = 0
				@geometry.each{|atom|
					next if atom.properties[:ghost]
					if ele_core[atom.element]
						core += ele_core[atom.element]
					else
						Cuby::error "No. of core orbitals not defined for element #{atom.element}"
					end
				}
				frozen_core = ";CORE,#{core}"
			end
		else
			# Disable frozen core
			frozen_core = ";CORE,0"
		end

		# Method-specific templates
		case @settings[:method]
		when :mp2
			# MP2 setup
			template = ERB.new(IO.read(interface_dir + "/templates/mp2.erb"))
			file.print(template.result(binding))
		when :mp3
			# MP3 setup
			template = ERB.new(IO.read(interface_dir + "/templates/mp3.erb"))
			file.print(template.result(binding))
		when :"ccsd"
			# CCSD setup
			template = ERB.new(IO.read(interface_dir + "/templates/ccsd.erb"))
			file.print(template.result(binding))
		when :"ccsd(t)"
			# CCSD(T) setup
			template = ERB.new(IO.read(interface_dir + "/templates/ccsdt.erb"))
			file.print(template.result(binding))
		when :"dft-sapt"
			# DFT-SAPT
			# Prepare sapt setup
			sapt = sapt_setup
			# Sapt options
			# Density fitting
			if @settings[:density_fitting] == :none
				df = ""
			elsif @settings[:density_fitting] == :both
				df = "-df"
			else
				Cuby::warning("Density fitting used globally in DFT-SAPT.")	
				df = "-df"
			end
			# Load template
			template = ERB.new(IO.read(interface_dir + "/templates/dft-sapt#{df}.erb"))

			file.print(template.result(binding))
		end

		file.close
	end

	def sapt_setup
		sapt = {}
		# Default setup
		sapt[:icpks] = 0;
		sapt[:fitlevel_hf] = 2;

		unless [@settings[:dft_sapt_level], 0].include? @settings[:molpro_sapt_df_level]
			Cuby.error "DFT-SAPT setup error:\nmolpro_sapt_df_level must be equal to dft_sapt_level or set to zero"
		end
		sapt[:fitlevel_dft] = @settings[:molpro_sapt_df_level];

		# Molecules setup
		prepare_dimer_settings

		# Dummy atoms
		sapt[:dummies_a] = @geometry.atomlist_from_selection(@settings[:molecule_b, :selection]).map{|x| x+1}.join(",")
		sapt[:dummies_b] = @geometry.atomlist_from_selection(@settings[:molecule_a, :selection]).map{|x| x+1}.join(",")

		# Frozen core setup
		frozena = 0
		frozenb = 0
		core_shells = [nil,0,1,5,9,18,27,43]
		reduce = @settings.element_hash(:reduce_core_element)
		reduce.default = 0
		@geometry.atomlist_from_selection(@settings[:molecule_a, :selection]).each{|i|
			frozena += core_shells[PeriodicTable.period(@geometry[i].element)]
			frozena -= reduce[@geometry[i].element]
		}
		@geometry.atomlist_from_selection(@settings[:molecule_b, :selection]).each{|i|
			frozenb += core_shells[PeriodicTable.period(@geometry[i].element)]
			frozenb -= reduce[@geometry[i].element]
		}
		sapt[:frozen_core_a] = frozena
		sapt[:frozen_core_b] = frozenb

		return sapt
	end

	def molpro_write_charges
		f = File.open(in_calc_dir("extcharges.txt"), "w+")
		f.puts "Charges"
		f.puts @point_charges.size
		@point_charges.each{|pch|
			f.puts "#{pch.x},#{pch.y},#{pch.z},#{pch.charge},1"
		}
		f.close
	end

	def molpro_run

		# delete molpro.out and molpro.xml
		["molpro.out", "molpro.xml"].each{|delfile|
			File.delete("#{calc_dir}/#{delfile}") if FileTest.exist?("#{calc_dir}/#{delfile}")
		}

		# Run calculation 
		# Common setup
		command = "cd #{calc_dir}; "
		command += "export LD_LIBRARY_PATH=#{@settings[:molpro_ld_path]}:LD_LIBRARY_PATH" if @settings[:molpro_ld_path] != ""

		# Parallel mode
		mode = @settings[:parallel_mode]
		mode = :shm if mode == :default
		case mode
		when :shm
			# Shared memory
			command += "export OMP_NUM_THREADS=#{@settings[:parallel]}; "
			command += "#{@settings[:molpro_exe]} -t #{@settings[:parallel]} -n 1"
		when :mpi
			# MPI
			command += "export OMP_NUM_THREADS=1; "
			command += "#{@settings[:molpro_exe]} --launcher \"#{@settings[:molpro_mpirun]} -np #{@settings[:parallel]} %x\""

		else
			Cuby::error("Molpro interface supports only following parallel modes: shm, mpi")
		end

		# Molpro options
	       	command += " -L #{@settings[:molpro_lib_dir]}" if @settings.set?(:molpro_lib_dir)
		command += " -k '#{@settings[:molpro_key]}'" if @settings[:molpro_key] != ''
		command += " #{MOLPRO_INPUT} > molpro.err 2>&1"

		unless system(command)
			# Look for common errors in the output (MOLPRO_OUTPUT may not be available)
			parser = OutputParser.new(in_calc_dir(MOLPRO_ERROUT), @name)
			parser.add_error(/unable to determine molpro_key/,			"MOLPRO didn't find the license key")
			parser.execute

			# Look for common errors in the output
			parser = OutputParser.new(in_calc_dir(MOLPRO_OUTPUT), @name)
			parser.add_error(/Cannot find default basis/,				"MOLPRO failed, unknown default basis set")
			parser.add_error(/Basis library exhausted/,				"MOLPRO failed, basis set not found")
			parser.add_error(/Unknown parameter .* for set HFOPT:DFUNC:KSOPT/,	"MOLPRO failed, unknown DFT functional")
			parser.add_error(/SPIN not possible/,					"MOLPRO failed, wrong multiplicity/charge combination")
			parser.add_error(/NOT ENOUGH MEMORY/,					"MOLPRO failed, not enough memory")
			parser.add_error(/words of memory are needed/,				"MOLPRO failed, not enough memory")
			parser.add_error(/Catastrophic lack of memory/,				"MOLPRO failed, not enough memory")
			parser.add_error(/insufficient memory available/,			"MOLPRO failed, not enough memory")
			parser.add_error(/I\/O error/,						"MOLPRO failed, I/O error")
			parser.execute


			# Nonzero exit code:
			# no erro for now, it seems that molpro returns it even when everything is OK
		end
	end

	def molpro_read
		results = Results.new
		parser = OutputParser.new(in_calc_dir(MOLPRO_OUTPUT), @name)

		# Open shell
		if @settings[:spin_restricted] == :rhf || (@settings[:spin_restricted] == :auto_uhf && @settings[:multiplicity] == 1)
			if @settings[:multiplicity] != 1
				openshell = "R"
			else
				openshell = ""
			end
		else
			openshell = "U"
		end

		# SCF energy
		parser.add_pattern(:e_scf, /^\s*SETTING ENERGYSCF += +([^ ]*)/, :get => :submatch)

		# Correlation energy
		if @settings[:method] == :mp2
			if @settings[:spin_restricted] == :rhf || (@settings[:spin_restricted] == :auto_uhf && @settings[:multiplicity] == 1)
				parser.add_pattern(:corr_mp2,  /^\s*[DF-]*#{openshell}MP2 correlation energy\s+([^ ]*)/, :get => :submatch)
				if @settings[:multiplicity] == 1
					parser.add_pattern(:corr_mp2s, /^\s*MP2 singlet pair energy\s+([^ ]*)/, :get => :submatch)
					parser.add_pattern(:corr_mp2t, /^\s*MP2 triplet pair energy\s+([^ ]*)/, :get => :submatch)
				end
			else
				parser.add_pattern(:corr_mp2,  /^\s*Correlation energy\s+([^ ]*)/, :get => :submatch)
			end
		end

		if @settings[:method] == :mp3
			parser.add_pattern(:corr_mp2,  /^\s*MP2:\s+([^ ]*)/, :get => :submatch)
			parser.add_pattern(:corr_mp3,  /^\s*MP3 correlation energy\s+([^ ]*)/, :get => :submatch)
			parser.add_pattern(:corr_mp3s, /^\s*MP3 singlet pair energy\s+([^ ]*)/, :get => :submatch)
			parser.add_pattern(:corr_mp3t, /^\s*MP3 triplet pair energy\s+([^ ]*)/, :get => :submatch)
		end

		if [:ccsd, :"ccsd(t)"].include? @settings[:method]
			if @settings[:multiplicity] == 1
				# MP2
				parser.add_pattern(:corr_mp2,  /^\s*MP2 correlation energy:\s+([^ ]*)/, :get => :submatch)
				parser.add_pattern(:corr_mp2s, /^\s*MP2 singlet pair energy:\s+([^ ]*)/, :get => :submatch)
				parser.add_pattern(:corr_mp2t, /^\s*MP2 triplet pair energy:\s+([^ ]*)/, :get => :submatch)
				# CCSD
				parser.add_pattern(:corr_ccsd,  /^\s*CCSD correlation energy\s+([^ ]*)/, :get => :submatch)
				parser.add_pattern(:corr_ccsd_s,  /^\s*CCSD singlet pair energy\s+([^ ]*)/, :get => :submatch)
				parser.add_pattern(:corr_ccsd_t,  /^\s*CCSD triplet pair energy\s+([^ ]*)/, :get => :submatch)
			else
				# MP2
				parser.add_pattern(:corr_mp2,  /^\s*RHF-RMP2 correlation energy\s+([^ ]*)/, :get => :submatch)
				# CCSD
				parser.add_pattern(:corr_ccsd,  /^\s*RCCSD correlation energy\s+([^ ]*)/, :get => :submatch)
			end
		end

		if @settings[:method] == :"ccsd(t)"
			# (T) and [T]
			parser.add_pattern(:corr_triples,  /^\s*Triples \(T\) contribution\s+([^ ]*)/, :get => :submatch)
			parser.add_pattern(:"energy_ccsd[t]",  /^\s*.*CCSD\[T\] energy\s+([^ ]*)/, :get => :submatch)
		end

		# Gradient
		if @what.include?(:gradient)
			case @settings[:method]
			when :hf
				grad_type = "SCF"
			when :mp2
				grad_type = "MP2"
			end
			results.gradient = Gradient.new
			parser.add_block(
				:gradient,
				/#{grad_type} GRADIENT FOR STATE/, 
				/^ +[0-9]+ +([-]*[0-9]*\.[0-9]+) +([-]*[0-9]*\.[0-9]+) +([-]*[0-9]*\.[0-9]+)/, 
				/Nuclear/,
				{:get => :match_array, :type => :float}
			)
		end

		parser.execute{|name, result, count|
			case name
			when :gradient
				x,y,z = result
				results.gradient << Coordinate[x,y,z] * (HARTREE2KCAL / BOHR2ANGSTROM)
			end
		}

		# SCF energy
		results.energy = parser[:e_scf] * HARTREE2KCAL
		results.energy_components[:scf_energy] = results.energy

		# Evaluate MP2 components
		if [:mp2, :ccsd, :"ccsd(t)"].include? @settings[:method]
			results.energy_components[:correlation_mp2] = parser[:corr_mp2] * HARTREE2KCAL
			# Spin components only for singlet
			if @settings[:multiplicity] == 1 && (@settings[:spin_restricted] == :rhf || @settings[:spin_restricted] == :auto_uhf)
				# Correct same-spin and other-spin components have to be calculated
				ss = parser[:corr_mp2t] * HARTREE2KCAL / 1.5
				results.energy_components[:correlation_mp2_os] = results.energy_components[:correlation_mp2] - ss
				results.energy_components[:correlation_mp2_ss] = ss
			end
		end

		if @settings[:method] == :mp3
			results.energy_components[:correlation_mp2] = parser[:corr_mp2] * HARTREE2KCAL
			results.energy_components[:correlation_mp3] = parser[:corr_mp3] * HARTREE2KCAL
			# Correct same-spin and other-spin components have to be calculated
			ss = parser[:corr_mp3t] * HARTREE2KCAL / 1.5
			results.energy_components[:correlation_mp3_os] = results.energy_components[:correlation_mp3] - ss
			results.energy_components[:correlation_mp3_ss] = ss
		end

		# CCSD components
		if [:ccsd, :"ccsd(t)"].include? @settings[:method]
			results.energy_components[:correlation_ccsd] = parser[:corr_ccsd] * HARTREE2KCAL
			# Correct same-spin and other-spin components have to be calculated
			if @settings[:multiplicity] == 1 && (@settings[:spin_restricted] == :rhf || @settings[:spin_restricted] == :auto_uhf)
				ss = parser[:corr_ccsd_t] * HARTREE2KCAL / 1.5
				results.energy_components[:correlation_ccsd_os] = results.energy_components[:correlation_ccsd] - ss
				results.energy_components[:correlation_ccsd_ss] = ss
			end
		end

		# Post-HF
		case @settings[:method]
		when :mp2
			# Add correlation to the SCF energy
			results.energy += results.energy_components[:correlation_mp2]
		when :ccsd
			# Add correlation to the SCF energy
			results.energy += results.energy_components[:correlation_ccsd]
		when :"ccsd(t)"
			results.energy_components[:"correlation_ccsd[t]"] = parser[:"energy_ccsd[t]"] * HARTREE2KCAL - results.energy_components[:scf_energy]
			results.energy_components[:"correlation_ccsd(t)"] = results.energy_components[:correlation_ccsd] + parser[:corr_triples] * HARTREE2KCAL
			# Add correlation to the SCF energy
			results.energy += results.energy_components[:"correlation_ccsd(t)"]
		end

		return results
	end

	def molpro_read_sapt
		results = Results.new
		parser = OutputParser.new(in_calc_dir(MOLPRO_OUTPUT), @name)

		parser.add_pattern(:e_E1pol, /^\s*E1pol +([-]?[0-9,.]+) +\(/, :get => :submatch, :multi=> :last)
		parser.add_pattern(:e_E1exch, /^\s*E1exch +([-]?[0-9,.]+) +\(/, :get => :submatch, :multi=> :last)
		parser.add_pattern(:e_E2ind, /^\s*E2ind +([-]?[0-9,.]+) +\(/, :get => :submatch, :multi=> :last)
		parser.add_pattern(:e_E2ind_exch, /^\s*E2ind-exch +([-]?[0-9,.]+) +\(/, :get => :submatch, :multi=> :last)
		parser.add_pattern(:e_E2disp, /^\s*E2disp +([-]?[0-9,.]+) +\(/, :get => :submatch, :multi=> :last)
		parser.add_pattern(:e_E2disp_exch, /^\s*E2disp-exch +([-]?[0-9,.]+) +\(/, :get => :submatch, :multi=> :last)
		parser.add_pattern(:e_dHF_mH, /^\s*SETTING DELTA_HF\s*=\s+([^ ]*)/, :get => :submatch)


		parser.execute

		# Read components
		results.energy_components[:E1pol] =  parser[:e_E1pol] * HARTREE2KCAL / 1000
		results.energy_components[:E1exch] =  parser[:e_E1exch] * HARTREE2KCAL / 1000
		results.energy_components[:E2ind] =  parser[:e_E2ind] * HARTREE2KCAL / 1000
		results.energy_components[:E2ind_exch] =  parser[:e_E2ind_exch] * HARTREE2KCAL / 1000
		results.energy_components[:E2disp] =  parser[:e_E2disp] * HARTREE2KCAL / 1000
		results.energy_components[:E2disp_exch] =  parser[:e_E2disp_exch] * HARTREE2KCAL / 1000
		results.energy_components[:deltaHF] = parser[:e_dHF_mH] * HARTREE2KCAL / 1000

		# Total energy
		results.energy = 0.0 +
			results.energy_components[:E1pol] +
			results.energy_components[:E1exch] +
			results.energy_components[:E2ind] +
			results.energy_components[:E2ind_exch] +
			results.energy_components[:E2disp] +
			results.energy_components[:E2disp_exch] +
			results.energy_components[:deltaHF]

		# Other useful components
		results.energy_components[:E1_sum] = results.energy_components[:E1pol] + results.energy_components[:E1exch]
		results.energy_components[:E2ind_sum] = results.energy_components[:E2ind] + results.energy_components[:E2ind_exch]
		results.energy_components[:E2disp_sum] = results.energy_components[:E2disp] + results.energy_components[:E2disp_exch]

		return results
	end

	def prepare_dimer_settings
		# Run only once
		return if @dimers_fixed
		@dimers_fixed = true

		# Fix settings for molecule_a and molecule_b
		if @settings.block(:molecule_a) && @settings.block(:molecule_b)
			# Both set
		elsif !@settings.block(:molecule_a) && !@settings.block(:molecule_b)
			# None set - automated setup
			@settings[:molecule_a, :selection] = "auto"
			@settings[:molecule_b, :selection] = "auto"
			# Charge
			if !@settings.set?(:charge) || @settings[:charge] == 0
				Cuby::warning("Charges not defined for monomers, using default value \"0\"")
				@settings[:molecule_a, :charge] = 0
				@settings[:molecule_b, :charge] = 0
			else
				Cuby::error("Charge of monomers can not be automatically assigned")
			end
			# Multiplicity
			if @settings[:multiplicity] != 1
				Cuby::error("Multiplicity of monomers can not be automatically assigned")
			end
		else
			Cuby::error("Two monomers must be defined in the input.")
		end

		# Auto selections
		@settings[:molecule_a, :selection] = "%molecule(1;2)" if @settings[:molecule_a, :selection] == "auto"
		@settings[:molecule_b, :selection] = "%molecule(2;2)" if @settings[:molecule_b, :selection] == "auto"
	end
end
