###############################################################################
#
# MRCC interface
#
# Author: Jan Rezac
# Date created: 2015-06-05
# License: Cuby4 license
# Description: Interface for external calculations in MRCC
# Status: Works
#
################################################################################

#===============================================================================
# Interface to the MRCC program (mainly coupled clusters calculations)
# Free for academic use (license must be signed)
# http://www.mrcc.hu
#===============================================================================

require "classes/tools/output_parser.rb"
require "fileutils"

# ToDo:
# SCF restart

module InterfaceMrcc
	#=======================================================================
	# Interface header
	#=======================================================================
	# Interface status
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "OK, more features to be implemented"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :ghost_atoms]
	MODIFIER = false
	DIRECTORY = "MRCC"
	# Methods provided by the interface:
	METHODS = {
		:"hf"		=> [:energy],
		:"mp2" 		=> [:energy],
		:"ccsd" 	=> [:energy],
		:"ccsd(t)" 	=> [:energy]
	}
	#=======================================================================

	# List of Coupled Clusters methods (which share commmon setup)
	CC_METHODS = [:"ccsd", :"ccsd(t)"]
	# Files to be cleaned up (not needed for restart)
	TEMP_FILES = "[0-9]* fort* DF* TEDAT TEINT abcd* abci abij aibj aijk OEINT SCHOL SROOT"

	def prepare_interface
		# Prepare calculation directory
		if calc_dir_mkdir("MINP", "mrcc.out") == :old
			calc_using_input # Old input should be reused
			return
		end
	
		# MP2 workaround - MP2 without DF ran as CCSD with no iterations
		if @settings[:method] == :mp2 && [:none, :scf].include?(@settings[:density_fitting])
			@mp2_from_cc = true
			Cuby::warning("MP2 job without DF ran using CCSD program")
		end

		# Write the keyword part of the input
		mrcc_write_keywords

		# Create complete input by combining mrcc.in and the current geometry
		mrcc_write_input_geo

		# Save info on the system upon writing the input
		calc_writing_input
	end

	def calculate_interface
		mrcc_calculate unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(calc_dir+"/mrcc.out")
		if @settings[:method] == :hf || (@settings[:method] == :mp2 && ! @mp2_from_cc)
			results = mrcc_read_results_iface
		else
			results = mrcc_read_results_output
		end

		# Check success of CCSD calculation
		@success_ccsd = false
		@success_ccsd = true if @settings[:method] == :ccsd && results.energy_components[:correlation_ccsd]
		@success_ccsd = true if @settings[:method] == :"ccsd(t)" && results.energy_components[:"correlation_ccsd(t)"]


		return results
	end

	def cleanup_interface
		# Delete large files
		if @settings[:job_cleanup]
			calc_dir_delete 
		elsif @settings[:delete_large_files]
			# Delete all but input and important output
			# Temporary files and integrals
			# CC restart if calculation was successful
			files = TEMP_FILES
			files += " ccsd.rst pt.rst" if @success_ccsd
			system("cd #{calc_dir}; rm -f #{files}")
		end
	end

	def priority_interface
		int = @geometry.size.to_f
		int *= @settings[:basisset_zeta] if @settings.set?(:basisset_zeta)
		case @settings[:method]
		when :hf
			return int**4
		when :mp2
			return int**5
		when :mp3
			return int**6
		when :"ccsd(t)"
			return int**7
		end
	end

	#=======================================================================
	# Private methods
	#=======================================================================

	MRCC_METHODS = {
		:hf => "HF",
		:mp2 => "MP2",
		:ccsd => "CCSD",
		:"ccsd(t)" => "CCSD(T)"
	}

	def mrcc_write_keywords
		filename = in_calc_dir("mrcc.in")
		f = File.open(filename,"w+")

		#f.puts "=#{@settings[:]}"
	
		# General setup
		f.puts "# MRCC calculation for Cuby"
		f.puts "# Memory and parallelization"

		# Memory and parallelization
		mem_mb = @settings[:mem] / @settings[:mrcc_parallel_mpi]
		f.puts "mem=#{mem_mb}MB"
		f.puts "mpitasks=#{@settings[:mrcc_parallel_mpi]}"
		# Parallelization for CCSD(T)
		f.puts "ccsdthreads=#{@settings[:mrcc_parallel_ccthreads]}"
		f.puts "ptthreads=#{@settings[:mrcc_parallel_ccthreads]}"


		# Method
		f.puts "# Method setup"
		if @mp2_from_cc
			# MP2 as CC with no iterations
			f.puts "calc=" + MRCC_METHODS[:ccsd]
			f.puts "ccmaxit=0"
		else
			f.puts "calc=" + MRCC_METHODS[@settings[:method]]
		end

		# SCF
		f.puts "# SCF"
		f.puts "scfmaxit=#{@settings[:scf_cycles]}"

		# Method-specific setup
		case @settings[:method]
		when :ccsd, :"ccsd(t)"
			f.puts "# Canonical coupled clusters"
			# Force direct algorithm for now
			if [:correlation, :both].include? @settings[:density_fitting]
				f.puts "ccsdalg=dfdirect"
			end
			# Restart - coupled clusters
			if File.exist?(in_calc_dir("ccsd.rst")) && @settings[:mrcc_restart_cc]
				Cuby::log.puts("MRCC interface: restarting CC calculation")
				f.puts "ccsdrest=ccsd"
			end
		end



		# Local CC
#		if @settings[:mrcc_local_cc] == :cim
#			# Cluster-in-molecules
#			# As described in http://dx.doi.org/10.1063/1.3632085
#			f.puts "localcc=on"
#			f.puts "lmp2dens=on"
#			f.puts "domrad=inf"
#			#lnoepso
#			#lnoepsv
#			#
#			raise "LCC/CIM setup not finished"
#		elsif @settings[:mrcc_local_cc] == :lno
#			# Local natural orbitals
#			# Setup from http://dx.doi.org/10.1063/1.4819401
#			# Confirmed by M. Kallay
#			f.puts "localcc=on"	# Switches Local CC on
#			f.puts "lmp2dens=on" 	# Default value
#			f.puts "domrad=#{@settings[:mrcc_lno_domrad]}"	# Default value 10
#			f.puts "bpcompo=0.985"	# Default value, in the original paper 0.98 was used
#			f.puts "lnoepso=1e-#{@settings[:mrcc_lno_eps_o]}"	# Threshold determining accuracy, default 5
#			f.puts "lnoepsv=1e-#{@settings[:mrcc_lno_eps_v]}"	# Threshold determining accuracy, default 6
#		end

		# Thresholds
		f.puts "# Thresholds"
		f.puts "scftol=#{@settings[:scf_convergence]}"
		f.puts "scfdtol=#{@settings[:density_convergence]}"
		f.puts "cctol=#{@settings[:correlation_convergence]}"

		# Basisset
		f.puts "# Basis set"
		mrcc_write_basis(f, "basis", "", :basisset, :basisset_elements)

		# Density fitting / auxiliary basis
		df_scf = false
		df_corr = false
		case @settings[:density_fitting]
		when :none
			f.puts "dfbasis_scf=none"
			f.puts "dfbasis_cor=none"
		when :scf
			f.puts "dfbasis_cor=none"
			df_scf = true
		when :correlation
			f.puts "dfbasis_scf=none"
			df_corr = true if @settings[:method] != :hf
		when :both
			df_scf = true
			df_corr = true if @settings[:method] != :hf
		end

		# DF SCF
		if df_scf
			unless @settings.set?(:auxiliary_basis_scf)
				# Automatic choice
				f.puts "dfbasis_scf=auto"
			else
				mrcc_write_basis(f, "dfbasis_scf", "-RI-JK", :auxiliary_basis_scf, :auxiliary_basis_scf_elements)
			end
		end

		# DF correlation
		if df_corr
			unless @settings.set?(:auxiliary_basis_mp2)
				# Automatic choice
				f.puts "dfbasis_cor=auto"
			else
				mrcc_write_basis(f, "dfbasis_cor", "-RI", :auxiliary_basis_mp2, :auxiliary_basis_mp2_elements)
			end
		end

		# Custom keywords
		f.puts @settings[:mrcc_extra_keywords]

		# System setup
		f.puts "# System setup"
		f.puts "charge=#{@settings[:charge]}"

		# Frozen core
		if @settings[:correlation_frozen_core]
			if @settings.set?(:frozen_core_elements)
				# user specified core
				ele_core = @settings.elements_hash(:frozen_core_elements)
				core = 0
				@geometry.each{|atom|
					next if atom.properties[:ghost]
					if ele_core[atom.element]
						core += ele_core[atom.element]
					else
						Cuby::error "No. of core orbitals not defined for element #{atom.element}"
					end
				}
				f.puts "core=#{core}"
			else
				# Default treatment
				f.puts "core=frozen"
			end
		else
			f.puts "core=0"
		end

		# Ghost atoms
		ghosts = []
		@geometry.each_index{|i|
			ghosts << (i + 1) if @geometry[i].properties[:ghost]
		}
		if ghosts.size > 0
			f.puts "# Ghost atoms"
			f.puts "ghost=serialno"
			f.puts ghosts.join(",")
		end

		# Geometry header
		f.puts "# Geometry"
		f.puts "unit=angs"
		f.puts "geom=xyz"

		f.close
	end

	def mrcc_write_basis(f, kw,suffix,basis,basis_elements)
		unless @settings.set?(:basisset_elements)
			# Simple case - one basis for everything
			b = @settings[basis]
			s = suffix
			# Skip suffix if already present
			s = "" if b =~ /#{suffix}$/
			s = "" if b =~ /-A[0-9]$/
			s = "" if b =~ /-A[0-9][a-z]$/
			s = "" if b =~ /-AUX$/
			f.puts "#{kw}=#{b}#{s}"
		else
			# Element-specific basis
			f.puts "#{kw}=atomtype"
			ele_basis = @settings.elements_hash(basis_elements)
			@geometry.elements_in_system.each{|element|
				if ele_basis[element]
					b = ele_basis[element]
				else
					b = @settings[basis]
				end
				s = suffix
				# Skip suffix if already present
				s = "" if b =~ /#{suffix}$/
				s = "" if b =~ /-A[0-9]$/
				s = "" if b =~ /-A[0-9][a-z]$/
				s = "" if b =~ /-AUX$/
				f.puts "#{element}:#{b}#{s}"
			}
		end
	end

	def mrcc_write_input_geo
		filename = in_calc_dir("MINP")
		f = File.open(filename,"w+")

		# Insert the contents of mrcc.in
		File.open(in_calc_dir("mrcc.in"),"r") {|f_in|
			while s = f_in.gets do
				f.print s
			end
		}

		# Add geometry
		@geometry.write_xyz(:file => f)

		f.close
	end

	def mrcc_calculate
		# Build input with a new geometry
		mrcc_write_input_geo

		# Prepare commandline to run
		command = "cd #{calc_dir};"
		command << "export PATH=#{@settings[:mrcc_bin_dir]}:$PATH;"
		command << "export OMP_NUM_THREADS=#{@settings[:parallel]/@settings[:mrcc_parallel_mpi]};"
		command << "export MKL_NUM_THREADS=#{@settings[:parallel]/@settings[:mrcc_parallel_mpi]};"
		command << "dmrcc > mrcc.out 2> mrcc.err;"

		# Run mrcc
		unless system(command)
			mrcc_clean_temp_files
			# Parse output
			parser = OutputParser.new(in_calc_dir("mrcc.out"), @name)
			parser.add_error(/Fatal error in exec drpa/, "MP2 failed with message 'Fatal error in exec drpa'.")
			parser.add_error(/Fatal error in scf./, "SCF failed with message 'Fatal error in scf.'.")
			parser.execute
			# If no specific error was caught, raise a general one
			Cuby::error "MRCC returned nonzero exit code (calculation #{@name})"
		end

		# Read errors
		parser = OutputParser.new(in_calc_dir("mrcc.out"), @name)
		parser.add_error(/THE SCF ITERATION HAS NOT CONVERGED/, "SCF not converged.")
		parser.add_error(/Convergence not achieved/, "Coupled clusters not converged.") unless @settings[:method] == :mp2
		begin
			parser.execute
		rescue
			mrcc_clean_temp_files
			raise
		end
	end

	def mrcc_clean_temp_files
		# Temporary files and integrals
		Cuby::log.puts("MRCC interface: Temporary files deleted on crash")
		system("cd #{calc_dir}; rm -f #{TEMP_FILES}")
	end

	def mrcc_read_results_iface
		results = Results.new

		parser = OutputParser.new(in_calc_dir("iface"), @name)

		# SCF
		parser.add_pattern(:e_scf, /^ENERGY *SCF *\S+ *\S+ *\S+ *([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => true})

		# Additional results
		case @settings[:method]
		when :mp2
			parser.add_pattern(:e_mp2, /^ENERGY *MP2 +\S+ *\S+ *\S+ *([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => true})
			parser.add_pattern(:corr_mp2s, /^ENERGY *MP2singlet +\S+ *\S+ *\S+ *([-]*[0-9]*\.[0-9,E,\+,-]+)/,   {:get => :submatch, :type => :float, :required => true})
			parser.add_pattern(:corr_mp2t, /^ENERGY *MP2triplet +\S+ *\S+ *\S+ *([-]*[0-9]*\.[0-9,E,\+,-]+)/,   {:get => :submatch, :type => :float, :required => true})
		end

		# Execute parser
		parser.execute


		# Get the total energy and its components
		case @settings[:method]
		when :hf
			results.energy = parser[:e_scf] * HARTREE2KCAL
		when :mp2
			results.energy_components[:scf_energy] = parser[:e_scf] * HARTREE2KCAL
			results.energy = parser[:e_mp2] * HARTREE2KCAL
			results.energy_components[:correlation_mp2] = results.energy - results.energy_components[:scf_energy]
			# Spin components:
			results.energy_components[:correlation_mp2_os] = parser[:corr_mp2s] * HARTREE2KCAL
			results.energy_components[:correlation_mp2_ss] = parser[:corr_mp2t] * HARTREE2KCAL
		end

		return results
	end

	def mrcc_read_results_output
		results = Results.new

		parser = OutputParser.new(in_calc_dir("mrcc.out"), @name)

		# SCF
		parser.add_pattern(:e_scf, /^ *\*\*\*FINAL HARTREE-FOCK ENERGY: +([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => true})

#		if @settings[:mrcc_local_cc] == :none
			# CCSD
			parser.add_pattern(:e_mp2, /^ *MP2 energy \[au\]: +([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => false})
			parser.add_pattern(:corr_ccsd, /^ *CCSD correlation energy \[au\]: +([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => false})
			# CCSD(T)
			parser.add_pattern(:"corr_ccsd(t)", /^ *CCSD\(T\) correlation energy \[au\]: +([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => false})
#		else
#			# Local CC - read only final local values
#			# MP2 correction
#			parser.add_pattern(:local_correction, /^ *Local MP2 correction \[au\]: +([-]*[0-9]*\.[0-9,E,\+]+)/, {:get => :submatch, :type => :float})
#			# CCSD
#			parser.add_pattern(:e_mp2, /^ *Total LMP2 energy \[au\]: +([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => false})
#			parser.add_pattern(:corr_ccsd, /^ *CCSD correlation energy \[au\]: +([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => false, :multi => :last})
#			# CCSD(T)
#			parser.add_pattern(:"corr_ccsd(t)", /^ *CCSD\(T\) correlation energy \[au\]: +([-]*[0-9]*\.[0-9,E,\+]+)/,   {:get => :submatch, :type => :float, :required => false, :multi => :last})
#		end

		# Execute parser
		parser.execute

		# SCF
		results.energy_components[:scf_energy] = parser[:e_scf] * HARTREE2KCAL
		# MP2
		results.energy_components[:correlation_mp2] = parser[:e_mp2] * HARTREE2KCAL - results.energy_components[:scf_energy]
		
		case @settings[:method]
		when :mp2
			results.energy = results.energy_components[:scf_energy] + results.energy_components[:correlation_mp2]
		when :ccsd
			results.energy_components[:correlation_ccsd] = parser[:corr_ccsd] * HARTREE2KCAL
			results.energy = results.energy_components[:scf_energy] + results.energy_components[:correlation_ccsd]
		when :"ccsd(t)"
			results.energy_components[:correlation_ccsd] = parser[:corr_ccsd] * HARTREE2KCAL
			results.energy_components[:"correlation_ccsd(t)"] = parser[:"corr_ccsd(t)"] * HARTREE2KCAL
			results.energy = results.energy_components[:scf_energy] +  results.energy_components[:"correlation_ccsd(t)"]
		end

		# Local CC
#		if @settings[:mrcc_local_cc] != :none
#			results.energy_components[:local_mp2_correction] = parser[:local_correction] * HARTREE2KCAL
#		end

		return results
	end

end
