################################################################################
#
# TorchMD-NET interface
#
# Author: Martin Novacek, Jan Rezac
# Date created: 2020-12-01
# License: Cuby4 license
# Description: Interface to TorchMD-NET
# Status: Works
#
################################################################################

require "open3" 
require "objspace"
require 'json'
require 'net/http'

module InterfaceTorchmdnet
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :warning
	DEVELOPMENT_STATUS = "Only basic functionality, no error checks"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient]
	MODIFIER = true
	DIRECTORY = "TORCHNET"
	# Methods provided by the interface:
	METHODS = {
		:"torchmd"		=> [:energy, :gradient],
	}
	#=======================================================================
	
	unless $skip_interface_init
		@@server_port = 8000
		port_ok = false
		10.times{
			if system("lsof -i:#{@@server_port}", out: '/dev/null')
				# Port is occupied
				@@server_port += 1
			else
				port_ok = true
				break
			end
		}
		unless port_ok
			Cuby::error "Torchmd-net interface: Free port for the server not found"
		end
		@@uri = URI("http://127.0.0.1:#{@@server_port}")
		@@headers = { 'Content-Type': 'application/json' }

		# Server initialization
		@@server = "My server"
		@@server_id = Process.spawn("python3 #{Cuby.install_dir}/interfaces/torchmdnet/http_server.py #{@@server_port}")

		# Wait for server init
		server_ok = false
		30.times{
			begin
				msg = ["ping"].to_json
				response = Net::HTTP.post(@@uri, msg, @@headers)
				out = JSON.parse(response.read_body)
			rescue
			end
			if out == 'OK'
				#puts "Calculation server ready at port #{@@server_port}"
				server_ok = true
				break
			end
			sleep(1)
		}
		unless server_ok
			Cuby::error "Torchmd-net interface: Server does not respond"
		end

		# Server "destructor"
		ObjectSpace.define_finalizer(@@server, proc {|id| 
			Process.kill('SIGHUP', @@server_id)
		})
	end
	
	def prepare_interface
		# Check config - presence of model file
		unless File.exist?(@settings[:torchmd_model])
			Cuby::error "Keyword 'torchmd_model' should point to a model file"
		end
		# Initialize model
		msg = ["load", @settings[:torchmd_model]].to_json
		# Query server
		response = Net::HTTP.post(@@uri, msg, @@headers)
		out = JSON.parse(response.read_body)
	end

	def calculate_interface
		return torchnet_run_and_read_http
		#return torchnet_run_and_read
	end

	def cleanup_interface
		# Nothing to clean
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
    def torchnet_run_and_read
		#command = "python3 #{interface_dir}/torchnet_on_xyz.py"
		command = "python3 #{interface_dir}/http_client.py"
		energy = nil
		results = Results.new
		Open3.popen3(command){|stdin, stdout, stderr, wait_thr|
			#stdin.puts @settings[:torchmd_model]
			@geometry.write_xyz(:file => stdin)
			stdin.close
			energy = stdout.gets
			if @what.include?(:gradient)
				results.gradient = Gradient.new
				@geometry.size.times{|i|
					gradline = stdout.gets
					x,y,z = gradline.strip.split.map{|x| x.to_f}
					results.gradient << Coordinate[x,y,z] * @settings[:torchmd_multiplier]
				}
			end


			# print error output, if it is not empty
			err = stderr.read
			err.each_line{|line|
				$stderr.puts line if line.strip != ""
			}
		}
		

		results.energy = energy.to_f * @settings[:torchmd_multiplier]
		return results
	end

	# Atom types used by the model
    ATOMTYPES = {:Br => 1, :C => 3, :Ca => 5,:Cl => 7, :F => 9, :H => 10, :I => 12, :K => 13, :Li => 14, :Mg => 15,:N => 17, :Na => 19, :O => 21, :P => 23, :S => 26} #Ca, K, Li, Mg, Na +1, other 0

    def torchnet_run_and_read_http
		results = Results.new
		# Convert geometry
		atomtypes = []
		coordinates = []
		@geometry.each{|atom|
			atomtypes << ATOMTYPES[atom.element]
			coordinates << atom.to_a
		}
		# Assemble the message
		msg = ["geom", [atomtypes, coordinates]].to_json
		# Query server
		#response = Net::HTTP.post(@@uri, msg, @@headers)
		#ene, forces = JSON.parse(response.read_body)

		# Query server, more complicated code which allows defining the timeout
		response = Net::HTTP.start("127.0.0.1", @@server_port, :read_timeout => 15*60) {|http| 
			http.request_post("127.0.0.1", msg)
		}
		ene, forces = JSON.parse(response.read_body)

		results.energy = ene.to_f * KJ2KCAL * @settings[:torchmd_multiplier]

		if @what.include?(:gradient)
			results.gradient = Gradient.new
			@geometry.each_index{|i|
				x = forces[i][0] * KJ2KCAL * -1.0
				y = forces[i][1] * KJ2KCAL * -1.0
				z = forces[i][2] * KJ2KCAL * -1.0
				results.gradient << Coordinate[x,y,z] * @settings[:torchmd_multiplier]
			}
		end

		return results
	end
end
