################################################################################
#
# Turbomole interface
#
# Author: Jan Rezac
# Date created: 2012-09-25
# License: Cuby4 license
# Description: Interface for external calculations in Turbomole
# Status: Works
#
################################################################################

#===============================================================================
# Interface to the Turbomole program
# Commercial
# http://www.turbomole-gmbh.com/
#===============================================================================

#===============================================================================
# ToDo
#-------------------------------------------------------------------------------
# ( ) Auxiliary basis set fix when uknown basis set is used
# ( ) Reuse previous MOs
# ( ) Save energy components
# ( ) MP2 without RI
# ( ) Basis sets from basis set exchange
#===============================================================================


require "classes/tools/output_parser.rb"
require "classes/misc/search_for_file.rb"
require "classes/misc/grep.rb"
require "classes/objects/electrostatic_potential.rb"
require "interfaces/turbomole/classes/turbomole_control.rb"
require "interfaces/turbomole/classes/turbomole_basis.rb"
require "classes/compatibility/erb_template.rb"

module InterfaceTurbomole
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "OK, more features to be implemented"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient, :ghost_atoms, :electric_field, :hessian, :atomic_charges, :point_charges, :point_charges_gradient, :mos]
	MODIFIER = false
	DIRECTORY = "TMOLE"
	# Methods provided by the interface:
	METHODS = {
		:"hf"		=> [:energy, :gradient, :hessian, :atomic_charges, :point_charges, :point_charges_gradient, :mos],
		:"mp2" 		=> [:energy, :gradient],
		:"mp3" 		=> [:energy],
		:"dft" 		=> [:energy, :gradient, :hessian, :atomic_charges, :point_charges, :point_charges_gradient, :mos],
		:"ccsd" 	=> [:energy],
		:"ccsd(t)" 	=> [:energy, :point_charges],
		:"rpa" 		=> [:energy],
		:"am1"		=> [:energy, :atomic_charges],
	}
	SOLVENTS = [:cosmo, :cosmo_rs]
	# Atomic charges
	ATOMIC_CHARGES = [:nbo, :mulliken, :esp]
	#=======================================================================
	
	# Extra output files
	FILE_POLARIZABILITY = "polarizability"

	# Deleting files
	# Large files - cleaned after successful run when only output is needed
	DELETE_LARGE_FILES = "mos CC* RIF* RIR* slave*.output master ewald_1 statistics MPI-TEMPDIR*"
	# Temporary files - cleaned when calculation fails
	DELETE_TEMP_FILES = "CC* RIF* RIR* MPI-TEMPDIR*"


	def prepare_interface
		# Check version and compatibility

		# DZVP-DFT basis automatically
		if @settings[:basisset].downcase == "dzvp-dft-auto"
			@settings[:basisset] = "DZVP-DFT"
			@settings[:basisset_file] = "DZVP-DFT.basis"
		end

		# COSMO-RS
		if @settings[:solvent_model] == :cosmo_rs
			# Input check
			Cuby::error("COSMO-RS interface supports only energy calculations") if @what.include?(:gradient)
			Cuby::error("COSMO-RS interface implements only solvent_name 'h2o'") unless @settings[:solvent_name].downcase == "h2o"

			# Check method setup			
			Cuby::recommendation("COSMO-RS is parametrized only for DFT calculations with BP functional", :once) unless @settings[:method] == :dft && @settings[:functional].downcase.gsub("-","") == "bp"
			# Check basis set
			case @settings[:cosmo_rs_params]
			when :bp_tzvpd_fine
				cosmobas = "def2-tzvpd"
			when :bp_tzvp_cosmo
				cosmobas = "def-tzvp"
			when :bp_svp_am1
				cosmobas = "def-svp"
			end
			Cuby::recommendation("The COSMO-RS parameters work only with #{cosmobas} basis set.", :once) unless cosmobas == @settings[:basisset].downcase

			# COSMO-RS needs a child calculation in gas phase
			gas_phase_settings = @settings.child_block_as_copy(:cosmors_gas_phase)
			gas_phase_settings[:solvent_model] = :none
			calc_name = (@name + "_gasphase")
			@cosmors_gas_phase_calc = Calculation.new(calc_name, gas_phase_settings, @geometry)
			@cosmors_gas_phase_calc.prepare([:energy])

			# Set epsilon to infinity
			@settings[:solvent_epsilon] = -1
		end

		# Multiplicity
		if @settings[:multiplicity] != 1 && @settings[:spin_restricted] == :rhf
			Cuby::error "For spin multiplicity > 1, only unrestricted calculations are supported,\n (set keyword 'spin_restricted: uhf')"
		end
		
		# RI
		@scf_ri = @settings[:density_fitting] == :scf || @settings[:density_fitting] == :both
		@correlation_ri = @settings[:density_fitting] == :correlation || @settings[:density_fitting] == :both

		if @scf_ri && @what.include?(:hessian)
			Cuby::error("Hessian calculation can not be performed with the RI approximation")
		end

		# Method checks
		#if @settings[:method] == :mp2 && !@correlation_ri
		#	# MP2 not implemented, only RI-MP2 works
		#	Cuby::error("MP2 calculation without RI currently not supported by the interface", self)
		#end

		if @settings[:method] == :rpa && (!@correlation_ri || !@scf_ri)
			# Only RI-RPA possible
			Cuby::error("RPA calculation without RI not supported", self)
		end
		if @settings[:spin_component_scaling] && ! (@settings[:method] == :mp2 && @correlation_ri)
			# SCS available only for RI-MP2
			Cuby::error("SCS is at present available only for RI-MP2")
		end
		if @settings[:method] == :mp3 && @settings[:turbomole_version] < 6.5
			Cuby::error("MP3 is not available in turbomole version lower than 6.5")
		end

		if @settings[:explicit_correlation] != :none
			unless @settings[:method] == :mp2 && ! @what.include?(:gradient) && @correlation_ri
				Cuby::error("Explicit corrlation is available only for RI-MP2 energy calculations for now")
			end
		end

		# MP2 module (needed in prepare as it affects also reading results)
		if @settings[:method] == :mp2 && !@correlation_ri
		  @correlation_module = :mpgrad
		else
		  if @settings[:turbomole_version] >= 7.0
			  if [:mp3, :ccsd, :"ccsd(t)"].include?(@settings[:method])
				  @correlation_module = :ccsdf12
			  else
				  @correlation_module = :ricc2
			  end
		  else
			  @correlation_module = :ricc2
		  end
		end

		# Symmetry
		if @settings[:use_symmetry]
			 if @settings[:explicit_correlation] != :none
				Cuby::error "F12 calculations possible only without use_symmetry (#{@name})"
			end
			 @geometry.each{|atom|
				if atom.properties[:ghost]
					Cuby::warning "Symmetry can't be used with ghost atoms in turbomole,\nswitching it off (#{@name})"
					@settings[:use_symmetry] = false
				end
			 }
		end

		# DFT setup
		if [:dft, :rpa].include?(@settings[:method])
			require "classes/calculation/dft_setup.rb"
			@dft_setup = DftSetup.new(interface_dir + "/dft_functionals.yaml", interface_dir + "/dft_grids.yaml")
		end

		# Prepare calculation directory
		if calc_dir_mkdir("control", "energy") == :old
			calc_using_input
			return
		end

		# Write geometry
		if @settings.set?(:turbomole_prep_geo)
			# Alternative geometry used just for prepare
			prepgeo = Geometry.load_from_settings(@settings, :turbomole_prep_geo)
			unless @geometry.same_size_as?(prepgeo)
				Cuby::error("Geometry from 'turbomole_prep_geo' does not match calculation geometry - difefrent size (calculation #{@name})")
			end
			unless @geometry.same_element_order?(prepgeo)
				Cuby::error("Geometry from 'turbomole_prep_geo' does not match calculation geometry - difefrent atom order (calculation #{@name})")
			end
			GeoFile::TurbomoleCoord.write(prepgeo, :file => calc_dir+"/coord")
			Cuby.log.puts_debug("Turbomole: using custom geometry for calculation preparation")
		else
			GeoFile::TurbomoleCoord.write(@geometry, :file => calc_dir+"/coord")
		end

		# Prepare input
		calc_writing_input
		turbomole_run_define
		turbomole_cosmoprep if @settings[:solvent_model] == :cosmo || @settings[:solvent_model] == :cosmo_rs
		turbomole_fix_control

		# If other geometry is used for prepare, replace it by real one
		if @settings.set?(:turbomole_prep_geo)
			GeoFile::TurbomoleCoord.write(@geometry, :file => calc_dir+"/coord")
		end

		# COSMO-RS
		if @settings[:solvent_model] == :cosmo_rs
			cosmo_rs_write_input
		end
	end

	def calculate_interface
		# Extra calculation for COSMO-RS
		if @settings[:solvent_model] == :cosmo_rs
			# Run the calculation now, bypassing the queue
			@cosmors_gas_phase_calc.calculate(1)
			cosmo_rs_write_energy_file(@cosmors_gas_phase_calc.results.energy)
		end

		turbomole_run_calculation unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(calc_dir+"/energy")

		results = turbomole_read_results

		if @settings[:solvent_model] == :cosmo_rs
			cosmo_rs_run
			cosmo_rs_read(results)
		end

		return results
	end

	def cleanup_interface
		if @settings[:job_cleanup]
			# Delete everything
			calc_dir_delete 
		elsif @settings[:delete_large_files]
			# Delete only large files (and some unnecessary files as well)
			system("cd #{calc_dir}; rm -rf #{DELETE_LARGE_FILES}")
		end

		# Extra calculation for COSMO-RS
		if @settings[:solvent_model] == :cosmo_rs
			@cosmors_gas_phase_calc.cleanup
		end
	end

	def priority_interface
		int = @geometry.size.to_f
		int *= @settings[:basisset_zeta] if @settings.set?(:basisset_zeta)
		case @settings[:method]
		when :dft
			return int**4
		when :hf
			return int**4
		when :mp2
			return int**5
		when :mp3
			return int**6
		when :"ccsd"
			return int**6
		when :"ccsd(t)"
			return int**7
		end
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def turbomole_run_define

		# Experimental: build own basis file
		#basis_builder = TurbomoleBasis.new(@geometry, settings)
		#basis_builder.build_basis_ao(in_calc_dir("basis2"))


		# Copy basis set file(s) if needed
		if @settings.set?(:basisset_file)
			filenames = []
			@settings[:basisset_file].split(/\s*[\,,;,:]\s*/).each{|fn|
				filenames << SearchForFile.search(
					:filename => fn,
					:locations => ['./', interface_dir + '/basissets', '/'],
					:extensions => ['', '.basis']
				)
			}
			# At least one file must be set up
			if filenames.size == 0
				Cuby::error("At least one basis set file must be provided in keyword 'basisset_file' (calculation #{@name})")
			end

			if filenames.size == 1
				# One file - simple copy
				bfn = filenames.first
				FileUtils.cp(bfn, calc_dir+"/basis")
			else
				# Multiple files - merge
				turbomole_merge_basis_files(filenames, calc_dir+"/basis")
			end
		end

		# Load input templates: general
		template_common = erb_template(IO.read(interface_dir + "/templates/define_common.erb"), nil, "-")

		# Create input for define
		File.open(calc_dir+"/define.in","w+") {|file|
			file.print(template_common.result(binding))
		}

		# Copy the control_template file
		FileUtils.cp(interface_dir+"/templates/control_template", calc_dir+"/control_template")

		# Run define to do the general setup
		system("cd #{calc_dir}; export TURBODIR=#{@settings[:turbomole_turbodir]}; cat define.in | #{@settings[:turbomole_bin_dir]}/define > define.out 2>&1");

		# Look for errors in define output
		unless Grep::file_unicode(calc_dir+"/define.out", /define ended normally/)
			Cuby::error("Turbomole define script failed, see #{calc_dir}/define.out for details (calculation #{@name})")
		end

		# Check for specific errors / define output
		parser = OutputParser.new(in_calc_dir("define.out"), @name)
		parser.add_error(/THERE ARE NO DATA SETS CATALOGUED IN FILE/, "Basis set not found in the library")
		parser.add_pattern(:multiplicity, /^ *FOUND .* WITH MULTIPLICITY +([0-9]+)/, :get => :submatch, :type => :as_is, :required => false)
		parser.add_pattern(:multiplicity_cs, /^ *FOUND CLOSED SHELL SYSTEM/, :get => :line, :type => :as_is, :required => false)
		parser.add_pattern(:electrons_in_atom, /^ *NUMBER OF ELECTRONS IN YOUR ATOM IS +([0-9]+)/, :get => :submatch, :type => :as_is, :required => false)
		parser.add_pattern(:multiplicity_s_trivial, /SINCE ALL MOLECULAR ORBITALS ARE DOUBLY/, :get => :line, :type => :as_is, :required => false)
		parser.execute

		# Check multiplicity
		if @settings[:multiplicity] == 1 && (@settings[:spin_restricted] == :rhf || @settings[:spin_restricted] == :auto_uhf)
			multiplicity = -1
			if parser[:multiplicity_cs] || parser[:multiplicity_s_trivial]
				multiplicity = 1 
			elsif parser[:multiplicity]
				multiplicity = parser[:multiplicity].to_i
			elsif parser[:electrons_in_atom]
				# It is an atom, multiplicity not in the output of define
				# -> not checked, assume the input is correct
				multiplicity = @settings[:multiplicity]
			end
			if multiplicity != @settings[:multiplicity]
				Cuby::error("Multiplicity determined by turbomole's define is different from the one\nspecified in the input. (#{@name})")
			end
		end

		if @settings[:turbomole_version] >= 7.3
			# In 7.3 and above, it is simple to specify memory regardless of paralellization mode
			if @settings.set?(:mem_core_scf)
				ri_scf_core =  @settings[:mem] * @settings[:mem_core_scf] / 100
			else
				ri_scf_core =  @settings[:mem] * @settings[:mem_core] / 100
			end

			if @settings.set?(:mem_core_correlation)
				ri_corr_core = @settings[:mem] * @settings[:mem_core_correlation] / 100
			else
				ri_corr_core = @settings[:mem] * @settings[:mem_core] / 100
			end
		else
			# Memory setup, depends on parallelization mode
			# The :mem keyword should specify total memory, per-core memory is calculated here
			# ri_scf_core and ri_corr_core variables have to be set

			# In serial jobs, MPI and SHM/GA parallelization, memory per core is used
			mem_divide = 1
			case @settings[:parallel_mode]
			when :mpi, :shm
				mem_divide = @settings[:parallel]
			end

			if @settings.set?(:mem_core_scf)
				ri_scf_core =  @settings[:mem] * @settings[:mem_core_scf] / 100 / mem_divide
			else
				ri_scf_core =  @settings[:mem] * @settings[:mem_core] / 100 / mem_divide
			end

			if @settings.set?(:mem_core_correlation)
				ri_corr_core = @settings[:mem] * @settings[:mem_core_correlation] / 100
			else
				ri_corr_core = @settings[:mem] * @settings[:mem_core] / 100
			end
			ri_corr_core /= mem_divide if @settings[:parallel_mode] == :mpi # divider applied only to MPI
		end


		# Other methods than HF: Define is run second time to update the control with method-specific setup
		# For some reason, it does not work when combined into one input
		
		# Load input templates
		method_templates = []
		case @settings[:method]
		when :hf
			if @scf_ri
				# RI-HF setup
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_rihf.erb"), nil, "-")
			else
				# Normal HF does not need additional setup
			end
		when :dft
			if @scf_ri
				# RI-DFT
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_ridft.erb"), nil, "-")
			else
				# Normal DFT
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_dft.erb"), nil, "-")
			end

			# Double-hybrid functionals
			if @dft_setup.functional(@settings) == "b2-plyp"
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_custom_cbas.erb"), nil, "-") if @settings.set?(:auxiliary_basis_mp2)
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_ricc2_rimp2.erb"), nil, "-")
			end

		when :mp2
			# Only RI-MP2 enabled, checked above
			if @scf_ri
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_rihf.erb"), nil, "-")
			end
            if @correlation_ri
			    method_templates << erb_template(IO.read(interface_dir + "/templates/define_custom_cbas.erb"), nil, "-") if @settings.set?(:auxiliary_basis_mp2)
			    method_templates << erb_template(IO.read(interface_dir + "/templates/define_ricc2_rimp2.erb"), nil, "-")
            else
                method_templates << erb_template(IO.read(interface_dir + "/templates/define_mp2.erb"), nil, "-")
            end
			if @settings[:explicit_correlation] != :none
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_f12.erb"), nil, "-")
			end
		when :mp3
			if @scf_ri
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_rihf.erb"), nil, "-")
			end
			method_templates << erb_template(IO.read(interface_dir + "/templates/define_custom_cbas.erb"), nil, "-") if @settings.set?(:auxiliary_basis_mp2)
			method_templates << erb_template(IO.read(interface_dir + "/templates/define_ricc2_mp3.erb"), nil, "-")
		when :"ccsd"
			if @scf_ri
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_rihf.erb"), nil, "-")
			end
			method_templates << erb_template(IO.read(interface_dir + "/templates/define_custom_cbas.erb"), nil, "-") if @settings.set?(:auxiliary_basis_mp2)
			method_templates << erb_template(IO.read(interface_dir + "/templates/define_ricc2_ccsd.erb"), nil, "-")
		when :"ccsd(t)"
			if @scf_ri
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_rihf.erb"), nil, "-")
			end
			method_templates << erb_template(IO.read(interface_dir + "/templates/define_custom_cbas.erb"), nil, "-") if @settings.set?(:auxiliary_basis_mp2)
			method_templates << erb_template(IO.read(interface_dir + "/templates/define_ricc2_ccsdt.erb"), nil, "-")
		when :rpa
			# DFT setup
			if @scf_ri
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_ridft.erb"), nil, "-")
			else
				method_templates << erb_template(IO.read(interface_dir + "/templates/define_dft.erb"), nil, "-")
			end
			# RPA setup
			method_templates << erb_template(IO.read(interface_dir + "/templates/define_rirpa.erb"), nil, "-")
		end

		# DFT-specific setup - make it accessible in templates
		if [:dft, :rpa].include?(@settings[:method])
			dft_setup = @dft_setup
		end

		method_templates.each_index{|i|
			# Create input for define
			File.open(calc_dir+"/define2_#{i}.in","w+") {|file|
				file.print(method_templates[i].result(binding))
			}
			# Run define
			system("cd #{calc_dir}; export TURBODIR=#{@settings[:turbomole_turbodir]}; cat define2_#{i}.in | #{@settings[:turbomole_bin_dir]}/define > define2_#{i}.out 2>&1");
			# Look for errors in define output
			unless Grep::file_unicode(calc_dir+"/define2_#{i}.out", /define ended normally/)
				Cuby::error("Turbomole define script failed, see #{calc_dir}/define2_#{i}.out for details")
			end
		}

		# Electrostatic filed
		if @settings.set?(:electric_field)
			template_ef = erb_template(IO.read(interface_dir + "/templates/define_electric_field.erb"))
			File.open(calc_dir+"/define_ef.in","w+") {|file| file.print(template_ef.result(binding))}
			system("cd #{calc_dir}; export TURBODIR=#{@settings[:turbomole_turbodir]}; cat define_ef.in | #{@settings[:turbomole_bin_dir]}/define > define_ef.out 2>&1");
			Cuby::error("Turbomole define script failed, see #{calc_dir}/define_ef.out for details") unless Grep::file_unicode(calc_dir+"/define_ef.out", /define ended normally/)
		end

		# Properties
		if @settings[:properties].include?(:static_polarizability)
			template_ef = erb_template(IO.read(interface_dir + "/templates/define_properties_polarizability.erb"))
			fn_part = "prop_polari"
			File.open(calc_dir+"/define_#{fn_part}.in","w+") {|file| file.print(template_ef.result(binding))}
			system("cd #{calc_dir}; export TURBODIR=#{@settings[:turbomole_turbodir]}; cat define_#{fn_part}.in | #{@settings[:turbomole_bin_dir]}/define > define_#{fn_part}.out 2>&1");
			Cuby::error("Turbomole define script failed, see #{calc_dir}/define_#{fn_part}.out for details") unless Grep::file_unicode(calc_dir+"/define_#{fn_part}.out", /define ended normally/)
		end
	end

	def turbomole_cosmoprep
		# Load ERB template
		template = erb_template(IO.read(interface_dir + "/templates/cosmoprep.erb"))
		# Parse it and write to file
		File.open(calc_dir+"/cosmoprep.in","w+") {|file|
			file.print(template.result(binding))
		}
		# Run cosmoprep
		system("cd #{calc_dir}; export TURBODIR=#{@settings[:turbomole_turbodir]}; cat cosmoprep.in | #{@settings[:turbomole_bin_dir]}/cosmoprep > cosmoprep.out 2>&1");
		# Look for errors in cosmoprep output
		#--------------------------------------------------
		# unless Grep::file_unicode(calc_dir+"/define2_#{i}.out", /define ended normally/)
		# 	Cuby::error("Turbomole define script failed, see #{calc_dir}/define2_#{i}.out for details")
		# end
		#-------------------------------------------------- 
	end

	def turbomole_fix_control
		# Check for presence of some important flags
		found_flags = []
		
		control = IO.readlines(calc_dir+"/control")
		f = File.open(calc_dir+"/control", "w+")
		control.each{|line|
			if line =~ /^\$ricc2/
				f.print line
				if @settings.set?(:turbomole_ccsd_cycles)
					f.puts "  maxiter=#{@settings[:turbomole_ccsd_cycles]}"
				end
				if @settings[:spin_component_scaling]
					# SCS caling for RICC2 module
					Cuby::error "Turbomole: spin_component_scaling and sos_scaling can not be set at the same time" if @settings[:sos_scaling]
					f.puts "  scs  cos=#{@settings[:scs_os]}d0  css=#{@settings[:scs_ss]}d0"
				end
				if @settings[:sos_scaling]
					# SOS caling for RICC2 module
					Cuby::error "Turbomole: spin_component_scaling and sos_scaling can not be set at the same time" if @settings[:spin_component_scaling]
					# Laplace transformation setup first
					if @settings[:sos_scaling_laplace] != 0
						f.puts "$laplace"
						f.puts "  conv=#{@settings[:sos_scaling_laplace]}"
					end
					# Then the ricc2 keyword
					f.puts "  sos"
				end
				if @settings[:method] == :"ccsd" && @settings[:turbomole_version] <= 6.2
					# Turbomole 6.2 define does not add the "ccsd" line to control
					f.puts "  ccsd"
				end
				if @settings[:method] == :"ccsd(t)" && @settings[:turbomole_version] <= 6.2
					# Turbomole 6.2 define does not add the "ccsd(t)" line to control
					f.puts "  ccsd(t)"
				end
			elsif line =~ /^\$scforbitalshift/ && @settings.set?(:scf_levelshift)
				f.puts "$scforbitalshift  automatic=#{@settings[:scf_levelshift]}"
			elsif line =~ /^\$drvopt/
				f.print line
				# Switch on calculation of gradient on point charges
				if @point_charges
					f.puts "   point charges"
				end
			elsif line =~ /^\$scfconv/
				# SCF convergence may have been changed accidentaly, set it back to cuby's value
				f.puts "$scfconv #{@settings[:scf_convergence]}"
				found_flags << "$scfconv"
			elsif line =~ /^\$denconv/
				# Density convergence may have been changed accidentaly, set it back to cuby's value
				f.puts "$denconv 1.0d-#{@settings[:density_convergence]}"
				found_flags << "$denconv"
			elsif line =~ /^\$dft/
				f.puts "$dft"
				if @settings[:turbomole_weight_derivatives]
					f.puts "   weight derivatives"
				end
			elsif line =~ /^\$maxcor/ && @settings[:turbomole_version] >= 7.3
				# Fix memory specification to be per node
				val = line.strip.split[1]
				f.puts "$maxcor #{val} MiB per_node"
			elsif line =~ /^\$ricore/ && @settings[:turbomole_version] >= 7.3
				# Fix memory specification to be per node
				val = line.strip.split[1]
				f.puts "$ricore #{val} MiB per_node"
			elsif line =~ /^\$freeze/ && @settings.set?(:frozen_core_elements)
				if @settings.set?(:reduce_core_element)
					Cuby::error "Keyword reduce_core_element can not combined with frozen_core_elements"
				end
				f.puts "$freeze"
				ele_core = @settings.elements_hash(:frozen_core_elements)
				core = 0
				@geometry.each{|atom|
					next if atom.properties[:ghost]
					if ele_core[atom.element]
						core += ele_core[atom.element]
					else
						Cuby::error "No. of core orbitals not defined for element #{atom.element}"
					end
				}
				f.puts " implicit core= #{core} virt= 0"
			elsif line =~ /^ *implicit core/ && @settings.set?(:frozen_core_elements)
				# Skip this line
			elsif line =~ /^ *implicit core/ && ! @settings.set?(:frozen_core_elements)
				# Modify frozen core
				if @settings.set?(:reduce_core_element)
					element_n = @settings.elements_hash(:reduce_core_element)
					element_n.default = 0
					reduce = 0
					@geometry.each{|atom|
						if element_n[atom.element] != 0 && ! atom.properties[:ghost]
							reduce += element_n[atom.element]
						end
					}
					# parse line
					s1,s2, core, s3, virt = line.strip.split
					f.puts " implicit core= #{core.to_i - reduce} virt= #{virt}"
				end
			elsif line =~ /^\$end/
				# Add things to the end of control
				# Missing but important flags
				f.puts "$scfconv #{@settings[:scf_convergence]}" unless found_flags.include?("$scfconv")
				f.puts "$denconv 1.0d-#{@settings[:density_convergence]}" unless found_flags.include?("$denconv")
				# Hessian calculations
				if @what.include?(:hessian)
					# Hessian calculation - force output to files (automatic in version >= 6.3)
					f.puts "$hessian (projected)   file=hessian"
					f.puts "$vibrational normal modes   file=vib_normal_modes"
					f.puts "$vibrational spectrum   file=vibspectrum"
					f.puts "$dipgrad   file=dipgrad"

					# Hessian calculation - save unprojected hessian
					f.puts "$noproj"
					f.puts "$nprhessian   file=hessian_npr"
					f.puts "$nprvibrational normal modes   file=vib_normal_modes_npr"
					f.puts "$nprvibrational spectrum   file=vibspectrum_npr"
				end
				# Atomic charges
				if what.include?(:atomic_charges)
					case @settings[:atomic_charges]
					when :mulliken
						f.puts "$pop"
					when :nbo
						f.puts "$pop nbo"
					when :esp
						f.puts "$esp_fit kolman"
					end
				end
				if @point_charges
					f.puts "$point_charges file=point_charges"
					f.puts "$point_charge_gradients file=point_charge_gradients"
				end

				# properties
				if @settings[:properties].include?(:static_polarizability)
					f.puts "$electronic polarizability   file=#{FILE_POLARIZABILITY}"
				end

				# All orbitals instead of just some
				if @what.include?(:mos)
					f.puts "$moprint" 
				end

				# COSMO-RS
				if @settings[:solvent_model] == :cosmo_rs
					if @settings[:cosmo_rs_params] == :bp_tzvpd_fine
						f.puts "$isorad"
					end
				end

				# ESP on grid
				if @settings.set?(:turbomole_esp_grid)
					turbomole_esp_grid_to_control(f, @settings[:turbomole_esp_grid])
				end

				# Add user-defined keywords
				f.puts @settings[:turbomole_keywords] if @settings.set?(:turbomole_keywords)
				f.print line
			else
				f.print line
			end
		}
		f.close
	end

	ERROR_CHECK_POSITIVE = {
		"dscf" => "dscf ended normally",
		"ridft" => "ridft : all done",
		"ricc2" => "ricc2 ended normally"

	}

	def turbomole_run_calculation
		# Write current geometry
		GeoFile::TurbomoleCoord.write(@geometry, :file => calc_dir+"/coord")

		# Write point charges
		if @point_charges
			write_point_charges
		end

		# Get list of executables to be run for selected method
		run_modules = []
		extra_scf_gradient = ''
		if [:hf, :dft, :rpa].include?(@settings[:method])
			if @scf_ri
				run_modules = ["ridft"]
				run_modules << "rdgrad" if @what.include?(:gradient)
			else
				run_modules = ["dscf"]
				run_modules << "grad" if @what.include?(:gradient)
			end

			if @settings[:method] == :dft && @dft_setup.functional(@settings) == "b2-plyp"
				run_modules << "ricc2"
			end

			if @settings[:method] == :rpa
				run_modules << "rirpa"
			end
		elsif [:mp2, :mp3, :ccsd, :"ccsd(t)"].include?(@settings[:method])
			# (RI-MP2 only)
			unless @settings[:turbomole_skip_scf]
				if @scf_ri
					run_modules = ["ridft"]
					if @what.include?(:gradient) && @settings[:turbomole_scf_grad]
						run_modules << "rdgrad"
						extra_scf_gradient = "rdgrad"
					end

				else
					run_modules = ["dscf"]
					if @what.include?(:gradient) && @settings[:turbomole_scf_grad]
						run_modules << "grad"
						extra_scf_gradient = "grad"
					end
				end
			end
			# In version 7.0, methods above MP2 are run using ccsdf12 module
            if @settings[:method] == :mp2 && !@correlation_ri
              # MP2 without correlation - needs mp2prep and mpgrad
              if @what.include?(:gradient)
                run_modules << "mp2prep -g"
              else
                run_modules << "mp2prep -e"
              end
              run_modules << "mpgrad"
              @correlation_module = :mpgrad
            else
              if @settings[:turbomole_version] >= 7.0
                  if [:mp3, :ccsd, :"ccsd(t)"].include?(@settings[:method])
                      run_modules << "ccsdf12"
                      @correlation_module = :ccsdf12
                  else
                      run_modules << "ricc2"
                      @correlation_module = :ricc2
                  end
              else
                  run_modules << "ricc2"
                  @correlation_module = :ricc2
              end
            end

		end

		if @what.include?(:hessian)
			run_modules << "aoforce"
		end

		# Properties
		run_modules << "escf" if @settings[:properties].include?(:static_polarizability)

		# Delete files with results so that new ones are created
		unless @settings[:turbomole_skip_scf]
			FileUtils.rm_f(calc_dir+"/energy")
			FileUtils.rm_f(calc_dir+"/gradient")
			FileUtils.rm_f(calc_dir+"/scf_gradient") if @settings[:turbomole_scf_grad]
		end

		# Some control keys have to be saved as they are modified for difefrent steps of calculations
		control = TurbomoleControl.new(calc_dir+"/control")
		save_maxcor = control.get_value("maxcor")
		save_ricore = control.get_value("ricore")
		#puts "maxcor value is #{save_maxcor}"
		#puts "ricore value is #{save_ricore}"


		# Run calculations
		run_modules.each_index{|mi|
            # Get module name, without arguments
            mod = run_modules[mi].gsub(/\s.*/,'')
            # Get arguments
            mod_arguments = run_modules[mi].gsub(/^\S*/,'')
			# Reduce $maxcor while doing SCF
			if @settings.set?(:mem_maxcor_max_scf)
				control = TurbomoleControl.new(calc_dir+"/control")
				if ["ridft", "dscf", "grad", "rdgrad"].include?(mod)
					control.set_value("maxcor", "#{@settings[:mem_maxcor_max_scf]}")
				else
					control.set_value("maxcor", save_maxcor)
				end
				control.write!
			end

			# Remove ricore for the SCF part
			control = TurbomoleControl.new(calc_dir+"/control")
			if ["ridft", "dscf", "grad", "rdgrad"].include?(mod)
				control.delete_key("ricore")
			else
				control.set_value("ricore", save_ricore)
			end
			control.write!


			# Parallel setup
			if @settings[:parallel] == 1 
				parallel = "export PARNODES=1; PARA_ARCH=\"\";"
			elsif @settings[:serial_scf] && ["dscf","ridft","rdgrad","grad"].include?(mod)
				# SCF will be serial / threaded only
				parallel = "export PARNODES=#{@settings[:parallel]}; export PARA_ARCH=\"\";"
			else
				mode = @settings[:parallel_mode]
				if mode == :default
					if defmode = @settings[:tm_parallel_mode_defaults][mod]
						mode = defmode.downcase.to_sym
					elsif defmode = @settings[:tm_parallel_mode_defaults]["default"]
						mode = defmode.downcase.to_sym
					else
						Cuby::error("Turbomole: default parallel mode not defined for module #{mod}")
					end
					#puts "parallel_mode defaults to #{mode} for module #{mod}"
				end
				case mode
				when :omp
					parallel = "export PARNODES=#{@settings[:parallel]}; export PARA_ARCH=\"\";"
				when :shm
					parallel = "export PARNODES=#{@settings[:parallel]}; export PARA_ARCH=SMP;"
				when :mpi
					parallel = "export PARNODES=#{@settings[:parallel]}; export PARA_ARCH=MPI;"
				end
			end

			# Name of the output file
			outfile = "#{mod}.out"

			# Build command running the module, save as a script
			# Save command as a script
			File.open(in_calc_dir("_run_#{mi}_#{mod}.sh"),"w+"){|f| 
				f.puts "#!/bin/bash"
				f.puts "export TURBODIR=#{@settings[:turbomole_turbodir]}; "
				f.puts parallel
				f.puts "source $TURBODIR/Config_turbo_env; "
				f.puts "export MPI_ROOT=$TURBODIR/mpirun_scripts/`sysname`/PMPI; "
				
				if @settings[:turbomole_save_system_info]
					f.puts "echo 'Node name:' > #{mod}.meminfo;"
					f.puts "uname -n >> #{mod}.meminfo; "
					f.puts "echo >> #{mod}.meminfo;"
					f.puts "echo 'Disk space (df)' >> #{mod}.meminfo;"
					f.puts "df -h >> #{mod}.meminfo; "
					f.puts "echo >> #{mod}.meminfo;"
					f.puts "echo 'Memory (free)' >> #{mod}.meminfo;"
					f.puts "free -g >> #{mod}.meminfo; >> #{mod}.meminfo;"
				end

				f.puts "which #{mod} > #{mod}.out; "
				f.puts "#{mod} #{mod_arguments} >> #{mod}.out 2>&1;"
			}

			# Run the module
			system("cd #{calc_dir}; bash _run_#{mi}_#{mod}.sh")

			# Copy SCF gradient if requested
			if mod == extra_scf_gradient
				system("cd #{calc_dir}; mv gradient scf_gradient")
			end

			# Check for errors
			if FileTest.exist?(calc_dir+"/dscf_problem")
				turbomole_clean_temp_files
				Cuby::error "SCF not converged (calculation #{@name})" 
			end

			# Check the output for errors
			if ERROR_CHECK_POSITIVE[mod]
				output_check = OutputParser.new(calc_dir+"/"+outfile, @name)
				output_check.add_pattern(:check_positive, /#{ERROR_CHECK_POSITIVE[mod]}/, {:get => :line, :multi => :last, :required => false})
				output_check.execute
				if output_check[:check_positive]
					Cuby::log.puts_debug("Turbomole module '#{mod}' finished successfully")
				else
					turbomole_clean_temp_files
					Cuby::error("Turbomole module '#{mod}' failed (calculation #{@name})")
				end
			end
		}

	end

	def turbomole_clean_temp_files
		# If requested, delete large files when the calculation fails
		if @settings[:delete_temp_files_on_fail]
			Cuby::log.puts "Large temporary files deleted (calculation #{@name})"
			system("cd #{calc_dir}; rm -rf #{DELETE_TEMP_FILES}")
		end
	end

	def write_point_charges
		f = File.open(in_calc_dir("point_charges"), "w+")
		f.puts "$point_charges"
		@point_charges.each { |pcharge|
			f.puts "   #{pcharge.x * ANGSTROM2BOHR} #{pcharge.y * ANGSTROM2BOHR} #{pcharge.z * ANGSTROM2BOHR} #{pcharge.charge}\n"	
		}
		f.puts "$end"
		f.close
	end

	def turbomole_read_gradient(filename)
		gradient = Gradient.new
		File.open(filename,"r"){|gradfile|
			# 2 lines header
			2.times {gradfile.gets}
			# Coordinates
			@geometry.size.times {gradfile.gets}
			# Gradient
			@geometry.size.times {
				g = gradfile.gets.strip.gsub("D","e").split.map{|x| x.to_f}
				gradient << Coordinate[g[0],g[1],g[2]] * HARTREE2KCAL / BOHR2ANGSTROM
			}
		}
		return gradient
	end

	def turbomole_read_point_charges_gradient(filename)
		gradient = Gradient.new
		File.open(filename,"r"){|gradfile|
			# 1 line header
			gradfile.gets
			# Gradient
			@point_charges.size.times {
				g = gradfile.gets.strip.gsub("D","e").split.map{|x| x.to_f}
				gradient << Coordinate[g[0],g[1],g[2]] * HARTREE2KCAL / BOHR2ANGSTROM
			}
		}
		return gradient
	end

	def turbomole_read_hessian_from_file(filename, projected = true)
		rows = []
		(@geometry.size * 3).times {|i| rows[i] = []}
		f = File.open(in_calc_dir(filename))
		read = false
		f.each_line{|line|
			if line =~ /\$hessian \(projected\)/ && projected || line =~ /\$nprhessian/ && !projected
				read = true
				next
			end
			if line =~ /^\$/ && read
				break
			end
			if read
				rowindex = line[0..2].to_i - 1
				line_a = line[5..-1].chomp.split
				line_a.each_index {|i|
					rows[rowindex] << line_a[i].to_f * HARTREE2KCAL / BOHR2ANGSTROM / BOHR2ANGSTROM
				}

			end

		}
		f.close
		hessian = Hessian.rows(rows)
		return hessian
	end

	def read_atomic_charges(filename)
		parser = OutputParser.new(filename, @name)
		charges = AtomicCharges.new(@settings[:atomic_charges])
		case @settings[:atomic_charges]
		when :nbo
			parser.add_block(
				:atomic_charges,
				/atomic populations from total density/, 
				/^ *[0-9]+ +[^ ]+ +([-]*[0-9]*\.[0-9]+)/, 
				/=====/,
				{:get => :submatch, :type => :float}
			)
		when :mulliken
			parser.add_block(
				:atomic_charges,
				/atomic populations from total density/, 
				/^ *[0-9][^ ]+ +([-]*[0-9]*\.[0-9]+)/, 
				/moments/,
				{:get => :submatch, :type => :float}
			)
		when :esp
			parser.add_block(
				:atomic_charges,
				/charges resulting from fit/, 
				/^ *[0-9]+ [a-z]+ +[0-9]*\.[0-9]+ +([-]*[0-9]*\.[0-9]+)/, 
				/Note/,
				{:get => :submatch, :type => :float}
			)
		end
		parser.execute{|name, result, count|
			charges << result
		}
		return charges
	end

	def turbomole_read_polarizability
		lines = IO.readlines(in_calc_dir(FILE_POLARIZABILITY)).map!{|s| s.strip.split}
		polarizability = Polarizability.new("a.u.")

		polarizability.matrix[0,0] = lines[1][0].to_f

		polarizability.matrix[1,0] = polarizability.matrix[0,1] = lines[2][0].to_f
		polarizability.matrix[1,1] = lines[2][1].to_f

		polarizability.matrix[2,0] = polarizability.matrix[0,2] = lines[3][0].to_f
		polarizability.matrix[2,1] = polarizability.matrix[1,2] = lines[3][1].to_f
		polarizability.matrix[2,2] = lines[3][2].to_f

		return polarizability
	end

	def turbomole_read_dipole(results, tm_module)
		case tm_module
		when :ridft
			filename = in_calc_dir("ridft.out")
		when :dscf
			filename = in_calc_dir("dscf.out")
		end

		if filename
			parser = OutputParser.new(filename, @name)
			dipole = []
			parser.add_block(
				:dipole,
				/^ *dipole moment *$/, 
				/^ *[x-z] +[-]*[0-9]*\.[0-9]+ +[-]*[0-9]*\.[0-9]+ +([-]*[0-9]*\.[0-9]+)/, 
				/debye/,
				{:get => :submatch, :type => :float}
			)
			parser.execute {|name, result, count|
				case name
				when :dipole
					dipole[count] = result
				end
			}
			results.multipoles[:dipole] = Multipoles::Dipole.new(
				dipole[0] * AU2DEBYE * DEBYE2CUBY,
				dipole[1] * AU2DEBYE * DEBYE2CUBY,
				dipole[2] * AU2DEBYE * DEBYE2CUBY
			)
		end
		return nil
	end

	def turbomole_read_orbitals(results, tm_module)
		results.molecular_orbitals = MolecularOrbitals.new

		case tm_module
		when :ridft
			filename = in_calc_dir("ridft.out")
		when :dscf
			filename = in_calc_dir("dscf.out")
		end

		if filename
			parser = OutputParser.new(filename, @name)
			moblocks = []
			parser.add_block(
				:mos,
				/^ *orbitals \$scfmo/, 
				/\S/, 
				/=====/,
				{:get => :line}
			)
			this_block = nil
			parser.execute {|name, result, count|
				case name
				when :mos
					if result =~ /irrep/
						# Initialize new block
						this_block = []
						moblocks << this_block
						# Read the line
						values = result.strip.split
						values.shift # remove line name
						this_block << values
					end

					if result =~ /eigenvalues H/
						# Read the line
						values = result.strip.split
						values.shift # remove line name
						values.shift # remove line name
						values.map!{|x| x.to_f}
						this_block << values

					end

					if result =~ /occupation/
						# Read the line
						values = result.strip.split
						values.shift # remove line name
						values.map!{|x| x.to_f}
						this_block << values
					end
				end
			}

			moblocks.each{|block|
				names, energies, occupations = block
				names.each_index{|i|
					name = names[i]
					energy = energies[i]
					if occupations && occupations[i]
						occ = occupations[i]
					else
						occ = 0.0
					end

					results.molecular_orbitals << MolecularOrbital.new(name, energy, occ)
				}
			}
		end
		results.molecular_orbitals.sort_by_e!
		results.molecular_orbitals = nil if results.molecular_orbitals.size == 0
		return nil
	end

	def turbomole_read_results
		results = Results.new

		# Read the energy file
		energy_file_data = nil
		IO.readlines(calc_dir+"/energy").each {|line|
			if line =~ / *[0-9]/
				energy_file_data = line.strip.split
			end
		}

		# Read the gradient file and save the gradient
		if @what.include?(:gradient)
			results.gradient = turbomole_read_gradient(in_calc_dir("gradient"))
			if @settings[:turbomole_scf_grad]
				results.gradient_components[:scf_gradient] = turbomole_read_gradient(in_calc_dir("scf_gradient"))
				results.gradient_components[:correlation_gradient] = results.gradient - results.gradient_components[:scf_gradient]
			end

			# Read gradient on point charges
			if @point_charges
				@point_charges.gradient = turbomole_read_point_charges_gradient(in_calc_dir("point_charge_gradients"))
			end
		end

		# Some results must be read from the output file of the correlation energy module
		if @settings[:method] == :dft && @dft_setup.functional(@settings) == "b2-plyp"
			ricc_reader = OutputParser.new(calc_dir+"/ricc2.out", @name)
			ricc_reader.add_pattern(:energy_b2plyp, /^ +\* *B2-PLYP energy.*: +(\S+)/, {:get => :match_array})
			ricc_reader.execute
		end

		if @settings[:method] == :mp2 && @correlation_module != :mpgrad
			ricc_reader = OutputParser.new(calc_dir+"/ricc2.out", @name)
			ricc_reader.add_pattern(:corr_mp2, /^ +\* +MP2 correlation energy.*: +(\S+)/, {:get => :match_array, :required => false})
			ricc_reader.add_pattern(:corr_total, /^ +\* +total correlation energy.*: +(\S+)/, {:get => :match_array, :required => false})
			# ricc_reader.add_pattern(:total_e_mp2, /^ +Total Energy +: +(\S+)/, {:get => :match_array, :required => false})
			if @settings[:explicit_correlation] != :none
				ricc_reader.add_pattern(:f12_hf_correction, / +\* +correction to HF from CABS singles.*: +(\S+)/, {:get => :match_array, :required => false})
				ricc_reader.add_pattern(:f12_mp2_correction, / +\* +fixed.*contr\..*: +(\S+)/, {:get => :match_array, :required => false})
			end
			#if @settings[:spin_component_scaling] && !(@what.include?(:gradient))
				ricc_reader.add_pattern(:corr_mp2_os_ss, /^.*E\(OS\) += +(\S+) +E\(SS\) += +(\S+)/, {:get => :match_array, :required => false})
				if @settings[:explicit_correlation] != :none
					ricc_reader.add_pattern(:corr_mp2f12_os_ss, /^.*F12\(OS\) += +(\S+) +F12\(SS\) += +(\S+)/, {:get => :match_array, :required => false})
				end
			#end
			ricc_reader.execute
		end
		if @settings[:method] == :mp2 && @correlation_module == :mpgrad
			ricc_reader = OutputParser.new(calc_dir+"/mpgrad.out", @name)
			if @settings[:multiplicity] == 1
				ricc_reader.add_pattern(:corr_mp2, /^ +\* +MP2-energy.*: +(\S+)/, {:get => :match_array, :required => false})
			else
				ricc_reader.add_pattern(:corr_mp2, /^ +\* +MP2-energy\(tot\).*: +(\S+)/, {:get => :match_array, :required => false})
			end
			ricc_reader.execute
        end

		if @settings[:method] == :mp3
			if @correlation_module == :ricc2
				ricc_reader = OutputParser.new(calc_dir+"/ricc2.out", @name)
			else # ccsdf12 module
				ricc_reader = OutputParser.new(calc_dir+"/ccsdf12.out", @name)
			end
			ricc_reader.add_pattern(:corr_mp2, /^ +\* *MP2 correlation energy.*: +(\S+)/, {:get => :match_array, :multi => :last})
			ricc_reader.add_pattern(:corr_mp2_os_ss, /^.*E\(OS\) += +(\S+) +E\(SS\) += +(\S+)/, {:get => :match_array})
			ricc_reader.add_pattern(:corr_mp3, /^ +\* *MP3 correlation energy.*: +(\S+)/, {:get => :match_array})
			ricc_reader.execute
		end

		if @settings[:method] == :"ccsd"
			if @correlation_module == :ricc2
				ricc_reader = OutputParser.new(calc_dir+"/ricc2.out", @name)
			else # ccsdf12 module
				ricc_reader = OutputParser.new(calc_dir+"/ccsdf12.out", @name)
			end
			ricc_reader.add_pattern(:corr_mp2, /^ +\* +MP2 correlation energy.*: +(\S+)/, {:get => :match_array})
			ricc_reader.add_pattern(:corr_mp2_os_ss, /^.*E\(OS\) += +(\S+) +E\(SS\) += +(\S+)/, {:get => :match_array})
			ricc_reader.add_pattern(:corr_ccsd, /^ +\* +correlation energy +: +(\S+)/, {:get => :match_array})
			ricc_reader.execute
		end

		if @settings[:method] == :"ccsd(t)"
			if @correlation_module == :ricc2
				ricc_reader = OutputParser.new(calc_dir+"/ricc2.out", @name)
			else # ccsdf12 module
				ricc_reader = OutputParser.new(calc_dir+"/ccsdf12.out", @name)
			end
			ricc_reader.add_pattern(:corr_mp2, /^ +\* +MP2 correlation energy.*: +(\S+)/, {:get => :match_array})
			ricc_reader.add_pattern(:corr_mp2_os_ss, /^.*E\(OS\) += +(\S+) +E\(SS\) += +(\S+)/, {:get => :match_array})
			ricc_reader.add_pattern(:corr_ccsd, /^ +\* +CCSD correlation energy +: +(\S+)/, {:get => :match_array})
			ricc_reader.add_pattern(:corr_ccsdt, /^ +\* +total correlation energy +: +(\S+)/, {:get => :match_array})
			ricc_reader.add_pattern(:ccsdt_e5, /^ +\* +E5 singles and triples contribution +: +(\S+)/, {:get => :match_array})
			ricc_reader.execute
		end


		# Save the energy
		case @settings[:method]
		when :hf
			results.energy = energy_file_data[1].to_f * HARTREE2KCAL
			results.atomic_charges = read_atomic_charges(@scf_ri ? calc_dir+"/ridft.out" : calc_dir+"/dscf.out") if @what.include?(:atomic_charges)
		when :dft
			results.energy = energy_file_data[1].to_f * HARTREE2KCAL
			results.atomic_charges = read_atomic_charges(@scf_ri ? calc_dir+"/ridft.out" : calc_dir+"/dscf.out") if @what.include?(:atomic_charges)
			if @dft_setup.functional(@settings) == "b2-plyp"
				results.energy = ricc_reader[:energy_b2plyp][0] * HARTREE2KCAL
			end
		when :rpa
			results.energy = energy_file_data[1].to_f * HARTREE2KCAL
			results.energy_components[:scf_energy] = results.energy
			e_corr_rpa = energy_file_data[4].to_f * HARTREE2KCAL
			results.energy_components[:correlation_rpa] = e_corr_rpa
			results.energy += e_corr_rpa
		when :mp2
			# SCF energy
			results.energy = energy_file_data[1].to_f * HARTREE2KCAL
			results.energy_components[:scf_energy] = results.energy
			# MP2 energy
			begin
				ec_mp2 = ricc_reader[:corr_total][0] * HARTREE2KCAL
			rescue
				begin
					ec_mp2 = ricc_reader[:corr_mp2][0] * HARTREE2KCAL
				rescue
					# Last option: read it from the energy file
					ec_mp2 = energy_file_data[4].to_f * HARTREE2KCAL
				end

			end
			results.energy += ec_mp2

			# Plain MP2 or scaled variants: name the component accordingly
			if @settings[:spin_component_scaling]
				results.energy_components[:correlation_scs_mp2] = ec_mp2
			elsif @settings[:sos_scaling]
				results.energy_components[:correlation_sos_mp2] = ec_mp2
			else
				results.energy_components[:correlation_mp2] = ec_mp2
			end

			# SCS-MP2 - spin components, available only in energy calculation
			# Gradient calcualtion yields only the final result
			#if !(@what.include?(:gradient)) && @settings[:spin_component_scaling] && 
			if @correlation_module == :ricc2 && ricc_reader[:corr_mp2_os_ss]
				#--------------------------------------------------
				# if ricc_reader[:corr_mp2f12_os_ss]
				# 	results.energy_components[:correlation_mp2_f12_os] = ricc_reader[:corr_mp2f12_os_ss][0] * HARTREE2KCAL
				# 	results.energy_components[:correlation_mp2_f12_ss] = ricc_reader[:corr_mp2f12_os_ss][1] * HARTREE2KCAL
				# 	results.energy_components[:correlation_mp2_f12] = results.energy_components[:correlation_mp2_f12_os] + results.energy_components[:correlation_mp2_f12_ss]
				# end
				#-------------------------------------------------- 
				results.energy_components[:correlation_mp2_os] = ricc_reader[:corr_mp2_os_ss][0] * HARTREE2KCAL
				results.energy_components[:correlation_mp2_ss] = ricc_reader[:corr_mp2_os_ss][1] * HARTREE2KCAL
				results.energy_components[:correlation_mp2] = results.energy_components[:correlation_mp2_os] + results.energy_components[:correlation_mp2_ss]
			end

			# F12
			if @settings[:explicit_correlation] != :none
				results.energy_components[:f12_hf_correction] = ricc_reader[:f12_hf_correction][0] * HARTREE2KCAL
				results.energy_components[:f12_mp2_correction] = ricc_reader[:f12_mp2_correction][0] * HARTREE2KCAL
				results.energy_components[:scf_energy_f12] = results.energy_components[:scf_energy] + results.energy_components[:f12_hf_correction]
				results.energy_components[:correlation_mp2_f12] = results.energy_components[:correlation_mp2] 
				results.energy_components[:correlation_mp2] -= results.energy_components[:f12_mp2_correction]
				results.energy += results.energy_components[:f12_hf_correction]
			end

		when :mp3
			results.energy_components[:scf_energy] = energy_file_data[1].to_f * HARTREE2KCAL
			# MP2
			results.energy_components[:correlation_mp2] = ricc_reader[:corr_mp2][0] * HARTREE2KCAL
			# MP2 - spin components
			results.energy_components[:correlation_mp2_os] = ricc_reader[:corr_mp2_os_ss][0] * HARTREE2KCAL
			results.energy_components[:correlation_mp2_ss] = ricc_reader[:corr_mp2_os_ss][1] * HARTREE2KCAL
			# MP3
			results.energy_components[:correlation_mp3] = ricc_reader[:corr_mp3][0] * HARTREE2KCAL + results.energy_components[:correlation_mp2] 
			# Total
			results.energy = results.energy_components[:scf_energy] + results.energy_components[:correlation_mp3]
		when :"ccsd"
			# SCF energy
			results.energy_components[:scf_energy] = energy_file_data[1].to_f * HARTREE2KCAL
			# MP2 - same as MP3
			results.energy_components[:correlation_mp2] = ricc_reader[:corr_mp2][0] * HARTREE2KCAL
			# MP2 - spin components - same as MP3
			results.energy_components[:correlation_mp2_os] = ricc_reader[:corr_mp2_os_ss][0] * HARTREE2KCAL
			results.energy_components[:correlation_mp2_ss] = ricc_reader[:corr_mp2_os_ss][1] * HARTREE2KCAL
			# CCSD
			results.energy_components[:correlation_ccsd] = ricc_reader[:corr_ccsd][0] * HARTREE2KCAL
			# Total
			results.energy = results.energy_components[:scf_energy] + results.energy_components[:correlation_ccsd]
		when :"ccsd(t)"
			# SCF energy
			results.energy_components[:scf_energy] = energy_file_data[1].to_f * HARTREE2KCAL
			# MP2 - same as MP3
			results.energy_components[:correlation_mp2] = ricc_reader[:corr_mp2][0] * HARTREE2KCAL
			# MP2 - spin components - same as MP3
			results.energy_components[:correlation_mp2_os] = ricc_reader[:corr_mp2_os_ss][0] * HARTREE2KCAL
			results.energy_components[:correlation_mp2_ss] = ricc_reader[:corr_mp2_os_ss][1] * HARTREE2KCAL
			# CCSD
			results.energy_components[:correlation_ccsd] = ricc_reader[:corr_ccsd][0] * HARTREE2KCAL
			# CCSD(T)
			ec_ccsdt = ricc_reader[:corr_ccsdt][0] * HARTREE2KCAL
			results.energy_components[:"correlation_ccsd[t]"] = ec_ccsdt - ricc_reader[:ccsdt_e5][0] * HARTREE2KCAL
			results.energy_components[:"correlation_ccsd(t)"] = ec_ccsdt
			# Total
			results.energy = results.energy_components[:scf_energy] + ec_ccsdt
		end

		# read dipole
		if [:hf, :dft].include?(@settings[:method])
			if @scf_ri
				turbomole_read_dipole(results, :ridft)
			else
				turbomole_read_dipole(results, :dscf)
			end
		end

		# read MOs
		if @what.include?(:mos)
			if @scf_ri
				turbomole_read_orbitals(results, :ridft)
			else
				turbomole_read_orbitals(results, :dscf)
			end
		end

		# Read electrostatic potential
		if @settings.set?(:turbomole_esp_grid)
			results.electrostatic_potential = turbomole_esp_grid_read(in_calc_dir("tp.xyz"))
			results.electrostatic_potential.write_xyzp("electrostatic_potential.dat")
		end

		# Read the hessian
		if @what.include?(:hessian)
			#results.hessian = turbomole_read_hessian_from_file("hessian", true) # Projected hessian
			results.hessian = turbomole_read_hessian_from_file("hessian_npr", false) # Unprojected hessian
		end

		# Gradient on point charges
		if @what.include?(:point_charges_gradient)
			results.point_charges_gradient = @point_charges.gradient
		end

		# properties
		if @settings[:properties].include?(:static_polarizability)
			results.static_polarizability = turbomole_read_polarizability
		end

		return results
	end

	def turbomole_merge_basis_files(filenames, outfile)
		#: merge basis sets from filenames array, creating a single file
		#: named outfile

		# Open output file
		of = File.open(outfile, "w+")

		# Write comment
		of.puts "# Basis set file build by Cuby, merging following files:"
		filenames.each{|fn|
			of.puts "# - #{fn}"
		}
		# Write header
		of.puts "$basis"

		filenames.each{|fn|
			copy = false
			infile = File.open(fn, "r")
			while line = infile.gets
				copy = false if line =~ /^\$end/
				if copy
					of.puts line
				end
				copy = true if line =~ /^\$basis/
			end
			infile.close
		}

		# Write footer
		of.puts "$end"
		# Close output file
		of.close
	end

	def turbomole_esp_grid_to_control(control_file, grid_file_name)
		# Write the grid for ESP calculation into the control file

		# Check if file exists
		unless FileTest.exist?(grid_file_name)
			Cuby::Error("File specified by the keyword turbomole_esp_grid does not exist")
		end

		# Write header
		control_file.puts '$pointval pot fmt=xyz geo=point'

		# Read grid file and write to control
		gridfile = File.open(grid_file_name)
		while line = gridfile.gets do
			line.strip!
			next if line == ""
			next if line =~ /^#/
			xyz = line.split.map{|v| v.to_f * ANGSTROM2BOHR}
			control_file.printf("  %12.6f%12.6f%12.6f\n", xyz[0], xyz[1], xyz[2])
		end
		gridfile.close
	end

	def turbomole_esp_grid_read(filename)
		esp = ElectrostaticPotential.new()
		espfile = File.open(filename)

		espfile = File.open(filename)
		while line = espfile.gets do
			line.strip!
			next if line == ""
			next if line =~ /^#/
			x, y, z, p = line.split.map{|v| v.to_f}
			esp.add_point_au(x,y,z,p)
		end
		espfile.close

		return esp
	end

	#=======================================================================
	# COSMO-RS
	#=======================================================================
	
	def cosmo_rs_write_energy_file(gas_phase_e)
		# Write energy
		File.open(in_calc_dir("out.energy"), "w+") {|f|
			f.puts gas_phase_e * KCAL2HARTREE
		}
		# Copy mos as well
		FileUtils.cp(@cosmors_gas_phase_calc.calc_dir + "/mos", in_calc_dir("mos"));

	end

	def cosmo_rs_write_input
		# Write COSMO-RS input
		# Now the method and solvent are hard-coded
		cdir = @settings[:turbomole_cosmotherm_dir] + "/COSMOtherm/CTDATA-FILES"
		case @settings[:cosmo_rs_params]
		when :bp_tzvpd_fine
			fdir = @settings[:turbomole_cosmotherm_dir] + "/COSMOtherm/DATABASE-COSMO/BP-TZVPD-FINE"
			ctdfile = "BP_TZVPD_FINE_C30_1701.ctd"
		when :bp_tzvp_cosmo
			fdir = @settings[:turbomole_cosmotherm_dir] + "/COSMOtherm/DATABASE-COSMO/BP-TZVP-COSMO"
			ctdfile = "BP_TZVP_C30_1701.ctd"
		when :bp_svp_am1
			fdir = @settings[:turbomole_cosmotherm_dir] + "/COSMOtherm/DATABASE-COSMO/BP-SVP-AM1"
			ctdfile = "BP_SVP_AM1_C30_1701.ctd"
		end

		ldir = @settings[:turbomole_cosmotherm_dir] + "/licensefiles"

		f = File.open(in_calc_dir("cosmo_rs.inp"), "w+")
		f.puts "ctd=#{ctdfile} cdir=#{cdir} fdir=#{fdir} ldir=#{ldir}"
		f.puts "AUTOC wcmn EHFile"
		f.puts "Title..."
		f.puts "f=h2o_c0"
		f.puts "f=out.ccf fdir=. ef=out.energy"
		f.puts "tc=25.00 Henry=1 GSOLV"
		f.close	
	end

	def cosmo_rs_run
		Cuby::log.puts_debug("Running cosmotherm...")
		command = "cd #{calc_dir}; #{@settings[:turbomole_cosmotherm_dir]}/COSMOtherm/BIN-LINUX/cosmotherm cosmo_rs.inp"
		system(command)
	end
	
	def cosmo_rs_read(results)
		# The last solvation energy belongs to the solute
		parser = OutputParser.new(calc_dir+"/cosmo_rs.out", @name)
		parser.add_pattern(:solv_g, /^ *Gibbs Free Energy of Solvation *: +(\S+)/, {:get => :submatch, :type => :float, :multi => :last})
		parser.execute

		# Save it instead of the total energy
		results.energy_components[:total_energy] = results.energy # It's already in kcal/mol
		results.energy = parser[:solv_g]
	end
end
