################################################################################
#
# xTB interface
#
# Author: Jan Rezac
# Date created: 2017-05-04
# License: Cuby4 license
# Description: Interface for external calculations using Grimme's xTB code
# Status: Works
#
################################################################################

#===============================================================================
# Interface to the xTB code
# DOI: 10.1021/acs.jctc.7b00118
#===============================================================================

require "classes/tools/output_parser.rb"
require "json"

module InterfaceXtb
	#=======================================================================
	# Interface header
	#=======================================================================
	DEVELOPMENT_FLAG = :ok
	DEVELOPMENT_STATUS = "Only basic functionality, no error checks"
	# Interface information
	INTERFACE = :calculation_external
	CAPABILITIES = [:energy, :gradient, :atomic_charges]
	MODIFIER = false
	DIRECTORY = "XTB"
	# Methods provided by the interface:
	METHODS = {
		:"gfn-xtb"		=> [:energy, :gradient, :atomic_charges],
		:"gfn2-xtb"		=> [:energy, :gradient, :atomic_charges],
		:"gfn-ff"		=> [:energy]
	}
	SOLVENTS = [:gbm, :alpb, :cosmo, :cpcmx]
	ATOMIC_CHARGES = [:mulliken]
	#=======================================================================
	
	XTB_SCRIPT = "xtb.sh"
	XTB_INPUT = "xtb.inp"
	XTB_OUT = "xtb.out"
	XTB_ERR = "xtb.err"

	def prepare_interface
		# Check config - presence of xtb executable
		if @settings[:xtb_home_dir] == ""
			Cuby::error "Path to xtb not set - use keyword 'xtb_home_dir' in your input or global config files."
		else
			unless File.exist?(@settings[:xtb_home_dir])
				Cuby::error "Keyword 'xtb_home_dir' points to a non-existent path"
			end
			unless File.directory?(@settings[:xtb_home_dir])
				Cuby::error "Keyword 'xtb_home_dir' does not point to a directory"
			end
			unless File.exist?(@settings[:xtb_home_dir]+"/bin/xtb")
				Cuby::error "Xtb executable not found at 'xtb_home_dir'/bin/xtb"
			end

		end

		# Prepare calculation directory
		if calc_dir_mkdir(XTB_SCRIPT, XTB_OUT) == :old
			calc_using_input # Old input should be reused
			return
		end

		# Write the input script
		xtb_write_input

		# Write the geometry
		xtb_write_geometry

		# Save info on the system upon writing the input
		calc_writing_input
	end

	def calculate_interface
		xtb_run unless @settings[:existing_calc_dir] == :read_results && FileTest.exist?(in_calc_dir(XTB_OUT))
		if @settings[:method] == :"gfn-ff"
			return xtb_read_old
		else
			return xtb_read
		end
	end

	def cleanup_interface
		calc_dir_delete if @settings[:job_cleanup]
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def xtb_write_input
		# Build commandline options for xtb
		options = []

		# extra options
		options << @settings[:xtb_options] 

		# Calculation
		if @what.include?(:gradient)
			options << "--grad"
		else
			options << "--sp"
		end
		options << "--pop" if @what.include?(:atomic_charges)

		# Solvent
		# Must be activated from commandline
		if @settings[:solvent_model] == :gbm
			options << "--gbsa #{@settings[:solvent_name]}"
        elsif @settings[:solvent_model] == :alpb
			options << "--alpb #{@settings[:solvent_name]}"
        elsif @settings[:solvent_model] == :cosmo
			options << "--cosmo #{@settings[:solvent_name]}"
        elsif @settings[:solvent_model] == :cpcmx
			options << "--cpcmx #{@settings[:solvent_name]}"
		end

		if @settings[:method] == :"gfn-ff"
			options << "--gfnff"
			options << "--chrg #{@settings[:charge]}"
		else
			# Write XTB input
			options << "--input  #{XTB_INPUT}"
			filename = in_calc_dir(XTB_INPUT)
			f = File.open(filename,"w+")
				# basic data
				f.puts "$chrg #{@settings[:charge]}"
				f.puts "$spin #{@settings[:multiplicity] - 1}"
				# method
				f.puts "$gfn"
				f.puts "  method=1" if @settings[:method] == :"gfn-xtb"
				f.puts "  method=2" if @settings[:method] == :"gfn-xtb2"
				# write control
				f.puts "$write"
				f.puts "  output file=properties.out"
				f.puts "  json=true"
				f.puts "$end"
			f.close
		end

		

		# Write a script that runs xtb with these options
		filename = in_calc_dir(XTB_SCRIPT)
		f = File.open(filename,"w+")
		f.puts "export XTBPATH=#{@settings[:xtb_home_dir]}"
		f.puts "export OMP_NUM_THREADS=#{@settings[:parallel]}"
		f.puts "export MKL_NUM_THREADS=#{@settings[:parallel]}"
		f.puts "export OMP_STACKSIZE=500m" if @settings[:parallel] > 1
		f.puts "#{@settings[:xtb_home_dir]}/bin/xtb " + options.join(" ") + " xtb_geo.xyz"
		f.close
	end

	def xtb_write_geometry
		filename = in_calc_dir("xtb_geo.xyz")
		@geometry.write_xyz(:file => filename)
	end

	def xtb_run
		xtb_write_geometry
		command = "cd #{calc_dir}; sh #{XTB_SCRIPT} > #{XTB_OUT} 2> #{XTB_ERR}" 

		unless system(command)
			Cuby::error "XTB returned nonzero exit code (calculation #{@name})"
		end
	end

	def xtb_read
		results = Results.new
		begin
			xtbdata = JSON.parse(File.read(calc_dir+"/xtbout.json"))
		rescue
			Cuby::error "Failed parsing the JSON file produced by XTB (calculation #{@name})"
		end

		# Energy
		results.energy = xtbdata["total energy"] * HARTREE2KCAL

                # Development test: reading total energy from the output
		#results.energy = `grep ":: total energy" #{calc_dir}/xtb.out`.strip.split[3].to_f * HARTREE2KCAL

		# Atomic charges
		results.atomic_charges = AtomicCharges.new(:mulliken)
		xtbdata["partial charges"].each{|ch| results.atomic_charges << ch}

		# Gradient
		if @what.include?(:gradient)
			results.gradient = xtb_read_gradient(in_calc_dir("gradient"))
		end

		return results
	end

	def xtb_read_old

		parser = OutputParser.new(calc_dir+"/"+XTB_OUT, @name)
		# Reading energy - try format from different versions
		parser.add_pattern(:energy, /TOTAL ENERGY *([-]*[0-9]*\.[0-9]+)/, {:required => false, :get => :submatch})
		parser.add_error(/convergence criteria cannot be satisfied within/, "XTB interface: SCC not converged")
		parser.execute

		results = Results.new
		if parser[:energy]
			results.energy = parser[:energy] * HARTREE2KCAL
		else
			Cuby::error "Energy not found in XTB output (#{@name})"
		end

		return results
	end

	def xtb_read_gradient(filename)
		gradient = Gradient.new
		File.open(filename,"r"){|gradfile|
			# 2 lines header
			2.times {gradfile.gets}
			# Coordinates
			@geometry.size.times {gradfile.gets}
			# Gradient
			@geometry.size.times {
				g = gradfile.gets.strip.gsub("D","e").split.map{|x| x.to_f}
				gradient << Coordinate[g[0],g[1],g[2]] * HARTREE2KCAL / BOHR2ANGSTROM
			}
		}
		return gradient
	end

end
