
module ProtocolAtomicCharges
	#=======================================================================
	# Protocol header
	#=======================================================================
	PROTOCOL_TYPE = :elementary_calculation
	# Input structure
	INPUT_BLOCKS = [
		InputBlock[:fragment_N,          :optional, "Blocks defining fragments, N=(1..X)"]
	]
	#=======================================================================

	def print_header_protocol
		# Print header
		Cuby.log.print_cuby_logo(:comment => false, :text => 'Atomic charges calculation')
	end

	def prepare_protocol
		# Read geometry
		load_geometry
		# Create calculation
		@calculation = Calculation.new(@name, @settings, @geometry)
		# Prepare calculation
		@calculation.prepare([:energy, :atomic_charges])
	end

	def run_protocol(queue)
		# Submit calculation
		@calculation.send_to_queue(queue)
	end

	def print_protocol
		# Print results
		results.print_energy
		Cuby::log.puts_v_only(:normal,"")
		results.print_energy_decomposition if @settings[:print].include?(:energy_decomposition)

        if @cap_messages && @cap_messages.size > 0
          Cuby::log.puts_v_only(:normal,"Charge capping activated:")
          @cap_messages.each{|s| Cuby::log.puts_v_only(:normal,s) }
          Cuby::log.puts_v_only(:normal,"")
        end

        if @at_correction_messages && @at_correction_messages.size > 0
          Cuby::log.puts_v_only(:normal,"Charge correction activated:")
          @at_correction_messages.each{|s| Cuby::log.puts_v_only(:normal,s) }
          Cuby::log.puts_v_only(:normal,"")
        end

		results.print_atomic_charges(@geometry)
		results.print_dipole if @settings[:print].include?(:dipole)

		Cuby::log.puts_v_only(:debug,"")
		Cuby::log.puts_v_only(:debug,"Total charge: #{'%.3f' % results.atomic_charges.total_charge}")
		Cuby::log.puts_v_only(:debug,"")

		# Print charges on fragments, if fragments are present
		print_fragment_charges

		# Save the atomic charges
		if @settings.set?(:atomic_charges_write)
			case @settings[:atomic_charges_write_format]
			when :charge_list
				results.atomic_charges.write_file(@settings[:atomic_charges_write])
			when :charges_only
				results.atomic_charges.write_file(@settings[:atomic_charges_write], false)
			when :mol2
				mol2geo = @geometry.deep_copy
				results.atomic_charges.write_to_geometry(mol2geo)
				comment = "Energy: #{results.energy} kcal/mol"
				mol2geo.write_file(@settings[:atomic_charges_write], :mol2, {:comment => comment})
			when :pdb
				geo2 = @geometry.deep_copy
				results.atomic_charges.write_to_geometry(geo2)
				geo2.write_file(@settings[:atomic_charges_write], :pdb, {:xyz_charge => true, :extra_columns => @settings[:pdb_extra_columns]})
			end
		end
		# Save as point charges
		if @settings.set?(:atomic_charges_write_pch)
			results.atomic_charges.write_file_xyzc(@settings[:atomic_charges_write_pch], @geometry)
		end
	end

	def cleanup_protocol
		@calculation.cleanup
	end

	def results
		# Results are collected only first time they are needed
		unless @results
			# Get results
			@results = @calculation.results
            if @settings.set?(:atomic_charges_cap)
              cap_charges(@results)
            end
			if @settings.set?(:atomic_charges_corrections)
				atomtype_correction(@results)
			end
		end
		return @results
	end

	#=======================================================================
	# Private
	#=======================================================================
	def print_fragment_charges
		return unless @settings.has_block?(:fragment_1)
		Cuby::log.puts_v_only(:normal,"")
		Cuby::log.puts_v_only(:normal,"Charge on fragments:")
		i = 1
		while @settings.has_block?("fragment_#{i}".to_sym)
			sum = 0.0
			atomlist = @geometry.atomlist_from_selection(@settings["fragment_#{i}".to_sym, :selection])
			atomlist.each{|chi|
				sum += @results.atomic_charges[chi]
			}
			Cuby::log.puts_v_only(:normal, sprintf("%4d%10.4f", i, sum))
			i += 1
		end
	end

	def cap_charges(results)
		conn = Connectivity.new(@geometry)
		fix_ele = @settings.element_hash(:atomic_charges_cap)
		excess = 0.0
		capped_atoms = []
		@cap_messages = []
		@geometry.each_with_index{|atom,i|
			next unless fix_ele.has_key?(atom.element)
			fix_val = fix_ele[atom.element].abs
			ch =  results.atomic_charges[i]
			if ch > fix_val
				dch = ch - fix_val
				results.atomic_charges[i] = fix_val
				@cap_messages << sprintf("   %-3s %3d  %8.3f -> %8.3f", atom.element.to_s, i+1, ch, fix_val)
				if @settings[:atomic_charges_cap_redistribute] == :neighbours
					n = conn.hybridization(i)
					conn.bound_atoms(i).each{|j|
						results.atomic_charges[j] += dch / n
					}
				end
				excess += dch
				capped_atoms << i
			end
			if ch < -fix_val
				dch = ch + fix_val
				results.atomic_charges[i] = -fix_val
				@cap_messages << sprintf("   %-3s %3d  %8.3f -> %8.3f", atom.element.to_s, i+1, ch, -fix_val)
				if @settings[:atomic_charges_cap_redistribute] == :neighbours
					n = conn.hybridization(i)
					conn.bound_atoms(i).each{|j|
						results.atomic_charges[j] += dch / n
					}
				end
				excess += dch
				capped_atoms << i
			end
		}
		case @settings[:atomic_charges_cap_redistribute]
		when :molecule
			n = @geometry.size - capped_atoms.size
			dch = excess / n
			@geometry.each_with_index{|atom,i|
				next if capped_atoms.include?(i)
				results.atomic_charges[i] += dch
			}
			@cap_messages << "Charge distributed to the remainder of the molecule as #{'%.3f' % dch} / atom"
		end
	end

	def atomtype_correction(results)
			connectivity = Connectivity.new(@geometry)
			atomtypes = connectivity.atom_types_simple(@geometry)

			@at_correction_messages = []

			@settings[:atomic_charges_corrections].each_pair{|key, value|
				# Parse atomptype for corrected atom and neigbour
				if key =~ /-.*-/
					Cuby::error("Multiple '-' not allowed in keyword atomic_charges_corrections key")
				elsif key =~ /-/
					atom, neighbour = key.split(/ *- */)
				else
					Cuby::error("Bond (indicated by '-') not found in keyword atomic_charges_corrections key")
				end
				@geometry.each_index{|i|
					if atomtypes[i].match_wildcard(atom)
						# Atom matches, check neigbours
						neighbours = []
						connectivity.bound_atoms(i).each{|j|
							if atomtypes[j].match_wildcard(neighbour)
								neighbours << j
							end
						}
						# Apply correction
						if neighbours.size > 0
							min, max = value.sort
							orig_charge = results.atomic_charges[i]
							if results.atomic_charges[i] < min
								redistribute = results.atomic_charges[i] - min
								results.atomic_charges[i] = min
							end
							if results.atomic_charges[i] > max
								redistribute = max - results.atomic_charges[i]
								results.atomic_charges[i] = max
							end
							neighbours.each{|n|
								results.atomic_charges[n] += redistribute / neighbours.size
							}
							@at_correction_messages << "Charge corrected on atom #{i+1} from #{orig_charge} to #{results.atomic_charges[i]}"
						end
					end
				}
			}
	end

end
