
module ProtocolAtomizationEnergy
	#=======================================================================
	# Protocol header
	#=======================================================================
	PROTOCOL_TYPE = :composite_calculation
	PROTOCOL_DESCRIPTION = "Calculation of atomization energy of a molecule"
	#=======================================================================
	
	def print_header_protocol
		# Print header
		Cuby.log.print_cuby_logo(:comment => false, :text => 'Atomization energy')
	end

	def prepare_protocol
		# Read geometry
		load_geometry

		# Get elements in the system
		@elements = @geometry.count_elements

		# Get element multiplicities
		element_multiplicity = {}
		@elements.each_key{|ele|
			configuration = PeriodicTable.electron_configuration(ele)
			unpaired = 0
			configuration.orbitals.each_index{|i|
				orb = configuration.orbitals[i]
				pop = configuration.occupations[i]
				case orb
				when "s"
					pop = 0 if pop == 2
				when "p"
					pop = 6 - pop if pop > 3
				when "d"
					pop = 10 - pop if pop > 5
				when "f"
					pop = 14 - pop if pop > 7
				end
				unpaired += pop

			}
			element_multiplicity[ele] = unpaired + 1
		}

		# Build energy calculations
		@calculations_element = {}
		@calculations = []

		# First, the elements
		@elements.each_key{|ele|
			blockname = ("calculation_ele_" + ele.to_s.downcase).to_sym
			@settings.new_block(blockname) unless @settings.has_block?(blockname)
			@settings.block(blockname).copy_from!(@settings, :keep, [/^calculation_ele_/])
			settings = @settings.block(blockname)
			settings[:charge] = 0
			settings[:multiplicity] = element_multiplicity[ele]
			geometry = Geometry.new
			geometry << Atom.new(ele)
			calculation = Calculation.new(@name + "_" + ele.to_s, settings, geometry)
			calculation.prepare([:energy])
			@calculations << calculation
			@calculations_element[ele] = calculation
		}

		# Second, the whole system
		blockname = :calculation_ene
		@settings.new_block(blockname)
		@settings.block(blockname).copy_from!(@settings, :keep, [/^calculation_ele_/, :calculation_ene])
		@calculation_ene = Calculation.new(@name + "_mol", settings, @geometry)
		@calculation_ene.prepare([:energy])
		@calculations << @calculation_ene
	end

	def run_protocol(queue)
		# Submit energy calculations
		@calculations.each{|calculation|
			calculation.send_to_queue(queue)
		}
	end

	def print_protocol
		# Print results
		Cuby::log.puts_v_only(:normal,"--------------------------------------------------------------------------------")
		Cuby::log.puts_v_only(:normal,"Molecule energy")
		Cuby::log.puts_v_only(:normal,"--------------------------------------------------------------------------------")
		@calculation_ene.results.print_energy
		Cuby::log.puts_v_only(:normal,"")

		Cuby::log.puts_v_only(:normal,"--------------------------------------------------------------------------------")
		Cuby::log.puts_v_only(:normal,"Element energies")
		Cuby::log.puts_v_only(:normal,"--------------------------------------------------------------------------------")
		@calculations_element.each_pair{|element, calculation|
			Cuby::log.puts_v_only(:normal, sprintf("%4s %16.6f kcal/mol", element.to_s, calculation.results.energy))
		}
		Cuby::log.puts_v_only(:normal,"")

		results.print_energy("Atomization energy")
		
	end

	def cleanup_protocol
		@calculations.each{|calculation|
			calculation.cleanup
		}
	end

	def results
		# Results are collected only first time they are needed
		unless @results
			# Get results
			@results = Results.new
			@results.energy = atomization_energy
		end
		return @results
	end

	#=======================================================================
	# Optional protocol functionality
	#=======================================================================
	
	def single_line_result
		return sprintf("Atomization energy: %20.6f kcal/mol", results.energy)
	end

	def single_number_result
		return results.energy
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def atomization_energy
		# Get atom energy corrections
		atom_energy_corrections = @settings.element_hash(:atom_energy_corrections)
		# Atomization energy
		e_at = -1.0 * (@calculation_ene.results.energy)
		@calculations_element.each_pair{|element, calculation| # + atomic energies
			e_at += @elements[element] * calculation.results.energy
			# Correction, if applicable
			e_at += @elements[element] * atom_energy_corrections[element] if atom_energy_corrections[element]
		}
		return e_at
	end
end
