require "classes/geometry/geometry_series.rb"

module ProtocolCycle
	#=======================================================================
	# Protocol header
	#=======================================================================
	PROTOCOL_TYPE = :batch
	PROTOCOL_DESCRIPTION = "Scan over a series of values of a keyword"
	# Input structure
	INPUT_BLOCKS = [
		InputBlock[:calculation, :required, "Setup for the calculation"]
	]
	#=======================================================================
	
	# TBD:
	# regexp in strings
	# changes in more keywords
	# (combined together, allow e.g. saving geometries from different optimizations to different files)
	# serial execution of jobs, with sequential printing


	def print_header_protocol
		# Print header
		Cuby.log.print_cuby_logo(:comment => false, :text => 'Cycle')
	end

	def prepare_protocol
		# Check the presence of the Calculation block
		unless @settings.has_block?(:calculation)
			Cuby::error("Input for a scan job must contain subsection named 'calculation' which contains the calculation setup")
		end
		unless @settings.set?(:calculation,:job)
			Cuby::error("Job type must be defined in the 'calculation' subsection of the input")
		end

		# Get the keyword to be changed
		if @settings.set?(:cycle_keyword)
			@cycle_keyword = @settings[:cycle_keyword].downcase.to_sym
		else
			@cycle_keyword = nil
		end

		# Determine the mode of operation and load the module implementing it
		case @settings[:cycle_type]
		when :list
			@value_list = cycle_values_from_list
		when :series
			@value_list = cycle_values_from_series
		when :files
			@value_list = cycle_values_files
		when :geometry
			@value_list = cycle_values_geometry
		end

		# Build list of values for the templates
		@template_value_lists = {}
		@settings[:cycle_templates].each_pair{|kw, template|
			# Fix keyword
			kw = kw.downcase.to_sym
			# Build values
			values = @value_list.map{|x| template.to_s.gsub("%",x.to_s)}
			# Check/fix values
			kw_definition = Settings.keyword(kw)
			values.map!{|x| Settings.fix_value(x, kw_definition, kw)}
			# Save values
			@template_value_lists[kw] = values
		}

		# Check/fix the values
		if @cycle_keyword
			kw_definition = Settings.keyword(@cycle_keyword)
			@value_list.each_index{|i|
				@value_list[i] = Settings.fix_value(@value_list[i], kw_definition, @cycle_keyword)
			}
		end

		# Check - either a keyword or a template must exist
		if !@cycle_keyword && @template_value_lists == 0
			Cuby::error("Cycle protocol: nothing to cycle, set keyword cycle_keyword or cycle_templates")
		end

		# Create child jobs
		@item_jobs = []
		@value_list.each_index{|i|
			# Create block for each value
			name = sprintf("cycle%03d",i)
			blockname = ("calculation_" + name).to_sym
			@settings.new_block(blockname)
			@settings.block(blockname).copy_from!(@settings.block(:calculation), :keep, [/^calculation_cycle/])
			settings = @settings.block(blockname)
			# Change the keyword and check if it was found
			if  @cycle_keyword
				unless cycle_set_keyword(settings, @cycle_keyword, @value_list[i])
					Cuby::error("Cycle protocol can not set the value of a keyword '#{@cycle_keyword}', it was not found in the input")
				end
			end
			# Change other keywords using templates
			@template_value_lists.each_pair{|kw, values|
				unless cycle_set_keyword(settings, kw, values[i])
					Cuby::error("Cycle protocol can not set the value of a keyword '#{kw}', it was not found in the input")
				end
			}
			# Add it to the list of jobs
			@item_jobs[i] = Job.new(@name + "_" + name, settings)
		}

		# Prepare child jobs
		@item_jobs.each{|job| job.prepare }

	end

	def run_protocol(queue)
		# Submit jobs to queue
		@item_jobs.each_index{|i| 
			if @settings[:cycle_header_before_run]
				Cuby::log.puts "Cycle #{i+1}/#{@item_jobs.size}"
			end
			job = @item_jobs[i]
			job.run
		}
	end

	def cleanup_protocol
		# Clean child jobs
		@item_jobs.each{|job| job.cleanup}
	end

	def print_protocol
		# Keys as strings
		keys = []
		@item_jobs.each_index{|i|
			if @value_list[i].class == Float
				keys << "%-20.6f" % @value_list[i]
			elsif @value_list[i].kind_of? Numeric
				keys << "%-20d" % @value_list[i]
			else
				keys << "%-20s" % @value_list[i].to_s
			end
		}
		# Print results
		if @settings[:cycle_print_mode] == :numbers && @item_jobs.first.respond_to?(:single_number_result)
			# simple oneline printing
			@item_jobs.each_index{|i|
				s = keys[i]
				s += "%20.6f" % @item_jobs[i].single_number_result
				Cuby::log.puts(s)
			}
		elsif @settings[:cycle_print_mode] == :simple && @item_jobs.first.respond_to?(:single_line_result)
			# simple oneline printing
			@item_jobs.each_index{|i|
				s = "%20s: "  % @cycle_keyword
				s += keys[i]
				s += @item_jobs[i].single_line_result
				Cuby::log.puts(s)
			}
		elsif @settings[:cycle_print_mode] == :full
			# printing using child protocol's print
			@value_list.each_index{|i|
				unless @settings[:cycle_type] == :geometry
					# Do not print geometries
					Cuby::log.puts "-" * 80
					Cuby::log.puts("#{@cycle_keyword}: #{@value_list[i]}")
				end
				@item_jobs[i].print
			}
		end
	end

	def results
		# Results are collected only first time they are needed
		unless @results
			# Get results
			@results = []
			@item_jobs.each{|job|
				@results << job.results
			}
		end
		return @results
	end

	#=======================================================================
	# Private methods
	#=======================================================================
	
	def cycle_set_keyword(settings, keyword, value)
		# Report if the keyword was actually set
		was_set = false
		# Set it at root level
		if settings.set?(keyword)
			settings[keyword] = value
			was_set = true
		end
		# Apply it recursively to all child blocks
		settings.blocks.each_pair{|name, childblock|
			was_set |= cycle_set_keyword(childblock, keyword, value)
		}
		return was_set
	end
	
	def cycle_values_from_list
		return @settings[:cycle_list]
	end

	def cycle_values_from_series
		series = []
		tolerance_end = @settings[:cycle_series_step] * 1.0e-3
		x = @settings[:cycle_series_start]
		while x <= @settings[:cycle_series_end] + tolerance_end
			series << x
			x += @settings[:cycle_series_step]
		end
		return series
	end

	def cycle_values_files
		files = []
		@settings[:cycle_files].split(/ *; */).each{|expr|
			files |= Dir.glob(expr)
		}
		return files
	end

	def cycle_values_geometry
		# Load a series from geometry keyword
		series = GeometrySeries.from_settings(@settings)
		series.size_check
		return series.map{|geo| geo.to_s}
	end
end
