module ActionsDifferences

	def rmsd_fit
		# Geometry2 is used as a template
		geometry2 = Geometry.from_file(@settings[:geometry2], @settings)

		if @settings.set?(:selection) && @settings[:selection] != 'auto'
			@messages << "Keyword 'selection' applied also to geometry2"
			geometry2 = geometry2.geometry_from_selection(@settings[:selection])
		end

		# Check if geometries are the same
		must_be_the_same!(geometry2)

		@messages << "RMSD before fit: #{@geometry.rmsd_noshift(geometry2)} A"
		@messages << "RMSD translated: #{@geometry.rmsd(geometry2)} A"
		fitted = @geometry.fit_rmsd(geometry2)
		@messages << "RMSD after fit:  #{fitted.rmsd(geometry2)} A"
		@geometry = fitted
	end

	def rmsd_fit_subset
		# Geometry2 is used as a template
		geometry2 = Geometry.from_file(@settings[:geometry2], @settings)

		if @settings.set?(:geometry_action_selection_array)
			@messages << "Working with selection from 'geometry_action_selection_array'"
			list1 = []
			list2 = []
			@settings[:geometry_action_selection_array].each{|sel|
				list1 = list1 + @geometry.atomlist_from_selection(sel)
				list2 = list2 + geometry2.atomlist_from_selection(sel)
			}
		elsif @settings.set?(:geometry_action_selection)
			@messages << "Working with selection from 'geometry_action_selection'"
			list1 = @geometry.atomlist_from_selection(@settings[:geometry_action_selection])
			list2 = geometry2.atomlist_from_selection(@settings[:geometry_action_selection])
		else
		end


		g_sel =  @geometry.geometry_from_list(list1)
		g2_sel = geometry2.geometry_from_list(list2)
		@messages << "RMSD before fit: #{g_sel.rmsd_noshift(g2_sel)} A"

		# Check if geometries are the same
		must_be_the_same!(g2_sel, g_sel)

		# Translate the geometries
		center_this  = g_sel.weighted_center(:unit)
		center_template = g2_sel.weighted_center(:unit)
		shifted_this  = g_sel.deep_copy.translate(center_this * -1.0)
		shifted_template = g2_sel.deep_copy.translate(center_template * -1.0)
		@messages << "RMSD translated: #{shifted_this.rmsd_noshift(shifted_template)} A"

		# Get RMSD transformation
		r_mat = shifted_this.quatfit_rotator(shifted_template)

		# Rotate
		@geometry.translate!(center_this * -1.0)
		@geometry.each {|atom| atom.rotate_by_matrix!(r_mat)}
		@geometry.translate!(center_template)

		# Final RMSD
		g_sel =  @geometry.geometry_from_list(list1)
		g2_sel = geometry2.geometry_from_list(list2)
		@messages << "RMSD after fit: #{g_sel.rmsd_noshift(g2_sel)} A"
	end

	def print_differences
		@write_geo = false # This action does not write geometry file

		# Calculate differences
		diffs = get_differences

		# Print differences	
		max = diffs.max
		@messages << "Differences in atom positions (A):"
		@geometry.each_with_index{|atom,i|
			msg = ""
			if diffs[i] == max && max > 0
				msg = "   <- max"
			end
			@messages << sprintf("%6d %-3s %10.4f", i+1, atom.element.to_s, diffs[i]) + msg
		}
	end

	def differences_as_geometry
		# Calculate differences
		diffs = get_differences

		diffgeo = Geometry.new

		@geometry.each_with_index{|atom,i|
			if diffs[i] >= @settings[:position_difference_threshold]
				diffgeo << atom
			end
		}

		@geometry = diffgeo
	end

	def differences_as_selection
		@write_geo = false # This action does not write geometry file

		# Calculate differences
		diffs = get_differences

		atomlist = []

		@geometry.each_with_index{|atom,i|
			if diffs[i] >= @settings[:position_difference_threshold]
				atomlist << i
			end
		}

		@messages << "Position of following atoms differ by more than #{'%.4f' % @settings[:position_difference_threshold]}"
		@messages <<  GeometrySelections.atomlist_to_string(atomlist)

	end

	#=======================================================================
	# Private
	#=======================================================================
	
	def get_differences
		# Get second geometry	
		geometry2 = Geometry.from_file(@settings[:geometry2], @settings)
		if @settings.set?(:selection) && @settings[:selection] != 'auto'
			@messages << "Keyword 'selection' applied also to geometry2"
			geometry2 = geometry2.geometry_from_selection(@settings[:selection])
		end

		# Check if geometries are the same
		must_be_the_same!(geometry2)
		
		# Calculate differences
		diffs = []
		@geometry.each_with_index{|atom,i|
			atom2 = geometry2[i]
			diffs[i] = (atom - atom2).abs
		}

		return diffs
	end
end
