module ActionsOrientation

	#=======================================================================
	# Actions - orienting molecules
	#=======================================================================
	
	def orient
		# Separate selections
		selections = @settings[:geometry_orientation].strip.split(";")
		unless selections.size == 3
			Cuby::error "Keyword geometry_orientation must define three selections"
		end

		# Convert them to atomlists
		atomlists = selections.map{|s|
			list = @geometry.atomlist_from_selection(s)
			Cuby::error "Selection in geometry_orientation selects no atoms" unless list.size > 0
			list
		}

		@geometry.orient!(atomlists[0],atomlists[1],atomlists[2])
	end

	#=======================================================================
	# Transformations
	#=======================================================================
	
	def rotate_around_bond
		# Select the bond
		sel1, sel2 = @settings[:bond_selection].split(/; */)
		sel1 = @geometry.atomlist_from_selection(sel1)
		sel2 = @geometry.atomlist_from_selection(sel2)
		Cuby::error "More than one atom selected in keyword bond_selection" if sel1.size > 1
		Cuby::error "More than one atom selected in keyword bond_selection" if sel2.size > 1
		Cuby::error "No tom selected in keyword bond_selection" if sel1.size < 1
		Cuby::error "No tom selected in keyword bond_selection" if sel2.size < 1
		sel1 = sel1[0]
		sel2 = sel2[0]

		# Build connectivity
		conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
		origsize = conn.molecules.size

		# Remove the bond from the connectivity
		conn.remove_bond(sel1, sel2)
		parts = conn.molecules
		unless parts.size == origsize + 1
			Cuby::error "Geometry action 'rotate_around_bond': the bond is part of a cycle"
		end

		# Find the part to rotate; it is the one containing the second atom
		rotpart = nil
		parts.each{|part|
			if part.include?(sel2)
				rotpart = part
				break
			end
		}


		# Create geometry for rotated part
		rot_geo = @geometry.geometry_from_list(rotpart)

		# Axis and angle
		axis = (@geometry[sel1] - @geometry[sel2]).normalize
		angle = @settings[:rot_angle] / 180.0 * Math::PI

		# Rotate
		rot_geo.rotate_angle_axis_center!(angle, axis, @geometry[sel1])
	end

	def rotate_xyz
		# Rotated part of the geometry
		rotgeo = @geometry.geometry_from_selection(@settings[:geometry_action_selection])

		# Rotation angles
		angles = @settings[:rot_angles_xyz].split(/\s*[,;]\s*/).map{|x| x.to_f / 180.0 * Math::PI}
		unless angles.size == 3
			Cuby::error "Keyword 'rot_angles_xyz' should provide three comma-separated values"
		end

		# Perform the rotations
		rotgeo.rotate_angle_axis!(angles[0], Coordinate[1,0,0])
		rotgeo.rotate_angle_axis!(angles[1], Coordinate[0,1,0])
		rotgeo.rotate_angle_axis!(angles[2], Coordinate[0,0,1])
	end

	def randomize_intermolecular
		angle = @settings[:rot_angle] / 180.0 * Math::PI

		# Get molecules
		conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
		molecules = conn.molecules
		geoms = conn.molecules.map{|list| @geometry.geometry_from_list(list) }

		if geoms.size != 2
			@messages << "not_dimer(#{geoms.size})"
			return
		end

		# Translation - move the molecule2 safely apart
		dist, i, j = @geometry.closest_contact(molecules[0], molecules[1])
		axis = @geometry[j] - @geometry[i]
		geoms[1].translate!(axis.normalize * 1000.0)

		# Rotations in cartesian space - iterate over molecules
		geoms.each_index{|m|
			com = geoms[m].com
			@geometry.translate!(-com)
			begin
				angle_range =  (-angle .. angle)
				rx = rand(angle_range)
				ry = rand(angle_range)
				rz = rand(angle_range)
			rescue
				# Fix for old ruby where rand does not accept range
				rx = (rand - 0.5) * 2 * angle
				ry = (rand - 0.5) * 2 * angle
				rz = (rand - 0.5) * 2 * angle
			end
			geoms[m].rotate_angle_axis!(rx, Coordinate[1,0,0])
			geoms[m].rotate_angle_axis!(ry, Coordinate[0,1,0])
			geoms[m].rotate_angle_axis!(rz, Coordinate[0,0,1])
		}

		# Get new intermol distance
		dist2, i, j = @geometry.closest_contact(molecules[0], molecules[1])
		axis = @geometry[j] - @geometry[i]
		rvdw = PeriodicTable.vdw_radius(@geometry[j].element) + PeriodicTable.vdw_radius(@geometry[i].element)
		
		# Move molecule 2 to vdw contact
		geoms[1].translate!(axis.normalize * -1 * (axis.abs - rvdw))

		# Center whole system
		com = @geometry.com
		@geometry.translate!(-com)
	end


end
