
module ProtocolGeometry
	#=======================================================================
	# Protocol header
	#=======================================================================
	PROTOCOL_TYPE = :tools
	PROTOCOL_DESCRIPTION = "Geometry manipulations"
	#=======================================================================
	
	require "protocols/geometry/classes/actions_atomtype.rb"
	include ActionsAtomtype
	require "protocols/geometry/classes/actions_zmat.rb"
	include ActionsZmat
	require "protocols/geometry/classes/actions_elements.rb"
	include ActionsElements
	require "protocols/geometry/classes/actions_connectors.rb"
	include ActionsConnectors
	require "protocols/geometry/classes/actions_orientation.rb"
	include ActionsOrientation
	require "protocols/geometry/classes/actions_molecules.rb"
	include ActionsMolecules
	require "protocols/geometry/classes/actions_series.rb"
	include ActionsSeries
	require "protocols/geometry/classes/actions_symmetry.rb"
	include ActionsSymmetry
	require "protocols/geometry/classes/actions_topology.rb"
	include ActionsTopology
	require "protocols/geometry/classes/actions_differences.rb"
	include ActionsDifferences
	require "protocols/geometry/classes/actions_noncovalent.rb"
	include ActionsNoncovalent


	def print_header_protocol
		# Print header
		Cuby.log.print_cuby_logo(:comment => false, :text => 'Geometry manipulation')
	end

	def prepare_protocol
		# Read geometry
		load_geometry

		# Messages to be printed in Print
		@messages = []

		# Write or not - default is yes
		@write_geo = true

		# Save geometry info
		if geometry.info[:xyz_remark]
			@geometry.info[:protocol_comment] = geometry.info[:xyz_remark]
		end

		# Perform action
		case @settings[:geometry_action]
		when :cat
			concatenate
		when :center
			center
		when :standard_orientation
			standard_orientation
		when :partial_pdb_update
			partial_pdb_update
		when :move_closest_contact_vdw_dist
			move_closest_contact_vdw_dist
		when :pdb_renumber
			pdb_renumber
		when :pdb_renumber_molecules
			pdb_renumber_molecules
		when :pdb_selection_as_chain
			pdb_selection_as_chain
		when :pdb_rename_residues
			pdb_rename_residues
		when :pdb_charge
			pdb_charge
		when :pdb_sort_residues
			pdb_sort_residues
		when :distribute_charge
			distribute_charge
		when :selection_to_atomlist
			selection_to_atomlist
		when :connect
			connect_connectors
		when :order_random
			order_random
		when :order_topo
			order_by_topology
		when :order_chain
			order_chain
		when :order_molecules
			order_molecules
		when :order_template
			order_template
		when :rmsd_fit
			rmsd_fit
		when :rmsd_fit_subset
			rmsd_fit_subset
		when :rmsd_fit_reorder
			rmsd_fit_topo
		when :differences
			print_differences
		when :diff_as_geo
			differences_as_geometry
		when :diff_as_selection
			differences_as_selection
		when :topologic_ranks
			topologic_ranks
		when :atomtypes
			atomtype_print
		when :atomtype_bonds
			atomtype_bonds
		when :atomtype_angles
			atomtype_angles
		when :build_zmat
			build_zmat
		when :zmat_update
			zmat_update
		when :moments_of_inertia
			moments_of_inertia
		when :element_list
			element_list
		when :element_counts
			element_counts
		when :element_match
			element_match
		when :atomization_energy
			atomization_energy
		when :add_connector
			add_connector
		when :set_connector_dist
			connector_set_distance
		when :symmetry
			print_symmetry
		when :symmetrize
			symmetrize
		when :orient
			orient
		when :rotate_around_bond
			rotate_around_bond
		when :rotate_xyz
			rotate_xyz
		when :randomize
			randomize
		when :random_translation
			random_translation
		when :randomize_intermolecular
			randomize_intermolecular
		when :sort_mol_by_distance
			sort_mol_by_distance
		when :expand_series
			expand_series
		when :molden
			visualize_molden
		when :print_valence
			print_valence
		when :checksum
			print_checksum
		when :print_center
			print_center
		when :close_contacts
			check_close_contacts
		when :get_info
			print_info
		when :list_hbonds
			print_hbonds
		when :displacement_in_zmat
			displacement_in_zmat
		when :series_energy_minimum
			series_energy_minimum
		when :list_nc_hbonds
			list_nc_hbonds
		when :list_nc_sigmahole_bonds
			list_nc_sigmahole_bonds
		when :list_nc_closest_vdw
			list_nc_closest_vdw
		when :update_charges
			update_charges
		when :sum_charges
			sum_charges
		when :sdf_check
			sdf_check
		when :sum_site_charges
			sum_site_charges
		end

		# Save geometry
		if @write_geo && @settings[:geometry_write] != ""
			filename = write_geometry
			@messages << "Geometry written to file '#{filename}'"
		end
	end

	def run_protocol(queue)
	end

	def print_protocol
		@messages.each{|message|
			Cuby::log.puts_v(:brief,  message)
		}
		Cuby::log.puts_v(:normal, "")
	end

	def cleanup_protocol
	end

	def results
		return nil
	end

	#=======================================================================
	# Helpers
	#=======================================================================
	def must_be_pdb!(geometry, name = "")
		unless  geometry[0].properties[:pdb_atom_name]
			Cuby::error("#{name} must be a PDB.")
		end
	end

	def must_be_the_same!(geometry2, geometry1 = nil)
		if !geometry1
			geometry1  = @geometry
		end

		unless geometry1.size == geometry2.size
			Cuby::error "The geometries must be of same size, now they have #{geometry1.size} and #{geometry2.size} aroms"
		end

		geometry1.each_with_index{|atom,i|
			atom2 = geometry2[i]
			unless atom.element == atom2.element
				Cuby::error "Geometry and geometry2 differ in element of atom #{i+1} (#{atom.element} vs. #{atom2.element})"
			end
		}
	end

	#=======================================================================
	# Actions
	#=======================================================================
	
	def concatenate
		@geometry.last.properties[:terafter] = true

		# Load secondary geometry
		geometry2 = Geometry.from_file(@settings[:geometry2], @settings)

		mol_a_end = @geometry.size
		selection1 = "1-#{mol_a_end}"

		geometry2.each{|atom|
			@geometry << atom
		}
		selection2 = "#{mol_a_end+1}-#{@geometry.size}"

		# Save selections to be printed in the file 
		@geometry.info[:protocol_comment] = "selection_a=#{selection1} selection_b=#{selection2}"
	end

	def center
		c = @geometry.center
		@geometry.translate!(-c)
	end

	def print_center
		@write_geo = false # This action does not write geometry file
		c = @geometry.center
		@messages << "#{c.x} #{c.y} #{c.z}"
	end

	def standard_orientation
		@geometry.standard_orientation!
	end
	
	def partial_pdb_update
		# Load secondary geometry
		geometry2 = Geometry.from_file(@settings[:geometry2])
		# Both must be PDBs
		must_be_pdb!(@geometry, "Primary geometry")
		must_be_pdb!(geometry2, "Geometry2")

		matched_count = 0;

		@geometry.each{|atom|
			geometry2.each{|atom2|
				if atom.properties[:pdb_atom_no] == atom2.properties[:pdb_atom_no] &&
				   atom.properties[:pdb_atom_name] == atom2.properties[:pdb_atom_name] &&
				   atom.properties[:pdb_res_no] == atom2.properties[:pdb_res_no] &&
				   atom.properties[:pdb_res_name] == atom2.properties[:pdb_res_name]
					matched_count += 1;
					atom.set_coord(atom2);
				end

			}
		}

		@messages << "Coordinates of #{matched_count} atoms updated"
	end

	def move_closest_contact_vdw_dist
		# Search for two molecules
		connectivity = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
		molecules = connectivity.molecules
		Cuby::error ("Geometry / move_closest_contact_vdw_dist: Only one molecule found") if molecules.size == 1
		Cuby::error ("Geometry / move_closest_contact_vdw_dist: More than two molecules found") if molecules.size == 1
		# Get closest contact and indexes
		distance, i, j = @geometry.closest_contact(molecules[0], molecules[1])
		# Compare with vdW distance
		vdw = PeriodicTable.vdw_radius(@geometry[i].element) + PeriodicTable.vdw_radius(@geometry[j].element)
		shift = distance - vdw
		direction = (@geometry[j] - @geometry[i]) / distance * -1
		shift = direction * shift
		# Make geometry for the second molecule
		mol2 = @geometry.geometry_from_list(molecules[1])
		mol2.translate!(shift)
		@messages << "Molecules moved from distance #{'%.2f' % distance} to #{'%.2f' % @geometry[i].distance(@geometry[j])} A"
	end

	def selection_to_atomlist
		@write_geo = false # This action does not write geometry file

		atomlist = @geometry.atomlist_from_selection(@settings[:geometry_action_selection])
		@messages <<  GeometrySelections.atomlist_to_string(atomlist)
	end

	def moments_of_inertia
		@write_geo = false # This action does not write geometry file

		axes, eigenvalues = geometry.moment_of_inertia_diagonalized

		if @settings.set?(:geometry2)
			# When a second geometry is provided, calculate differences
			geometry2 = Geometry.from_file(@settings[:geometry2])
			axes2, eigenvalues2 = geometry2.moment_of_inertia_diagonalized

			differences = []
			eigenvalues.each_index{|i|
				differences << eigenvalues[i] - eigenvalues2[i]
			}

			@messages << "Difference of principal moments of inertia, g/mol * A^2"
			differences.each{|x| @messages << sprintf("%16.6f", x * UNIT2GMOL)}
		else
			# Just print the moments
			@messages << "Principal moments of inertia, g/mol * A^2"
			eigenvalues.each{|x| @messages << sprintf("%16.6f", x * UNIT2GMOL)}
		end
	end

	def check_close_contacts
		@write_geo = false # This action does not write geometry file
		conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])

		vdw_limit = 0.8

		conn.each_nonbonded_above_1_4{|i,j|
			a1 = @geometry.at(i)
			a2 = @geometry.at(j)

			distance = a1.distance(a2)
			vdw_distance = PeriodicTable.vdw_radius(a1.element) + PeriodicTable.vdw_radius(a2.element)
			if distance < vdw_limit * vdw_distance
				printf("%8d%8d%4s%4s    distance: %8.3f  =  %6.3f vdW radii sum",
				       i+1,
				       j+1,
				       a1.element.to_s,
				       a2.element.to_s,
				       distance,
				       distance / vdw_distance
				      )
				if conn.is_1_5?(i,j)
					printf("    1-5") 
				end
				if conn.is_h_bond?(geometry, i,j, 2.5, 60.0, [:O,:N,:S])
					printf("    H-bond")
				end
				puts
			end


		}
	end

	#=======================================================================
	# Actions - PDB
	#=======================================================================
	
	def pdb_renumber
		@geometry.renumber_pdb_atoms!
		@geometry.renumber_pdb_residues!(@settings[:pdb_resno_start])
	end

	def pdb_renumber_molecules
		@geometry.renumber_pdb_atoms!
		conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
		molecules = conn.molecules
		molecules.each_index{|m|
			molecules[m].each{|i|
				@geometry[i].properties[:pdb_res_no] = m+1
			}

		}
	end

	def pdb_selection_as_chain
		chain = @settings[:pdb_set_chain]
		Cuby::error "PDB chain must be one character" unless chain.length == 1
		g2 = @geometry.geometry_from_selection(@settings[:geometry_action_selection])
		g2.each{|atom|
			atom.properties[:pdb_chain] = chain
		}
		@messages << "PDB chain set to #{chain} on #{g2.size} atoms"
	end

	def pdb_rename_residues
		count = 0
		@settings[:pdb_rename_residues].each_pair{|selection, newname|
			g2 = @geometry.geometry_from_selection(selection)
			count += g2.size
			g2.each{|atom|
				atom.properties[:pdb_res_name] = newname
			}
		}
		@messages << "PDB residue names changed for #{count} atoms"
	end

	def pdb_sort_residues
		@geometry.sort!{|a,b| a.properties[:pdb_res_no] <=> b.properties[:pdb_res_no]}
	end

	def pdb_charge
		@write_geo = false
		charge = @geometry.pdb_protein_charge

		# Print result
		case @settings[:verbosity] 
		when :minimal
			Cuby::log.puts(charge)
		else
			Cuby::log.puts("PDB charge: #{charge}")
		end
	end

	#=======================================================================
	# Atomic charges
	#=======================================================================
	
	def distribute_charge
		at_c = @settings[:charge].to_f / @geometry.size
		@geometry.each{|atom|
			atom.properties[:charge] = at_c
		}
	end


	#=======================================================================
	# Actions - ordering
	#=======================================================================
	
	def order_random
		@geometry.shuffle!
	end

	def order_molecules
		newgeo = Geometry.new
		conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
		conn.molecules.each{|molecule|
			molecule.each{|i|
				newgeo << @geometry[i]
			}
		}
		@geometry = newgeo
	end

	def order_chain
		# Split geometry into terminal an backbone
		conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])

		Cuby::error "Not implemented yet"

	end

	#=======================================================================
	# Actions - randomization
	#=======================================================================
	
	def randomize
		@geometry.each_index{|i|
			f = 2.0 * @settings[:geometry_randomize_interval]
			dc = Coordinate.new((rand() - 0.5)*f, (rand() - 0.5)*f, (rand() - 0.5)*f)
			@geometry[i].plus!(dc)
		}
	end

	def random_translation
		f = 2.0 * @settings[:geometry_randomize_interval]
		dc = Coordinate.new((rand() - 0.5)*f, (rand() - 0.5)*f, (rand() - 0.5)*f)
		@geometry.each_index{|i|
			@geometry[i].plus!(dc)
		}
	end

	#=======================================================================
	# Atomic charges
	#=======================================================================
	
	def update_charges
		charges = IO.readlines(@settings[:atomic_charges_read])
		if @geometry.size != charges.size
			Cuby::error("List of atomic charges has #{charges.size} entries, geometry has #{@geometry.size}")
		end
		@geometry.each_index{|i|
			@geometry[i].properties[:charge] = charges[i].to_f
		}
	end

	def sum_charges
		@write_geo = false
		charge = 0.0
		@geometry.each_index{|i|
			if @geometry[i].properties[:charge]
				charge += @geometry[i].properties[:charge] 
			else
				Cuby::error("Atom no. #{i+1} does not have atomic charge information")
			end
		}
		@messages << "Net charge:     #{'%.3f' % charge}"
		@messages << "Rounded charge: #{charge.round}"
	end

	
	#=======================================================================
	# Visualization
	#=======================================================================
	
	def visualize_molden
		@write_geo = false
		molden_file = Cuby.make_tmpname('molden_')
		@geometry.write_xyz(:file => molden_file)
		`molden #{molden_file}; rm #{molden_file}`
	end

	#=======================================================================
	# Printing
	#=======================================================================
	
	def print_valence
		@write_geo = false
		Cuby::log.puts "Valence numbers from connectivity:"
		conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
		val = []
		@geometry.each_index{|i| val[i] = conn.hybridization(i)}
		Cuby::log.puts "[ " + val.join(", ") + " ]"
	end

	def print_checksum
		@write_geo = false
		@messages << "Checksum: " + @geometry.checksum
	end

	def print_info
		@write_geo = false
		if info = @geometry.get_extended_info
			@messages << info.to_s
		else
			@messages << "No system information available in the geometry file."
		end
	end

	def print_hbonds
		@write_geo = false
		conn = Connectivity.new(@geometry, @settings[:connectivity_tolerance])
		molecules = conn.molecules
		hbonds = []
		conn.each_h_bond(@geometry, 2.5, 60.0, [:O, :N, :C]){|d,h,a|
			d_a = @geometry[d]
			h_a = @geometry[h]
			a_a = @geometry[a]
			elements = "#{d_a.element}#{h_a.element}-#{a_a.element}"
			distance = h_a.distance(a_a)
			reldist = distance / (PeriodicTable.vdw_radius(h_a.element) + PeriodicTable.vdw_radius(a_a.element)) * 100
			angle = Coordinate.angle(d_a, h_a, a_a) / Math::PI * 180


			# Molecules
			m1 = nil
			m2 = nil
			molecules.each_index{|i|
				m1 = i if molecules[i].include?(d)
				m2 = i if molecules[i].include?(a)
			}
			text = ""
			text = "intermolecular" if m1 != m2

			hbonds << [reldist, sprintf("%-8s%10.3f%8.1f%8.1f  %-20s", elements, distance, reldist, angle, text)]
		}
		hbonds.sort!{|a,b| a[0] <=> b[0]}.reverse
		@messages << "H-bonds found: #{hbonds.size}"
		if hbonds.size > 0
			@messages << sprintf("%-8s%10s%8s%8s", "type", "dist.", "%vdW", "angle")
		end
		hbonds.each{|hb| @messages << hb[1]}
	end

	def sdf_check
		valences = {
			:H => [1],
			:N => [3],
			:C => [4],
			:O => [2],
			:S => [2,6],
			:F => [1],
			:Cl => [1],
			:Br => [1],
			:I => [1]
		}

		@write_geo = false
		unless geometry.info[:sdf_bonds]
			Cuby::error("Action 'sdf_check' works only on sdf files")
		end

		# Calculate valence
		valence = Array.new(@geometry.size, 0)
		geometry.info[:sdf_bonds].each_pair{|atoms, order|
			a1, a2 = atoms
			valence[a1 - 1] += order
			valence[a2 - 1] += order
		}

		# Adjust valence for charge,
		# get total charge
		total_charge = 0
		geometry.info[:sdf_charges].each_pair{|index, chg|
			valence[index] -= chg
			total_charge += chg
		}

		# Check and report valence
		@geometry.each_with_index{|atom,i|
			if valences[atom.element] # Element is tabulated
				if not valences[atom.element].include?(valence[i])
					chg = ""
					if geometry.info[:sdf_charges][i]
						chg = geometry.info[:sdf_charges][i]
						if chg > 0
							chg="(+#{chg})"
						else
							chg="(#{chg})"
						end
					end
					puts "Error in valence, atom #{i+1}: #{atom.element} #{valence[i]} #{chg}"
				end
			end
		}

		# Check total charge
		if geometry.info[:sdf_data]['charge']
			if geometry.info[:sdf_data]['charge'].to_i != total_charge
				puts "Error: charge data field present but does not match atom charges"
			end
		end

	end

	def sum_site_charges

		@write_geo = false
		unless geometry.info[:sdf_charges]
			Cuby::error("Action 'sum_site_charges' works only on sdf files")
		end

		# Get total charge
		total_charge = 0
		abs_charge = 0
		geometry.info[:sdf_charges].each_pair{|index, chg|
			total_charge += chg
			abs_charge += chg.abs
		}

		if abs_charge == total_charge.abs
			puts total_charge
		else
			puts "#{total_charge}*"
		end
	end

end
